@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2015-2022 The Fluent Bit Authors
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -370,6 +370,8 @@ struct flb_output_instance {
     struct cmt_gauge   *cmt_upstream_total_connections;
     /* m: output_upstream_busy_connections */
     struct cmt_gauge   *cmt_upstream_busy_connections;
+    /* m: output_chunk_available_capacity_percent */
+    struct cmt_gauge   *cmt_chunk_available_capacity_percent;
 
     /* OLD Metrics API */
 #ifdef FLB_HAVE_METRICS
