@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -51,7 +50,13 @@
 #define FLB_TASK_SET(ret, task_id, out_id)              \
     (uint32_t) ((ret << 28) | (task_id << 14) | out_id)
 
+/* Route status */
+#define FLB_TASK_ROUTE_INACTIVE 0
+#define FLB_TASK_ROUTE_ACTIVE   1
+#define FLB_TASK_ROUTE_DROPPED  2
+
 struct flb_task_route {
+    int status;
     struct flb_output_instance *out;
     struct mk_list _head;
 };
@@ -73,23 +78,49 @@ struct flb_task_retry {
 
 /* A task takes a buffer and sync input and output instances to handle it */
 struct flb_task {
-    int id;                             /* task id                   */
-    uint64_t ref_id;                    /* external reference id     */
-    uint8_t status;                     /* new task or running ?     */
-    int users;                          /* number of users (threads) */
-    char *tag;                          /* record tag                */
-    int tag_len;                        /* tag length                */
-    const char *buf;                    /* buffer                    */
-    size_t size;                        /* buffer data size          */
-    void *ic;                           /* input chunk */
+    int id;                              /* task id                   */
+    uint64_t ref_id;                     /* external reference id     */
+    uint8_t status;                      /* new task or running ?     */
+    int users;                           /* number of users (threads) */
+    struct flb_event_chunk *event_chunk; /* event chunk context       */
+    void *ic;                            /* input chunk context       */
 #ifdef FLB_HAVE_METRICS
-    int records;                        /* numbers of records in 'buf'   */
+    int records;                         /* numbers of records in 'buf'   */
 #endif
-    struct mk_list routes;              /* routes to dispatch data       */
-    struct mk_list retries;             /* queued in-memory retries      */
-    struct mk_list _head;               /* link to input_instance        */
-    struct flb_input_instance *i_ins;   /* input instance                */
-    struct flb_config *config;          /* parent flb config             */
+    struct mk_list routes;               /* routes to dispatch data       */
+    struct mk_list retries;              /* queued in-memory retries      */
+    struct mk_list _head;                /* link to input_instance        */
+    struct flb_input_instance *i_ins;    /* input instance                */
+    struct flb_config *config;           /* parent flb config             */
+    pthread_mutex_t lock;
+};
+
+/*
+ * A queue of flb_task_enqueued tasks
+ *
+ * This structure is currently used to track pending flushes when FLB_OUTPUT_SYNCHRONOUS
+ * is used.
+ */
+struct flb_task_queue {
+    struct mk_list pending;
+    struct mk_list in_progress;
+};
+
+/*
+ * An enqueued task is a task that is not yet dispatched to a thread
+ * or started on the engine.
+ *
+ * There may be multiple enqueued instances of the same task on different out instances.
+ *
+ * This structure is currently used to track pending flushes when FLB_OUTPUT_SYNCHRONOUS
+ * is used.
+ */
+struct flb_task_enqueued {
+    struct flb_task *task;
+    struct flb_task_retry *retry;
+    struct flb_output_instance *out_instance;
+    struct flb_config *config;
+    struct mk_list _head;
 };
 
 int flb_task_running_count(struct flb_config *config);
@@ -108,6 +139,8 @@ void flb_task_add_coro(struct flb_task *task, struct flb_coro *coro);
 
 void flb_task_destroy(struct flb_task *task, int del);
 
+struct flb_task_queue* flb_task_queue_create();
+void flb_task_queue_destroy(struct flb_task_queue *queue);
 struct flb_task_retry *flb_task_retry_create(struct flb_task *task,
                                              struct flb_output_instance *ins);
 
@@ -151,5 +184,96 @@ static inline void flb_task_users_dec(struct flb_task *task, int release_check)
     }
 }
 
+static inline void flb_task_acquire_lock(struct flb_task *task)
+{
+    pthread_mutex_lock(&task->lock);
+}
+
+static inline void flb_task_release_lock(struct flb_task *task)
+{
+    pthread_mutex_unlock(&task->lock);
+}
+
+static FLB_INLINE int flb_task_get_active_route_count(
+                        struct flb_task *task)
+{
+    struct mk_list        *iterator;
+    int                    result;
+    struct flb_task_route *route;
+
+    result = 0;
+
+    mk_list_foreach(iterator, &task->routes) {
+        route = mk_list_entry(iterator, struct flb_task_route, _head);
+
+        if (route->status == FLB_TASK_ROUTE_ACTIVE) {
+            result++;
+        }
+    }
+
+    return result;
+}
+
+static FLB_INLINE size_t flb_task_get_route_status(
+                            struct flb_task *task,
+                            struct flb_output_instance *o_ins)
+{
+    struct mk_list        *iterator;
+    size_t                 result;
+    struct flb_task_route *route;
+
+    result = FLB_TASK_ROUTE_INACTIVE;
+
+    mk_list_foreach(iterator, &task->routes) {
+        route = mk_list_entry(iterator, struct flb_task_route, _head);
+
+        if (route->out == o_ins) {
+            result = route->status;
+            break;
+        }
+    }
+
+    return result;
+}
+
+static FLB_INLINE void flb_task_set_route_status(
+                        struct flb_task *task,
+                        struct flb_output_instance *o_ins,
+                        int new_status)
+{
+    struct mk_list        *iterator;
+    struct flb_task_route *route;
+
+    mk_list_foreach(iterator, &task->routes) {
+        route = mk_list_entry(iterator, struct flb_task_route, _head);
+
+        if (route->out == o_ins) {
+            route->status = new_status;
+            break;
+        }
+    }
+}
+
+
+static FLB_INLINE void flb_task_activate_route(
+                        struct flb_task *task,
+                        struct flb_output_instance *o_ins)
+{
+    flb_task_set_route_status(task, o_ins, FLB_TASK_ROUTE_ACTIVE);
+}
+
+static FLB_INLINE void flb_task_deactivate_route(
+                        struct flb_task *task,
+                        struct flb_output_instance *o_ins)
+{
+    flb_task_set_route_status(task, o_ins, FLB_TASK_ROUTE_INACTIVE);
+}
+
+static FLB_INLINE void flb_task_drop_route(
+                        struct flb_task *task,
+                        struct flb_output_instance *o_ins)
+{
+    flb_task_set_route_status(task, o_ins, FLB_TASK_ROUTE_DROPPED);
+}
 
 #endif
