@@ -105,6 +105,10 @@ static inline int _mk_event_add(struct mk_event_ctx *ctx, int fd,
     int ret;
     struct mk_event *event;
     struct epoll_event ep_event;
+    memset(&ep_event, 0, sizeof(ep_event));
+
+    mk_bug(ctx == NULL);
+    mk_bug(data == NULL);
 
     /* Verify the FD status and desired operation */
     event = (struct mk_event *) data;
@@ -138,6 +142,13 @@ static inline int _mk_event_add(struct mk_event_ctx *ctx, int fd,
     }
 
     event->mask = events;
+    event->priority = MK_EVENT_PRIORITY_DEFAULT;
+
+    /* Remove from priority queue */
+    if (!mk_list_entry_is_orphan(&event->_priority_head)) {
+        mk_list_del(&event->_priority_head);
+    }
+
     return ret;
 }
 
@@ -146,19 +157,29 @@ static inline int _mk_event_del(struct mk_event_ctx *ctx, struct mk_event *event
 {
     int ret;
 
-    if ((event->status & MK_EVENT_REGISTERED) == 0) {
+    mk_bug(ctx == NULL);
+    mk_bug(event == NULL);
+
+    if (!MK_EVENT_IS_REGISTERED(event)) {
         return 0;
     }
 
     ret = epoll_ctl(ctx->efd, EPOLL_CTL_DEL, event->fd, NULL);
+
     MK_TRACE("[FD %i] Epoll, remove from QUEUE_FD=%i, ret=%i",
              event->fd, ctx->efd, ret);
+
     if (ret < 0) {
 #ifdef MK_HAVE_TRACE
         mk_libc_warn("epoll_ctl");
 #endif
     }
 
+    /* Remove from priority queue */
+    if (!mk_list_entry_is_orphan(&event->_priority_head)) {
+        mk_list_del(&event->_priority_head);
+    }
+
     MK_EVENT_NEW(event);
 
     return ret;
@@ -175,7 +196,8 @@ static inline int _mk_event_timeout_create(struct mk_event_ctx *ctx,
     struct timespec now;
     struct mk_event *event;
 
-    mk_bug(!data);
+    mk_bug(data == NULL);
+
     memset(&its, '\0', sizeof(struct itimerspec));
 
     if (clock_gettime(CLOCK_MONOTONIC, &now) != 0) {
@@ -275,6 +297,8 @@ static inline int _mk_event_timeout_create(struct mk_event_ctx *ctx,
     struct fd_timer *timer;
     pthread_t tid;
 
+    mk_bug(data == NULL);
+
     timer = mk_mem_alloc(sizeof(struct fd_timer));
     if (!timer) {
         return -1;
@@ -317,7 +341,7 @@ static inline int _mk_event_timeout_destroy(struct mk_event_ctx *ctx, void *data
 {
     struct mk_event *event;
 
-    if (!data) {
+    if (data == NULL) {
         return 0;
     }
 
@@ -334,6 +358,8 @@ static inline int _mk_event_channel_create(struct mk_event_ctx *ctx,
     int fd[2];
     struct mk_event *event;
 
+    mk_bug(data == NULL);
+
     ret = pipe(fd);
     if (ret < 0) {
         mk_libc_error("pipe");
@@ -359,12 +385,35 @@ static inline int _mk_event_channel_create(struct mk_event_ctx *ctx,
     return 0;
 }
 
+static inline int _mk_event_channel_destroy(struct mk_event_ctx *ctx,
+                                            int r_fd, int w_fd, void *data)
+{
+    struct mk_event *event;
+    int ret;
+
+
+    event = (struct mk_event *)data;
+    if (event->fd != r_fd) {
+        return -1;
+    }
+
+    ret = _mk_event_del(ctx, event);
+    if (ret != 0) {
+        return ret;
+    }
+
+    close(r_fd);
+    close(w_fd);
+
+    return 0;
+}
+
 static inline int _mk_event_inject(struct mk_event_loop *loop,
                                    struct mk_event *event,
                                    int mask,
                                    int prevent_duplication)
 {
-    size_t               index;
+    int                  index;
     struct mk_event_ctx *ctx;
 
     ctx = loop->data;
@@ -386,12 +435,24 @@ static inline int _mk_event_inject(struct mk_event_loop *loop,
     return 0;
 }
 
-static inline int _mk_event_wait(struct mk_event_loop *loop)
+static inline int _mk_event_wait_2(struct mk_event_loop *loop, int timeout)
 {
     struct mk_event_ctx *ctx = loop->data;
+    int ret = 0;
 
-    loop->n_events = epoll_wait(ctx->efd, ctx->events, ctx->queue_size, -1);
-    return loop->n_events;
+    while(1) {
+        ret = epoll_wait(ctx->efd, ctx->events, ctx->queue_size, timeout);
+        if (ret >= 0) {
+            break;
+        }
+        else if(ret < 0 && errno != EINTR) {
+            mk_libc_error("epoll_wait");
+            break;
+        }
+        /* retry when errno is EINTR */
+    }
+    loop->n_events = ret;
+    return ret;
 }
 
 static inline char *_mk_event_backend()
