@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -29,6 +28,7 @@
 #include <fluent-bit/flb_io.h>
 #include <fluent-bit/flb_sds.h>
 #include <fluent-bit/flb_regex.h>
+#include <fluent-bit/flb_hash_table.h>
 
 /*
  * Since this filter might get a high number of request per second,
@@ -55,9 +55,6 @@
 #define FLB_API_PORT  443
 #define FLB_API_TLS   FLB_TRUE
 
-/* Kubelet info */
-#define FLB_KUBELET_HOST  "127.0.0.1"
-
 /*
  * Default expected Kubernetes tag prefix, this is used mostly when source
  * data comes from in_tail with custom tags like: kube.service.*
@@ -83,6 +80,7 @@ struct flb_kube {
     int dummy_meta;
     int tls_debug;
     int tls_verify;
+    int kube_token_ttl;
     flb_sds_t meta_preload_cache_dir;
 
     /* Configuration proposed through Annotations (boolean) */
@@ -156,6 +154,7 @@ struct flb_kube {
 
     int use_tag_for_meta;
     int use_kubelet;
+    char *kubelet_host;
     int kubelet_port;
 
     int kube_meta_cache_ttl;
@@ -163,7 +162,7 @@ struct flb_kube {
     struct flb_tls *tls;
 
     struct flb_config *config;
-    struct flb_hash *hash_table;
+    struct flb_hash_table *hash_table;
     struct flb_upstream *upstream;
     struct flb_filter_instance *ins;
 };
