@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -25,6 +24,8 @@
 #include <fluent-bit/flb_pack.h>
 #include <fluent-bit/flb_parser.h>
 #include <fluent-bit/flb_unescape.h>
+#include <fluent-bit/flb_log_event_decoder.h>
+#include <fluent-bit/flb_log_event_encoder.h>
 
 #include "kube_conf.h"
 #include "kube_meta.h"
@@ -132,7 +133,7 @@ static int merge_log_handler(msgpack_object o,
         ret = flb_parser_do(parser, ctx->unesc_buf, ctx->unesc_buf_len,
                             out_buf, out_size, log_time);
         if (ret >= 0) {
-            if (flb_time_to_double(log_time) == 0.0) {
+            if (flb_time_to_nanosec(log_time) == 0L) {
                 flb_time_get(log_time);
             }
             return MERGE_PARSED;
@@ -143,7 +144,7 @@ static int merge_log_handler(msgpack_object o,
                             ctx->unesc_buf, ctx->unesc_buf_len,
                             out_buf, out_size, log_time);
         if (ret >= 0) {
-            if (flb_time_to_double(log_time) == 0.0) {
+            if (flb_time_to_nanosec(log_time) == 0L) {
                 flb_time_get(log_time);
             }
             return MERGE_PARSED;
@@ -152,7 +153,7 @@ static int merge_log_handler(msgpack_object o,
     else { /* Default JSON parser */
         ret = flb_pack_json_recs(ctx->unesc_buf, ctx->unesc_buf_len,
                                  (char **) out_buf, out_size, &root_type,
-                                 &records);
+                                 &records, NULL);
         if (ret == 0 && root_type != FLB_PACK_JSON_OBJECT) {
             flb_plg_debug(ctx->ins, "could not merge JSON, root_type=%i",
                       root_type);
@@ -210,7 +211,7 @@ static int cb_kube_init(struct flb_filter_instance *f_ins,
     return 0;
 }
 
-static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
+static int pack_map_content(struct flb_log_event_encoder *log_encoder,
                             msgpack_object source_map,
                             const char *kube_buf, size_t kube_size,
                             struct flb_kube_meta *meta,
@@ -218,10 +219,12 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
                             struct flb_parser *parser,
                             struct flb_kube *ctx)
 {
+    int append_original_objects;
+    int scope_opened;
+    int ret;
     int i;
     int map_size = 0;
     int merge_status = -1;
-    int new_map_size = 0;
     int log_index = -1;
     int log_buf_entries = 0;
     size_t off = 0;
@@ -273,19 +276,23 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
 
     /* Append record timestamp */
     if (merge_status == MERGE_PARSED) {
-        if (flb_time_to_double(&log_time) == 0.0) {
-            flb_time_append_to_msgpack(time_lookup, pck, 0);
+        if (flb_time_to_nanosec(&log_time) == 0L) {
+            ret = flb_log_event_encoder_set_timestamp(
+                    log_encoder, time_lookup);
         }
         else {
-            flb_time_append_to_msgpack(&log_time, pck, 0);
+            ret = flb_log_event_encoder_set_timestamp(
+                    log_encoder, &log_time);
         }
     }
     else {
-        flb_time_append_to_msgpack(time_lookup, pck, 0);
+        ret = flb_log_event_encoder_set_timestamp(
+                log_encoder, time_lookup);
     }
 
-    /* Determinate the size of the new map */
-    new_map_size = map_size;
+    if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+        return -1;
+    }
 
     /* If a merged status exists, check the number of entries to merge */
     if (log_index != -1) {
@@ -305,30 +312,15 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
         }
     }
 
-    /* Kubernetes metadata */
-    if (kube_buf && kube_size > 0) {
-        new_map_size++;
-    }
-
-    if (log_buf_entries > 0) {
-        if (ctx->merge_log_key != NULL) {
-            new_map_size++;
-        }
-        else {
-            new_map_size += log_buf_entries;
-        }
-    }
-
     if ((merge_status == MERGE_PARSED || merge_status == MERGE_MAP) &&
         ctx->keep_log == FLB_FALSE) {
-        new_map_size--;
     }
 
-    /* Pack Map */
-    msgpack_pack_map(pck, new_map_size);
-
     /* Original map */
-    for (i = 0; i < map_size; i++) {
+    for (i = 0;
+         i < map_size &&
+         ret == FLB_EVENT_ENCODER_SUCCESS;
+         i++) {
         k = source_map.via.map.ptr[i].key;
         v = source_map.via.map.ptr[i].val;
 
@@ -337,37 +329,60 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
          * will depend on merge_status. If the parsing failed we cannot
          * merge so we keep the 'log' key/value.
          */
+        append_original_objects = FLB_FALSE;
+
         if (log_index == i) {
             if (ctx->keep_log == FLB_TRUE) {
-                msgpack_pack_object(pck, k);
                 if (merge_status == MERGE_NONE || merge_status == MERGE_PARSED){
-                    msgpack_pack_str(pck, ctx->unesc_buf_len);
-                    msgpack_pack_str_body(pck, ctx->unesc_buf,
-                                          ctx->unesc_buf_len);
+                    ret = flb_log_event_encoder_append_body_values(
+                            log_encoder,
+                            FLB_LOG_EVENT_MSGPACK_OBJECT_VALUE(&k),
+                            FLB_LOG_EVENT_STRING_VALUE(ctx->unesc_buf,
+                                                       ctx->unesc_buf_len));
                 }
                 else {
-                    msgpack_pack_object(pck, v);
+                    append_original_objects = FLB_TRUE;
                 }
             }
             else if (merge_status == MERGE_NONE) {
-                msgpack_pack_object(pck, k);
-                msgpack_pack_object(pck, v);
+                append_original_objects = FLB_TRUE;
             }
         }
         else {
-            msgpack_pack_object(pck, k);
-            msgpack_pack_object(pck, v);
+            append_original_objects = FLB_TRUE;
+        }
+
+        if (append_original_objects) {
+            ret = flb_log_event_encoder_append_body_values(
+                    log_encoder,
+                    FLB_LOG_EVENT_MSGPACK_OBJECT_VALUE(&k),
+                    FLB_LOG_EVENT_MSGPACK_OBJECT_VALUE(&v));
         }
     }
 
+    if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+        return -2;
+    }
+
+    scope_opened = FLB_FALSE;
     /* Merge Log */
     if (log_index != -1) {
         if (merge_status == MERGE_PARSED) {
             if (ctx->merge_log_key && log_buf_entries > 0) {
-                msgpack_pack_str(pck, flb_sds_len(ctx->merge_log_key));
-                msgpack_pack_str_body(pck, ctx->merge_log_key,
-                                      flb_sds_len(ctx->merge_log_key));
-                msgpack_pack_map(pck, log_buf_entries);
+                ret = flb_log_event_encoder_append_body_string(
+                        log_encoder,
+                        ctx->merge_log_key,
+                        flb_sds_len(ctx->merge_log_key));
+
+                if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+                    ret = flb_log_event_encoder_body_begin_map(log_encoder);
+                }
+
+                if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+                    return -3;
+                }
+
+                scope_opened = FLB_TRUE;
             }
 
             off = 0;
@@ -375,9 +390,18 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
             msgpack_unpack_next(&result, log_buf, log_size, &off);
             root = result.data;
 
-            for (i = 0; i < log_buf_entries; i++) {
+            for (i = 0;
+                 i < log_buf_entries &&
+                 ret == FLB_EVENT_ENCODER_SUCCESS;
+                 i++) {
                 k = root.via.map.ptr[i].key;
-                msgpack_pack_object(pck, k);
+
+                ret = flb_log_event_encoder_append_body_msgpack_object(
+                        log_encoder, &k);
+
+                if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+                    return -4;
+                }
 
                 v = root.via.map.ptr[i].val;
 
@@ -387,49 +411,107 @@ static int pack_map_content(msgpack_packer *pck, msgpack_sbuffer *sbuf,
                  */
                 if (v.type == MSGPACK_OBJECT_STR &&
                     ctx->merge_log_trim == FLB_TRUE) {
-                    int s = value_trim_size(v);
-                    msgpack_pack_str(pck, s);
-                    msgpack_pack_str_body(pck, v.via.str.ptr, s);
+                    ret = flb_log_event_encoder_append_body_string(
+                            log_encoder,
+                            (char *) v.via.str.ptr,
+                            value_trim_size(v));
                 }
                 else {
-                    msgpack_pack_object(pck, v);
+                    ret = flb_log_event_encoder_append_body_msgpack_object(
+                            log_encoder, &v);
                 }
             }
+
             msgpack_unpacked_destroy(&result);
+
             flb_free(log_buf);
+
+            if (scope_opened && ret == FLB_EVENT_ENCODER_SUCCESS) {
+                ret = flb_log_event_encoder_body_commit_map(log_encoder);
+            }
+
+            if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+                return -5;
+            }
         }
         else if (merge_status == MERGE_MAP) {
             msgpack_object map;
 
             if (ctx->merge_log_key && log_buf_entries > 0) {
-                msgpack_pack_str(pck, flb_sds_len(ctx->merge_log_key));
-                msgpack_pack_str_body(pck, ctx->merge_log_key,
-                                      flb_sds_len(ctx->merge_log_key));
-                msgpack_pack_map(pck, log_buf_entries);
+                ret = flb_log_event_encoder_append_body_string(
+                        log_encoder,
+                        ctx->merge_log_key,
+                        flb_sds_len(ctx->merge_log_key));
+
+                if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+                    ret = flb_log_event_encoder_body_begin_map(log_encoder);
+                }
+
+                if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+                    return -6;
+                }
+
+                scope_opened = FLB_TRUE;
             }
 
             map = source_map.via.map.ptr[log_index].val;
-            for (i = 0; i < map.via.map.size; i++) {
+            for (i = 0;
+                 i < map.via.map.size &&
+                 ret == FLB_EVENT_ENCODER_SUCCESS;
+                 i++) {
                 k = map.via.map.ptr[i].key;
                 v = map.via.map.ptr[i].val;
-                msgpack_pack_object(pck, k);
-                msgpack_pack_object(pck, v);
+
+                ret = flb_log_event_encoder_append_body_values(
+                        log_encoder,
+                        FLB_LOG_EVENT_MSGPACK_OBJECT_VALUE(&k),
+                        FLB_LOG_EVENT_MSGPACK_OBJECT_VALUE(&v));
+            }
+
+            if (scope_opened && ret == FLB_EVENT_ENCODER_SUCCESS) {
+                ret = flb_log_event_encoder_body_commit_map(log_encoder);
+            }
+
+            if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+                return -7;
             }
         }
     }
 
     /* Kubernetes */
     if (kube_buf && kube_size > 0) {
-        msgpack_pack_str(pck, 10);
-        msgpack_pack_str_body(pck, "kubernetes", 10);
+        ret = flb_log_event_encoder_append_body_cstring(
+                log_encoder,
+                "kubernetes");
 
         off = 0;
         msgpack_unpacked_init(&result);
         msgpack_unpack_next(&result, kube_buf, kube_size, &off);
-        msgpack_pack_object(pck, result.data);
+
+        if (kube_size != off) {
+            /* This buffer should contain a single map and we shouldn't
+             * have to unpack it in order to ensure that we are appending
+             * a single map but considering that the current code only
+             * appends the first entry without taking any actions I think
+             * we should warn the user if there is more than one entry in
+             * it so in the future we can remove the unpack code and just
+             * use flb_log_event_encoder_append_body_raw_msgpack with
+             * kube_size.
+             */
+        }
+
+        if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+            ret = flb_log_event_encoder_append_body_raw_msgpack(log_encoder,
+                    (char *) kube_buf, off);
+        }
+
         msgpack_unpacked_destroy(&result);
     }
 
+    if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+        return -8;
+    }
+
     return 0;
 }
 
@@ -447,17 +529,15 @@ static int cb_kube_filter(const void *data, size_t bytes,
     char *dummy_cache_buf = NULL;
     const char *cache_buf = NULL;
     size_t cache_size = 0;
-    msgpack_unpacked result;
     msgpack_object map;
-    msgpack_object root;
-    msgpack_sbuffer tmp_sbuf;
-    msgpack_packer tmp_pck;
-    msgpack_object *obj;
     struct flb_parser *parser = NULL;
     struct flb_kube *ctx = filter_context;
     struct flb_kube_meta meta = {0};
     struct flb_kube_props props = {0};
-    struct flb_time time_lookup;
+    struct flb_log_event_encoder log_encoder;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
+
     (void) f_ins;
     (void) i_ins;
     (void) config;
@@ -479,22 +559,36 @@ static int cb_kube_filter(const void *data, size_t bytes,
         }
     }
 
-    /* Create temporary msgpack buffer */
-    msgpack_sbuffer_init(&tmp_sbuf);
-    msgpack_packer_init(&tmp_pck, &tmp_sbuf, msgpack_sbuffer_write);
+    ret = flb_log_event_decoder_init(&log_decoder, (char *) data, bytes);
 
-    /* Iterate each item array and append meta */
-    msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
-        root = result.data;
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
 
-        if (root.type != MSGPACK_OBJECT_ARRAY ||
-            root.via.array.size != 2 ||
-            root.via.array.ptr[1].type != MSGPACK_OBJECT_MAP) {
-            flb_plg_warn(ctx->ins, "unexpected record format");
-            continue;
-        }
+        flb_kube_meta_release(&meta);
+        flb_kube_prop_destroy(&props);
 
+        return FLB_FILTER_NOTOUCH;
+    }
+
+    ret = flb_log_event_encoder_init(&log_encoder,
+                                     FLB_LOG_EVENT_FORMAT_DEFAULT);
+
+    if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event encoder initialization error : %d", ret);
+
+        flb_log_event_decoder_destroy(&log_decoder);
+        flb_kube_meta_release(&meta);
+        flb_kube_prop_destroy(&props);
+
+        return FLB_FILTER_NOTOUCH;
+    }
+
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
+        off = log_decoder.offset;
         /*
          * Journal entries can be origined by different Pods, so we are forced
          * to parse and check it metadata.
@@ -516,13 +610,14 @@ static int cb_kube_filter(const void *data, size_t bytes,
 
         parser = NULL;
 
-        switch (get_stream(root.via.array.ptr[1].via.map)) {
+        switch (get_stream(log_event.body->via.map)) {
         case FLB_KUBE_PROP_STREAM_STDOUT:
             {
                 if (props.stdout_exclude == FLB_TRUE) {
                     /* Skip this record */
                     if (ctx->use_journal == FLB_TRUE) {
                         flb_kube_meta_release(&meta);
+                        flb_kube_prop_destroy(&props);
                     }
                     continue;
                 }
@@ -534,6 +629,11 @@ static int cb_kube_filter(const void *data, size_t bytes,
         case FLB_KUBE_PROP_STREAM_STDERR:
             {
                 if (props.stderr_exclude == FLB_TRUE) {
+                    /* Skip this record */
+                    if (ctx->use_journal == FLB_TRUE) {
+                        flb_kube_meta_release(&meta);
+                        flb_kube_prop_destroy(&props);
+                    }
                     continue;
                 }
                 if (props.stderr_parser != NULL) {
@@ -555,41 +655,46 @@ static int cb_kube_filter(const void *data, size_t bytes,
             break;
         }
 
-        /*
-         * Temporal time lookup in case a parser comes up with a new
-         * timestamp for the record.
-         */
-        flb_time_pop_from_msgpack(&time_lookup, &result, &obj);
-
         /* get records map */
-        map  = root.via.array.ptr[1];
+        map  = *log_event.body;
 
-        /* Compose the new array (0=timestamp, 1=record) */
-        msgpack_pack_array(&tmp_pck, 2);
+        ret = flb_log_event_encoder_begin_record(&log_encoder);
 
+        if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+            break;
+        }
 
-        ret = pack_map_content(&tmp_pck, &tmp_sbuf,
+        ret = pack_map_content(&log_encoder,
                                map,
                                cache_buf, cache_size,
-                               &meta, &time_lookup, parser, ctx);
-        if (ret == -1) {
-            msgpack_sbuffer_destroy(&tmp_sbuf);
-            msgpack_unpacked_destroy(&result);
+                               &meta, &log_event.timestamp, parser, ctx);
+        if (ret != 0) {
+            flb_log_event_decoder_destroy(&log_decoder);
+            flb_log_event_encoder_destroy(&log_encoder);
+
             if (ctx->dummy_meta == FLB_TRUE) {
                 flb_free(dummy_cache_buf);
             }
 
             flb_kube_meta_release(&meta);
             flb_kube_prop_destroy(&props);
+
             return FLB_FILTER_NOTOUCH;
         }
 
+        ret = flb_log_event_encoder_commit_record(&log_encoder);
+
+        if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+            flb_log_event_encoder_rollback_record(&log_encoder);
+
+            break;
+        }
+
         if (ctx->use_journal == FLB_TRUE) {
             flb_kube_meta_release(&meta);
             flb_kube_prop_destroy(&props);
         }
     }
-    msgpack_unpacked_destroy(&result);
 
     /* Release meta fields */
     if (ctx->use_journal == FLB_FALSE) {
@@ -597,14 +702,18 @@ static int cb_kube_filter(const void *data, size_t bytes,
         flb_kube_prop_destroy(&props);
     }
 
-    /* link new buffers */
-    *out_buf   = tmp_sbuf.data;
-    *out_bytes = tmp_sbuf.size;
-
     if (ctx->dummy_meta == FLB_TRUE) {
         flb_free(dummy_cache_buf);
     }
 
+    *out_buf   = log_encoder.output_buffer;
+    *out_bytes = log_encoder.output_length;
+
+    flb_log_event_encoder_claim_internal_buffer_ownership(&log_encoder);
+
+    flb_log_event_decoder_destroy(&log_decoder);
+    flb_log_event_encoder_destroy(&log_encoder);
+
     return FLB_FILTER_MODIFIED;
 }
 
@@ -840,6 +949,15 @@ static struct flb_config_map config_map[] = {
      0, FLB_TRUE, offsetof(struct flb_kube, use_kubelet),
      "use kubelet to get metadata instead of kube-server"
     },
+    /*
+     * The kubelet host for /pods endpoint, default is 127.0.0.1
+     * Will only check when "use_kubelet" config is set to true
+     */
+    {
+     FLB_CONFIG_MAP_STR, "kubelet_host", "127.0.0.1",
+     0, FLB_TRUE, offsetof(struct flb_kube, kubelet_host),
+     "kubelet host to connect with when using kubelet"
+    },
     /*
      * The kubelet port for /pods endpoint, default is 10250
      * Will only check when "use_kubelet" config is set to true
@@ -849,6 +967,11 @@ static struct flb_config_map config_map[] = {
      0, FLB_TRUE, offsetof(struct flb_kube, kubelet_port),
      "kubelet port to connect with when using kubelet"
     },
+    {
+     FLB_CONFIG_MAP_TIME, "kube_token_ttl", "10m",
+     0, FLB_TRUE, offsetof(struct flb_kube, kube_token_ttl),
+     "kubernetes token ttl, until it is reread from the token file. Default: 10m"
+    },
     /*
      * Set TTL for K8s cached metadata 
      */
