@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -20,6 +19,8 @@
 
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_input.h>
+#include <fluent-bit/flb_plugin.h>
+#include <fluent-bit/flb_processor.h>
 #include <fluent-bit/flb_filter_plugin.h>
 #include <fluent-bit/flb_metrics.h>
 #include <fluent-bit/flb_storage.h>
@@ -27,6 +28,8 @@
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_time.h>
 #include <fluent-bit/flb_record_accessor.h>
+#include <fluent-bit/flb_log_event_decoder.h>
+#include <fluent-bit/flb_log_event_encoder.h>
 #include <msgpack.h>
 
 #include "rewrite_tag.h"
@@ -94,6 +97,8 @@ static int emitter_create(struct flb_rewrite_tag *ctx)
     if (ret == -1) {
         flb_plg_error(ctx->ins, "cannot initialize storage for stream '%s'",
                       ctx->emitter_name);
+        flb_input_instance_exit(ins, ctx->config);
+        flb_input_instance_destroy(ins);
         return -1;
     }
     ctx->ins_emitter = ins;
@@ -132,6 +137,11 @@ static int process_config(struct flb_rewrite_tag *ctx)
 
         /* key */
         entry = flb_slist_entry_get(val->val.list, 0);
+        if (entry == NULL) {
+            flb_plg_error(ctx->ins, "failed to get entry");
+            flb_free(rule);
+            return -1;
+        }
         rule->ra_key = flb_ra_create(entry->str, FLB_FALSE);
         if (!rule->ra_key) {
             flb_plg_error(ctx->ins, "invalid record accessor key ? '%s'",
@@ -215,9 +225,7 @@ static int cb_rewrite_tag_init(struct flb_filter_instance *ins,
         return -1;
     }
     if (is_wildcard(ins->match)) {
-        flb_plg_error(ins, "'Match' causes infinite loop. abort.");
-        flb_free(ctx);
-        return -1;
+        flb_plg_warn(ins, "'Match' may cause infinite loop.");
     }
     ctx->ins = ins;
     ctx->config = config;
@@ -308,6 +316,38 @@ static int cb_rewrite_tag_init(struct flb_filter_instance *ins,
     return 0;
 }
 
+static int ingest_inline(struct flb_rewrite_tag *ctx,
+                         flb_sds_t out_tag,
+                         const void *buf, size_t buf_size)
+{
+    struct flb_input_instance *input_instance;
+    struct flb_processor_unit *processor_unit;
+    struct flb_processor      *processor;
+    int                        result;
+
+    if (ctx->ins->parent_processor != NULL) {
+        processor_unit = (struct flb_processor_unit *) \
+                            ctx->ins->parent_processor;
+        processor = (struct flb_processor *) processor_unit->parent;
+        input_instance = (struct flb_input_instance *) processor->data;
+
+        if (processor->source_plugin_type == FLB_PLUGIN_INPUT) {
+            result = flb_input_log_append_skip_processor_stages(
+                        input_instance,
+                        processor_unit->stage + 1,
+                        out_tag, flb_sds_len(out_tag),
+                        buf, buf_size);
+
+            if (result == 0) {
+                return FLB_TRUE;
+            }
+        }
+    }
+
+    return FLB_FALSE;
+}
+
+
 /*
  * On given record, check if a rule applies or not to the map, if so, compose
  * the new tag, emit the record and return FLB_TRUE, otherwise just return
@@ -315,7 +355,7 @@ static int cb_rewrite_tag_init(struct flb_filter_instance *ins,
  */
 static int process_record(const char *tag, int tag_len, msgpack_object map,
                           const void *buf, size_t buf_size, int *keep,
-                          struct flb_rewrite_tag *ctx)
+                          struct flb_rewrite_tag *ctx, int *matched)
 {
     int ret;
     flb_sds_t out_tag;
@@ -323,6 +363,11 @@ static int process_record(const char *tag, int tag_len, msgpack_object map,
     struct rewrite_rule *rule = NULL;
     struct flb_regex_search result = {0};
 
+    if (matched == NULL) {
+        return FLB_FALSE;
+    }
+    *matched = FLB_FALSE;
+
     mk_list_foreach(head, &ctx->rules) {
         rule = mk_list_entry(head, struct rewrite_rule, _head);
         if (rule) {
@@ -341,6 +386,7 @@ static int process_record(const char *tag, int tag_len, msgpack_object map,
     if (!rule) {
         return FLB_FALSE;
     }
+    *matched = FLB_TRUE;
 
     /* Compose new tag */
     out_tag = flb_ra_translate(rule->ra_tag, (char *) tag, tag_len, map, &result);
@@ -353,9 +399,16 @@ static int process_record(const char *tag, int tag_len, msgpack_object map,
         return FLB_FALSE;
     }
 
-    /* Emit record with new tag */
-    ret = in_emitter_add_record(out_tag, flb_sds_len(out_tag), buf, buf_size,
-                                ctx->ins_emitter);
+    ret = ingest_inline(ctx, out_tag, buf, buf_size);
+
+    if (!ret) {
+        /* Emit record with new tag */
+        ret = in_emitter_add_record(out_tag, flb_sds_len(out_tag), buf, buf_size,
+                                    ctx->ins_emitter);
+    }
+    else {
+        ret = 0;
+    }
 
     /* Release the tag */
     flb_sds_destroy(out_tag);
@@ -364,7 +417,6 @@ static int process_record(const char *tag, int tag_len, msgpack_object map,
         return FLB_FALSE;
     }
 
-
     return FLB_TRUE;
 }
 
@@ -376,38 +428,60 @@ static int cb_rewrite_tag_filter(const void *data, size_t bytes,
                                  void *filter_context,
                                  struct flb_config *config)
 {
-    int ret;
     int keep;
-    int emitted = 0;
+    int emitted_num = 0;
+    int is_matched = FLB_FALSE;
+    int is_emitted = FLB_FALSE;
     size_t pre = 0;
     size_t off = 0;
 #ifdef FLB_HAVE_METRICS
     uint64_t ts;
     char *name;
 #endif
-    msgpack_sbuffer mp_sbuf;
-    msgpack_packer mp_pck;
     msgpack_object map;
-    msgpack_object root;
-    msgpack_unpacked result;
-    struct flb_rewrite_tag *ctx = (struct flb_rewrite_tag *) filter_context;
+    struct flb_rewrite_tag *ctx;
+    struct flb_log_event_encoder log_encoder;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
+    int ret;
+
     (void) config;
     (void) i_ins;
 
+    ctx = (struct flb_rewrite_tag *) filter_context;
+
 #ifdef FLB_HAVE_METRICS
-    ts = cmt_time_now();
+    ts = cfl_time_now();
     name = (char *) flb_filter_name(f_ins);
 #endif
 
-    /* Create temporal msgpack buffer */
-    msgpack_sbuffer_init(&mp_sbuf);
-    msgpack_packer_init(&mp_pck, &mp_sbuf, msgpack_sbuffer_write);
+    ret = flb_log_event_decoder_init(&log_decoder, (char *) data, bytes);
 
-    msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
-        root = result.data;
-        map = root.via.array.ptr[1];
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
 
+        return FLB_FILTER_NOTOUCH;
+    }
+
+    ret = flb_log_event_encoder_init(&log_encoder,
+                                     FLB_LOG_EVENT_FORMAT_DEFAULT);
+
+    if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event encoder initialization error : %d", ret);
+
+        flb_log_event_decoder_destroy(&log_decoder);
+
+        return FLB_FILTER_NOTOUCH;
+    }
+
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
+        off = log_decoder.offset;
+        map = *log_event.body;
+        is_matched = FLB_FALSE;
         /*
          * Process the record according the defined rules. If it returns FLB_TRUE means
          * the record was emitter with a different tag.
@@ -415,10 +489,10 @@ static int cb_rewrite_tag_filter(const void *data, size_t bytes,
          * If a record was emitted, the variable 'keep' will define if the record must
          * be preserved or not.
          */
-        ret = process_record(tag, tag_len, map, (char *) data + pre, off - pre, &keep, ctx);
-        if (ret == FLB_TRUE) {
+        is_emitted = process_record(tag, tag_len, map, (char *) data + pre, off - pre, &keep, ctx, &is_matched);
+        if (is_emitted == FLB_TRUE) {
             /* A record with the new tag was emitted */
-            emitted++;
+            emitted_num++;
         }
 
         /*
@@ -427,33 +501,57 @@ static int cb_rewrite_tag_filter(const void *data, size_t bytes,
          * - record with new tag was emitted and the rule says it must be preserved
          * - record was not emitted
          */
-        if (keep == FLB_TRUE) {
-            msgpack_sbuffer_write(&mp_sbuf, (char *) data + pre, off - pre);
+        if (keep == FLB_TRUE || is_matched != FLB_TRUE) {
+            ret = flb_log_event_encoder_emit_raw_record(
+                    &log_encoder,
+                    log_decoder.record_base,
+                    log_decoder.record_length);
         }
 
         /* Adjust previous offset */
         pre = off;
     }
-    msgpack_unpacked_destroy(&result);
 
-    if (emitted == 0) {
-        msgpack_sbuffer_destroy(&mp_sbuf);
+    if (emitted_num == 0) {
+        flb_log_event_decoder_destroy(&log_decoder);
+        flb_log_event_encoder_destroy(&log_encoder);
+
         return FLB_FILTER_NOTOUCH;
     }
 #ifdef FLB_HAVE_METRICS
-    else if (emitted > 0) {
-        cmt_counter_add(ctx->cmt_emitted, ts, emitted,
+    else if (emitted_num > 0) {
+        cmt_counter_add(ctx->cmt_emitted, ts, emitted_num,
                         1, (char *[]) {name});
 
         /* OLD api */
-        flb_metrics_sum(FLB_RTAG_METRIC_EMITTED, emitted, ctx->ins->metrics);
+        flb_metrics_sum(FLB_RTAG_METRIC_EMITTED, emitted_num, ctx->ins->metrics);
     }
 #endif
 
-    *out_buf = mp_sbuf.data;
-    *out_bytes = mp_sbuf.size;
+    if (ret == FLB_EVENT_DECODER_ERROR_INSUFFICIENT_DATA &&
+        log_decoder.offset == bytes) {
+        ret = FLB_EVENT_ENCODER_SUCCESS;
+    }
 
-    return FLB_FILTER_MODIFIED;
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        *out_buf   = log_encoder.output_buffer;
+        *out_bytes = log_encoder.output_length;
+
+        ret = FLB_FILTER_MODIFIED;
+
+        flb_log_event_encoder_claim_internal_buffer_ownership(&log_encoder);
+    }
+    else {
+        flb_plg_error(ctx->ins,
+                      "Log event encoder error : %d", ret);
+
+        ret = FLB_FILTER_NOTOUCH;
+    }
+
+    flb_log_event_decoder_destroy(&log_decoder);
+    flb_log_event_encoder_destroy(&log_encoder);
+
+    return ret;
 }
 
 /* Destroy rules from context */
