@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -20,6 +19,8 @@
 
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_input.h>
+#include <fluent-bit/flb_engine.h>
+#include <fluent-bit/flb_downstream.h>
 #include <fluent-bit/flb_input_plugin.h>
 #include <fluent-bit/flb_network.h>
 #include <msgpack.h>
@@ -27,6 +28,7 @@
 #ifdef FLB_HAVE_UNIX_SOCKET
 #include <sys/socket.h>
 #include <sys/un.h>
+#include <sys/stat.h>
 #endif
 
 #include "fw.h"
@@ -34,38 +36,80 @@
 #include "fw_config.h"
 
 #ifdef FLB_HAVE_UNIX_SOCKET
+static int remove_existing_socket_file(char *socket_path)
+{
+    struct stat file_data;
+    int         result;
+
+    result = stat(socket_path, &file_data);
+
+    if (result == -1) {
+        if (errno == ENOENT) {
+            return 0;
+        }
+
+        flb_errno();
+
+        return -1;
+    }
+
+    if (S_ISSOCK(file_data.st_mode) == 0) {
+        return -2;
+    }
+
+    result = unlink(socket_path);
+
+    if (result != 0) {
+        return -3;
+    }
+
+    return 0;
+}
+
 static int fw_unix_create(struct flb_in_fw_config *ctx)
 {
-    flb_sockfd_t fd = -1;
-    unsigned long len;
-    size_t address_length;
-    struct sockaddr_un address;
+    int ret;
+
+    ret = remove_existing_socket_file(ctx->unix_path);
+
+    if (ret != 0) {
+        if (ret == -2) {
+            flb_plg_error(ctx->ins,
+                          "%s exists and it is not a unix socket. Aborting",
+                          ctx->unix_path);
+        }
+        else {
+            flb_plg_error(ctx->ins,
+                          "could not remove existing unix socket %s. Aborting",
+                          ctx->unix_path);
+        }
 
-    fd = flb_net_socket_create(AF_UNIX, FLB_TRUE);
-    if (fd == -1) {
         return -1;
     }
 
-    ctx->server_fd = fd;
+    ctx->downstream = flb_downstream_create(FLB_TRANSPORT_UNIX_STREAM,
+                                            ctx->ins->flags,
+                                            ctx->unix_path,
+                                            0,
+                                            ctx->ins->tls,
+                                            ctx->ins->config,
+                                            &ctx->ins->net_setup);
 
-    /* Prepare the unix socket path */
-    unlink(ctx->unix_path);
-    len = strlen(ctx->unix_path);
-
-    address.sun_family = AF_UNIX;
-    sprintf(address.sun_path, "%s", ctx->unix_path);
-    address_length = sizeof(address.sun_family) + len + 1;
-    if (bind(fd, (struct sockaddr *) &address, address_length) != 0) {
-        flb_errno();
-        close(fd);
+    if (ctx->downstream == NULL) {
         return -1;
     }
 
-    if (listen(fd, 5) != 0) {
-        flb_errno();
-        close(fd);
-        return -1;
+    if (ctx->unix_perm_str) {
+        if (chmod(ctx->unix_path, ctx->unix_perm)) {
+            flb_errno();
+
+            flb_plg_error(ctx->ins, "cannot set permission on '%s' to %04o",
+                          ctx->unix_path, ctx->unix_perm);
+
+            return -1;
+        }
     }
+
     return 0;
 }
 #endif
@@ -78,27 +122,34 @@ static int fw_unix_create(struct flb_in_fw_config *ctx)
 static int in_fw_collect(struct flb_input_instance *ins,
                          struct flb_config *config, void *in_context)
 {
-    int fd;
-    struct fw_conn *conn;
-    struct flb_in_fw_config *ctx = in_context;
+    struct flb_connection   *connection;
+    struct fw_conn          *conn;
+    struct flb_in_fw_config *ctx;
+
+    ctx = in_context;
+
+    connection = flb_downstream_conn_get(ctx->downstream);
+
+    if (connection == NULL) {
+        flb_plg_error(ctx->ins, "could not accept new connection");
 
-    /* Accept the new connection */
-    fd = flb_net_accept(ctx->server_fd);
-    if (fd == -1) {
-        flb_plg_error(ins, "could not accept new connection");
         return -1;
     }
 
-    if (config->is_ingestion_active == FLB_FALSE) {
-        mk_event_closesocket(fd);
+    if (!config->is_ingestion_active) {
+        flb_downstream_conn_release(connection);
+
         return -1;
     }
 
-    flb_plg_trace(ins, "new TCP connection arrived FD=%i", fd);
-    conn = fw_conn_add(fd, ctx);
+    flb_plg_trace(ins, "new TCP connection arrived FD=%i", connection->fd);
+
+    conn = fw_conn_add(connection, ctx);
+
     if (!conn) {
         return -1;
     }
+
     return 0;
 }
 
@@ -106,8 +157,10 @@ static int in_fw_collect(struct flb_input_instance *ins,
 static int in_fw_init(struct flb_input_instance *ins,
                       struct flb_config *config, void *data)
 {
-    int ret;
+    unsigned short int       port;
+    int                      ret;
     struct flb_in_fw_config *ctx;
+
     (void) data;
 
     /* Allocate space for the configuration */
@@ -115,6 +168,8 @@ static int in_fw_init(struct flb_input_instance *ins,
     if (!ctx) {
         return -1;
     }
+
+    ctx->coll_fd = -1;
     ctx->ins = ins;
     mk_list_init(&ctx->connections);
 
@@ -140,33 +195,55 @@ static int in_fw_init(struct flb_input_instance *ins,
 #endif
     }
     else {
-        /* Create TCP server */
-        ctx->server_fd = flb_net_server(ctx->tcp_port, ctx->listen);
-        if (ctx->server_fd > 0) {
+        port = (unsigned short int) strtoul(ctx->tcp_port, NULL, 10);
+
+        ctx->downstream = flb_downstream_create(FLB_TRANSPORT_TCP,
+                                                ctx->ins->flags,
+                                                ctx->listen,
+                                                port,
+                                                ctx->ins->tls,
+                                                config,
+                                                &ctx->ins->net_setup);
+
+        if (ctx->downstream == NULL) {
+            flb_plg_error(ctx->ins,
+                          "could not initialize downstream on unix://%s. Aborting",
+                          ctx->listen);
+
+            fw_config_destroy(ctx);
+
+            return -1;
+        }
+
+        if (ctx->downstream != NULL) {
             flb_plg_info(ctx->ins, "listening on %s:%s",
                          ctx->listen, ctx->tcp_port);
         }
         else {
             flb_plg_error(ctx->ins, "could not bind address %s:%s. Aborting",
                           ctx->listen, ctx->tcp_port);
+
             fw_config_destroy(ctx);
+
             return -1;
         }
     }
-    flb_net_socket_nonblocking(ctx->server_fd);
 
-    ctx->evl = config->evl;
+    flb_input_downstream_set(ctx->downstream, ctx->ins);
+
+    flb_net_socket_nonblocking(ctx->downstream->server_fd);
 
     /* Collect upon data available on the standard input */
     ret = flb_input_set_collector_socket(ins,
                                          in_fw_collect,
-                                         ctx->server_fd,
+                                         ctx->downstream->server_fd,
                                          config);
     if (ret == -1) {
         flb_plg_error(ctx->ins, "could not set server socket collector");
         fw_config_destroy(ctx);
         return -1;
     }
+
     ctx->coll_fd = ret;
 
     return 0;
@@ -185,9 +262,7 @@ static void in_fw_pause(void *data, struct flb_config *config)
      * refactored shortly.
      */
     if (config->is_ingestion_active == FLB_FALSE) {
-        mk_event_closesocket(ctx->server_fd);
         fw_conn_del_all(ctx);
-
     }
 }
 
@@ -217,6 +292,11 @@ static struct flb_config_map config_map[] = {
     0, FLB_TRUE, offsetof(struct flb_in_fw_config, unix_path),
     "The path to unix socket to receive a Forward message."
    },
+   {
+    FLB_CONFIG_MAP_STR, "unix_perm", (char *)NULL,
+    0, FLB_TRUE, offsetof(struct flb_in_fw_config, unix_perm_str),
+    "Set the permissions for the UNIX socket"
+   },
    {
     FLB_CONFIG_MAP_SIZE, "buffer_chunk_size", FLB_IN_FW_CHUNK_SIZE,
     0, FLB_TRUE, offsetof(struct flb_in_fw_config, buffer_chunk_size),
@@ -241,5 +321,5 @@ struct flb_input_plugin in_forward_plugin = {
     .cb_pause     = in_fw_pause,
     .cb_exit      = in_fw_exit,
     .config_map   = config_map,
-    .flags        = FLB_INPUT_NET
+    .flags        = FLB_INPUT_NET_SERVER | FLB_IO_OPT_TLS
 };
