@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -24,6 +23,6 @@
 #include "fw_conn.h"
 
 int fw_prot_parser(struct fw_conn *conn);
-int fw_prot_process(struct fw_conn *conn);
+int fw_prot_process(struct flb_input_instance *ins, struct fw_conn *conn);
 
 #endif
