@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -110,9 +109,6 @@ static int single_value_per_record(struct flb_input_instance *i_ins,
                                    struct flb_in_head_config *ctx)
 {
     int ret = -1;
-    int num_map = 1;
-    msgpack_packer mp_pck;
-    msgpack_sbuffer mp_sbuf;
 
     ctx->buf[0] = '\0'; /* clear buf */
     ctx->buf_len =   0;
@@ -127,41 +123,50 @@ static int single_value_per_record(struct flb_input_instance *i_ins,
     flb_plg_trace(ctx->ins, "%s read_len=%zd buf_size=%zu", __FUNCTION__,
                   ctx->buf_len, ctx->buf_size);
 
-    if (ctx->add_path == FLB_TRUE) {
-        num_map++;
+    ret = flb_log_event_encoder_begin_record(&ctx->log_encoder);
+
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        ret = flb_log_event_encoder_set_current_timestamp(
+                &ctx->log_encoder);
     }
 
-    /* Initialize local msgpack buffer */
-    msgpack_sbuffer_init(&mp_sbuf);
-    msgpack_packer_init(&mp_pck, &mp_sbuf, msgpack_sbuffer_write);
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        ret = flb_log_event_encoder_append_body_values(
+                &ctx->log_encoder,
+                FLB_LOG_EVENT_CSTRING_VALUE(ctx->key),
+                FLB_LOG_EVENT_STRING_VALUE(ctx->buf, ctx->buf_len));
 
-    /* Pack data */
-    msgpack_pack_array(&mp_pck, 2);
-    flb_pack_time_now(&mp_pck);
-    msgpack_pack_map(&mp_pck, num_map);
-
-    msgpack_pack_str(&mp_pck, ctx->key_len);
-    msgpack_pack_str_body(&mp_pck, ctx->key,
-                          ctx->key_len);
-    msgpack_pack_str(&mp_pck, ctx->buf_len);
-    msgpack_pack_str_body(&mp_pck,
-                          ctx->buf, ctx->buf_len);
-
-    if (ctx->add_path == FLB_TRUE) {
-        msgpack_pack_str(&mp_pck, 4);
-        msgpack_pack_str_body(&mp_pck, "path", 4);
-        msgpack_pack_str(&mp_pck, ctx->path_len);
-        msgpack_pack_str_body(&mp_pck,
-                              ctx->filepath, ctx->path_len);
     }
 
-    ret = 0;
+    if (ctx->add_path) {
+        if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+            ret = flb_log_event_encoder_append_body_values(
+                    &ctx->log_encoder,
+                    FLB_LOG_EVENT_CSTRING_VALUE("path"),
+                    FLB_LOG_EVENT_STRING_VALUE(ctx->filepath, ctx->path_len));
+        }
+    }
 
-    flb_input_chunk_append_raw(i_ins, NULL, 0, mp_sbuf.data, mp_sbuf.size);
-    msgpack_sbuffer_destroy(&mp_sbuf);
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        ret = flb_log_event_encoder_commit_record(&ctx->log_encoder);
+    }
+
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        flb_input_log_append(i_ins, NULL, 0,
+                             ctx->log_encoder.output_buffer,
+                             ctx->log_encoder.output_length);
+
+        ret = 0;
+    }
+    else {
+        flb_plg_error(i_ins, "Error encoding record : %d", ret);
+
+        ret = -1;
+    }
+
+    flb_log_event_encoder_reset(&ctx->log_encoder);
 
     return ret;
-
 }
 
 #define KEY_LEN_MAX 32
@@ -170,13 +175,11 @@ static int split_lines_per_record(struct flb_input_instance *i_ins,
 {
     FILE *fp = NULL;
     int i;
+    int ret;
     size_t str_len;
     size_t key_len;
-    int num_map = ctx->lines;
     char *ret_buf;
     char key_str[KEY_LEN_MAX] = {0};
-    msgpack_packer mp_pck;
-    msgpack_sbuffer mp_sbuf;
 
     fp = fopen(ctx->filepath, "r");
     if (fp == NULL) {
@@ -184,25 +187,20 @@ static int split_lines_per_record(struct flb_input_instance *i_ins,
         return -1;
     }
 
-    if (ctx->add_path == FLB_TRUE) {
-        num_map++;
+    ret = flb_log_event_encoder_begin_record(&ctx->log_encoder);
+
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        ret = flb_log_event_encoder_set_current_timestamp(
+                &ctx->log_encoder);
     }
 
-    /* Initialize local msgpack buffer */
-    msgpack_sbuffer_init(&mp_sbuf);
-    msgpack_packer_init(&mp_pck, &mp_sbuf, msgpack_sbuffer_write);
-
-    /* Pack data */
-    msgpack_pack_array(&mp_pck, 2);
-    flb_pack_time_now(&mp_pck);
-    msgpack_pack_map(&mp_pck, num_map);
-
-    if (ctx->add_path == FLB_TRUE) {
-        msgpack_pack_str(&mp_pck, 4);
-        msgpack_pack_str_body(&mp_pck, "path", 4);
-        msgpack_pack_str(&mp_pck, ctx->path_len);
-        msgpack_pack_str_body(&mp_pck,
-                              ctx->filepath, ctx->path_len);
+    if (ctx->add_path) {
+        if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+            ret = flb_log_event_encoder_append_body_values(
+                    &ctx->log_encoder,
+                    FLB_LOG_EVENT_CSTRING_VALUE("path"),
+                    FLB_LOG_EVENT_STRING_VALUE(ctx->filepath, ctx->path_len));
+        }
     }
 
     for (i = 0; i < ctx->lines; i++) {
@@ -221,17 +219,36 @@ static int split_lines_per_record(struct flb_input_instance *i_ins,
             key_len = KEY_LEN_MAX;
         }
 
-        msgpack_pack_str(&mp_pck, key_len);
-        msgpack_pack_str_body(&mp_pck, key_str, key_len);
-        msgpack_pack_str(&mp_pck, str_len);
-        msgpack_pack_str_body(&mp_pck,
-                              ctx->buf, str_len);
+        if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+            ret = flb_log_event_encoder_append_body_values(
+                    &ctx->log_encoder,
+                    FLB_LOG_EVENT_CSTRING_VALUE(key_str),
+                    FLB_LOG_EVENT_STRING_VALUE(ctx->buf, ctx->buf_len));
+        }
     }
 
-    flb_input_chunk_append_raw(i_ins, NULL, 0, mp_sbuf.data, mp_sbuf.size);
-    msgpack_sbuffer_destroy(&mp_sbuf);
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        ret = flb_log_event_encoder_commit_record(&ctx->log_encoder);
+    }
+
+    if (ret == FLB_EVENT_ENCODER_SUCCESS) {
+        flb_input_log_append(i_ins, NULL, 0,
+                             ctx->log_encoder.output_buffer,
+                             ctx->log_encoder.output_length);
+
+        ret = 0;
+    }
+    else {
+        flb_plg_error(i_ins, "Error encoding record : %d", ret);
+
+        ret = -1;
+    }
+
+    flb_log_event_encoder_reset(&ctx->log_encoder);
+
     fclose(fp);
-    return 0;
+
+    return ret;
 }
 
 
@@ -256,81 +273,39 @@ static int in_head_collect(struct flb_input_instance *i_ins,
 static int in_head_config_read(struct flb_in_head_config *ctx,
                                struct flb_input_instance *in)
 {
-    const char *filepath = NULL;
-    const char *pval = NULL;
-
-    /* filepath setting */
-    filepath = flb_input_get_property("file", in);
-    if (!filepath) {
+    int ret;
+    /* Load the config map */
+    ret = flb_input_config_map_set(in, (void *)ctx);
+    if (ret == -1) {
+        flb_plg_error(in, "unable to load configuration");
         return -1;
     }
-    ctx->filepath = filepath;
 
-    pval = flb_input_get_property("key", in);
-    if (pval) {
-        ctx->key      = pval;
-        ctx->key_len  = strlen(pval);
-    }
-    else {
-        ctx->key      = "head";
-        ctx->key_len  = 4;
-    }
 
-    /* buffer size setting */
-    pval = flb_input_get_property("buf_size", in);
-    if (pval != NULL && atoi(pval) > 0) {
-        ctx->buf_size = atoi(pval);
-    }
-    else {
-        ctx->buf_size = DEFAULT_BUF_SIZE;
-    }
+    ctx->key_len = strlen(ctx->key);
 
-    /* interval settings */
-    pval = flb_input_get_property("interval_sec", in);
-    if (pval != NULL && atoi(pval) >= 0) {
-        ctx->interval_sec = atoi(pval);
-    }
-    else {
-        ctx->interval_sec = DEFAULT_INTERVAL_SEC;
-    }
-
-    pval = flb_input_get_property("interval_nsec", in);
-    if (pval != NULL && atoi(pval) >= 0) {
-        ctx->interval_nsec = atoi(pval);
-    }
-    else {
-        ctx->interval_nsec = DEFAULT_INTERVAL_NSEC;
-    }
-
-    pval = flb_input_get_property("split_line", in);
-    if (pval != NULL && flb_utils_bool(pval)) {
-        ctx->split_line = FLB_TRUE;
+    /* only set lines if not explicitly set */
+    if (ctx->split_line && ctx->lines <= 0) {
         ctx->lines      = 10;
     }
-    else {
-        ctx->split_line = FLB_FALSE;
-    }
-
-    pval = flb_input_get_property("lines", in);
-    if (pval != NULL && atoi(pval) >= 0) {
-        ctx->lines = atoi(pval);
-    }
-    else {
-        ctx->lines = 0; /* read bytes mode */
-    }
 
     if (ctx->interval_sec <= 0 && ctx->interval_nsec <= 0) {
         /* Illegal settings. Override them. */
-        ctx->interval_sec = DEFAULT_INTERVAL_SEC;
-        ctx->interval_nsec = DEFAULT_INTERVAL_NSEC;
+        ctx->interval_sec = atoi(DEFAULT_INTERVAL_SEC);
+        ctx->interval_nsec = atoi(DEFAULT_INTERVAL_NSEC);
     }
 
-    pval = flb_input_get_property("add_path", in);
-    if (pval) {
-        if (strcasecmp(pval, "true") == 0 || strcasecmp(pval, "on") == 0) {
-            ctx->add_path = FLB_TRUE;
-            ctx->path_len = strlen(ctx->filepath);
-        }
+    if (ctx->add_path) {
+        ctx->path_len = strlen(ctx->filepath);
+    }
+
+    ret = flb_log_event_encoder_init(&ctx->log_encoder,
+                                     FLB_LOG_EVENT_FORMAT_DEFAULT);
+
+    if (ret != FLB_EVENT_ENCODER_SUCCESS) {
+        flb_plg_error(ctx->ins, "error initializing event encoder : %d", ret);
+
+        return -1;
     }
 
     flb_plg_debug(ctx->ins, "buf_size=%zu path=%s",
@@ -347,10 +322,13 @@ static void delete_head_config(struct flb_in_head_config *ctx)
         return;
     }
 
+    flb_log_event_encoder_destroy(&ctx->log_encoder);
+
     /* release buffer */
     if (ctx->buf) {
         flb_free(ctx->buf);
     }
+
     flb_free(ctx);
 }
 
@@ -435,6 +413,51 @@ static int in_head_exit(void *data, struct flb_config *config)
     return 0;
 }
 
+/* Configuration properties map */
+static struct flb_config_map config_map[] = {
+    {
+     FLB_CONFIG_MAP_STR, "file", NULL,
+     0, FLB_TRUE, offsetof(struct flb_in_head_config, filepath),
+     "Set the file"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "key", "head",
+     0, FLB_TRUE, offsetof(struct flb_in_head_config, key),
+     "Set the record key"
+    },
+    {
+      FLB_CONFIG_MAP_INT, "buf_size", DEFAULT_BUF_SIZE,
+      0, FLB_TRUE, offsetof(struct flb_in_head_config, buf_size),
+      "Set the read buffer size"
+    },
+    {
+      FLB_CONFIG_MAP_BOOL, "split_line", "false",
+      0, FLB_TRUE, offsetof(struct flb_in_head_config, split_line),
+      "generate key/value pair per line"
+    },
+    {
+      FLB_CONFIG_MAP_INT, "lines", "0",
+      0, FLB_TRUE, offsetof(struct flb_in_head_config, lines),
+      "Line number to read"
+    },
+    {
+      FLB_CONFIG_MAP_BOOL, "add_path", "false",
+      0, FLB_TRUE, offsetof(struct flb_in_head_config, add_path),
+      "append filepath to records"
+    },
+    {
+      FLB_CONFIG_MAP_INT, "interval_sec", DEFAULT_INTERVAL_SEC,
+      0, FLB_TRUE, offsetof(struct flb_in_head_config, interval_sec),
+      "Set the collector interval"
+    },
+    {
+      FLB_CONFIG_MAP_INT, "interval_nsec", DEFAULT_INTERVAL_NSEC,
+      0, FLB_TRUE, offsetof(struct flb_in_head_config, interval_nsec),
+      "Set the collector interval (nanoseconds)"
+    },
+    /* EOF */
+    {0}
+};
 
 struct flb_input_plugin in_head_plugin = {
     .name         = "head",
@@ -445,5 +468,6 @@ struct flb_input_plugin in_head_plugin = {
     .cb_flush_buf = NULL,
     .cb_pause     = in_head_pause,
     .cb_resume    = in_head_resume,
+    .config_map   = config_map,
     .cb_exit      = in_head_exit
 };
