@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -37,11 +36,19 @@ static int http_conn_event(void *data)
     char *tmp;
     char *request_end;
     size_t request_len;
-    struct http_conn *conn = data;
+    struct flb_connection *connection;
+    struct http_conn *conn;
     struct mk_event *event;
-    struct flb_http *ctx = conn->ctx;
+    struct flb_http *ctx;
+
+    connection = (struct flb_connection *) data;
+
+    conn = connection->user_data;
+
+    ctx = conn->ctx;
+
+    event = &connection->event;
 
-    event = &conn->event;
     if (event->mask & MK_EVENT_READ) {
         available = (conn->buf_size - conn->buf_len) - 1;
         if (available < 1) {
@@ -59,7 +66,7 @@ static int http_conn_event(void *data)
                 flb_errno();
                 return -1;
             }
-            flb_plg_trace(ctx->ins, "fd=%i buffer realloc %i -> %i",
+            flb_plg_trace(ctx->ins, "fd=%i buffer realloc %i -> %zu",
                           event->fd, conn->buf_size, size);
 
             conn->buf_data = tmp;
@@ -68,14 +75,17 @@ static int http_conn_event(void *data)
         }
 
         /* Read data */
-        bytes = recv(conn->fd, conn->buf_data + conn->buf_len, available, 0);
+        bytes = flb_io_net_read(connection,
+                                (void *) &conn->buf_data[conn->buf_len],
+                                available);
+
         if (bytes <= 0) {
             flb_plg_trace(ctx->ins, "fd=%i closed connection", event->fd);
             http_conn_del(conn);
             return -1;
         }
 
-        flb_plg_trace(ctx->ins, "read()=%i pre_len=%i now_len=%i",
+        flb_plg_trace(ctx->ins, "read()=%zi pre_len=%i now_len=%zi",
                       bytes, conn->buf_len, conn->buf_len + bytes);
         conn->buf_len += bytes;
         conn->buf_data[conn->buf_len] = '\0';
@@ -124,7 +134,7 @@ static int http_conn_event(void *data)
                  * handled, the additional memset intends to wipe any left over data
                  * from the headers parsed in the previous request.
                  */
-                memset(&conn->session.parser, 0, sizeof(mk_http_parser));
+                memset(&conn->session.parser, 0, sizeof(struct mk_http_parser));
                 mk_http_parser_init(&conn->session.parser);
                 http_conn_request_init(&conn->session, &conn->request);
             }
@@ -136,7 +146,7 @@ static int http_conn_event(void *data)
              * handled, the additional memset intends to wipe any left over data
              * from the headers parsed in the previous request.
              */
-            memset(&conn->session.parser, 0, sizeof(mk_http_parser));
+            memset(&conn->session.parser, 0, sizeof(struct mk_http_parser));
             mk_http_parser_init(&conn->session.parser);
             http_conn_request_init(&conn->session, &conn->request);
         }
@@ -199,11 +209,11 @@ static void http_conn_request_init(struct mk_http_session *session,
     request->session = session;
 }
 
-struct http_conn *http_conn_add(int fd, struct flb_http *ctx)
+struct http_conn *http_conn_add(struct flb_connection *connection,
+                                struct flb_http *ctx)
 {
-    int ret;
     struct http_conn *conn;
-    struct mk_event *event;
+    int               ret;
 
     conn = flb_calloc(1, sizeof(struct http_conn));
     if (!conn) {
@@ -211,40 +221,47 @@ struct http_conn *http_conn_add(int fd, struct flb_http *ctx)
         return NULL;
     }
 
+    conn->connection = connection;
+
     /* Set data for the event-loop */
-    event = &conn->event;
-    MK_EVENT_NEW(event);
-    event->fd      = fd;
-    event->type    = FLB_ENGINE_EV_CUSTOM;
-    event->handler = http_conn_event;
+    MK_EVENT_NEW(&connection->event);
+
+    connection->user_data     = conn;
+    connection->event.type    = FLB_ENGINE_EV_CUSTOM;
+    connection->event.handler = http_conn_event;
 
     /* Connection info */
-    conn->fd      = fd;
     conn->ctx     = ctx;
     conn->buf_len = 0;
 
     conn->buf_data = flb_malloc(ctx->buffer_chunk_size);
     if (!conn->buf_data) {
         flb_errno();
-        flb_socket_close(fd);
+
         flb_plg_error(ctx->ins, "could not allocate new connection");
         flb_free(conn);
+
         return NULL;
     }
     conn->buf_size = ctx->buffer_chunk_size;
 
     /* Register instance into the event loop */
-    ret = mk_event_add(ctx->evl, fd, FLB_ENGINE_EV_CUSTOM, MK_EVENT_READ, conn);
+    ret = mk_event_add(flb_engine_evl_get(),
+                       connection->fd,
+                       FLB_ENGINE_EV_CUSTOM,
+                       MK_EVENT_READ,
+                       &connection->event);
     if (ret == -1) {
         flb_plg_error(ctx->ins, "could not register new connection");
-        flb_socket_close(fd);
+
         flb_free(conn->buf_data);
         flb_free(conn);
+
         return NULL;
     }
 
     /* Initialize HTTP Session: this is a custom context for Monkey HTTP */
-    http_conn_session_init(&conn->session, ctx->server, conn->fd);
+    http_conn_session_init(&conn->session, ctx->server, conn->connection->fd);
 
     /* Initialize HTTP Request: this is the initial request and it will be reinitialized
      * automatically after the request is handled so it can be used for the next one.
@@ -253,18 +270,23 @@ struct http_conn *http_conn_add(int fd, struct flb_http *ctx)
 
     /* Link connection node to parent context list */
     mk_list_add(&conn->_head, &ctx->connections);
+
     return conn;
 }
 
 int http_conn_del(struct http_conn *conn)
 {
-    struct flb_http *ctx;
+    if (conn->session.channel != NULL) {
+        mk_channel_release(conn->session.channel);
+    }
 
-    ctx = conn->ctx;
+    /* The downstream unregisters the file descriptor from the event-loop
+     * so there's nothing to be done by the plugin
+     */
+    flb_downstream_conn_release(conn->connection);
 
-    mk_event_del(ctx->evl, &conn->event);
     mk_list_del(&conn->_head);
-    flb_socket_close(conn->fd);
+
     flb_free(conn->buf_data);
     flb_free(conn);
 
