@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -24,9 +23,13 @@
 #include <fluent-bit/flb_config.h>
 #include <fluent-bit/flb_input.h>
 #include <fluent-bit/flb_pack.h>
+#include <fluent-bit/flb_log_event_encoder.h>
+#include <fluent-bit/flb_pthread.h>
 
 #define LIB_BUF_CHUNK   65536
 
+pthread_key_t flb_active_lib_context;
+
 /* Library input configuration & context */
 struct flb_in_lib_config {
     int fd;                     /* instance input channel  */
@@ -34,6 +37,7 @@ struct flb_in_lib_config {
     int buf_len;                /* read buffer length      */
     char *buf_data;             /* the real buffer         */
 
+    struct flb_log_event_encoder log_encoder;
     struct flb_pack_state state;
     struct flb_input_instance *ins;
 };
