@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -22,6 +21,7 @@
 #include <fluent-bit/flb_input_plugin.h>
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_network.h>
+#include <fluent-bit/flb_downstream.h>
 #include <fluent-bit/flb_config_map.h>
 
 #include "mqtt.h"
@@ -32,8 +32,10 @@
 static int in_mqtt_init(struct flb_input_instance *in,
                         struct flb_config *config, void *data)
 {
-    int ret;
+    unsigned short int         port;
+    int                        ret;
     struct flb_in_mqtt_config *ctx;
+
     (void) data;
 
     /* Allocate space for the configuration */
@@ -47,23 +49,33 @@ static int in_mqtt_init(struct flb_input_instance *in,
     /* Set the context */
     flb_input_set_context(in, ctx);
 
-    /* Create TCP server */
-    ctx->server_fd = flb_net_server(ctx->tcp_port, ctx->listen);
-    if (ctx->server_fd > 0) {
-        flb_plg_info(ctx->ins, "listening on %s:%s", ctx->listen, ctx->tcp_port);
-    }
-    else {
-        flb_plg_error(ctx->ins, "could not bind address %s:%s",
+    /* Create downstream */
+    port = (unsigned short int) strtoul(ctx->tcp_port, NULL, 10);
+
+    ctx->downstream = flb_downstream_create(FLB_TRANSPORT_TCP,
+                                            in->flags,
+                                            ctx->listen,
+                                            port,
+                                            in->tls,
+                                            config,
+                                            &in->net_setup);
+
+    if (ctx->downstream == NULL) {
+        flb_plg_error(ctx->ins,
+                      "could not initialize downstream on %s:%s. Aborting",
                       ctx->listen, ctx->tcp_port);
+
         mqtt_config_free(ctx);
+
         return -1;
     }
-    ctx->evl = config->evl;
+
+    flb_input_downstream_set(ctx->downstream, ctx->ins);
 
     /* Collect upon data available on the standard input */
     ret = flb_input_set_collector_event(in,
                                         in_mqtt_collect,
-                                        ctx->server_fd,
+                                        ctx->downstream->server_fd,
                                         config);
     if (ret == -1) {
         flb_plg_error(ctx->ins, "could not set collector for MQTT input plugin");
@@ -82,22 +94,30 @@ static int in_mqtt_init(struct flb_input_instance *in,
 int in_mqtt_collect(struct flb_input_instance *ins,
                     struct flb_config *config, void *in_context)
 {
-    int fd;
-    struct flb_in_mqtt_config *ctx = in_context;
-    struct mqtt_conn *conn;
+    struct flb_connection     *connection;
+    struct mqtt_conn          *conn;
+    struct flb_in_mqtt_config *ctx;
 
-    /* Accept the new connection */
-    fd = flb_net_accept(ctx->server_fd);
-    if (fd == -1) {
+    ctx = in_context;
+
+    connection = flb_downstream_conn_get(ctx->downstream);
+
+    if (connection == NULL) {
         flb_plg_error(ctx->ins, "could not accept new connection");
+
         return -1;
     }
 
-    flb_plg_debug(ctx->ins, "[fd=%i] new TCP connection", fd);
-    conn = mqtt_conn_add(fd, ctx);
+    flb_plg_debug(ctx->ins, "[fd=%i] new TCP connection", connection->fd);
+
+    conn = mqtt_conn_add(connection, ctx);
+
     if (!conn) {
+        flb_downstream_conn_release(connection);
+
         return -1;
     }
+
     return 0;
 }
 
@@ -111,6 +131,7 @@ static int in_mqtt_exit(void *data, struct flb_config *config)
     }
 
     mqtt_conn_destroy_all(ctx);
+
     mqtt_config_free(ctx);
 
     return 0;
@@ -118,7 +139,11 @@ static int in_mqtt_exit(void *data, struct flb_config *config)
 
 /* Configuration properties map */	
 static struct flb_config_map config_map[] = {	
-        	
+    {
+     FLB_CONFIG_MAP_STR, "payload_key", NULL,
+     0, FLB_TRUE, offsetof(struct flb_in_mqtt_config, payload_key),
+     "Key where the payload will be preserved"
+    },
     /* EOF */	
     {0}	
 };
@@ -133,5 +158,5 @@ struct flb_input_plugin in_mqtt_plugin = {
     .cb_flush_buf = NULL,
     .cb_exit      = in_mqtt_exit,
     .config_map   = config_map,
-    .flags        = FLB_INPUT_NET,
+    .flags        = FLB_INPUT_NET_SERVER | FLB_IO_OPT_TLS
 };
