@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -21,6 +20,7 @@
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_config.h>
 #include <fluent-bit/flb_input.h>
+#include <fluent-bit/flb_input_plugin.h>
 #include <fluent-bit/flb_str.h>
 #include <fluent-bit/flb_log.h>
 #include <fluent-bit/flb_parser.h>
@@ -33,37 +33,72 @@
 struct flb_syslog *syslog_conf_create(struct flb_input_instance *ins,
                                       struct flb_config *config)
 {
-    const char *tmp;
+    int ret;
     char port[16];
     struct flb_syslog *ctx;
 
     ctx = flb_calloc(1, sizeof(struct flb_syslog));
-    if (!ctx) {
+
+    if (ctx == NULL) {
         flb_errno();
+
         return NULL;
     }
-    ctx->evl = config->evl;
+
     ctx->ins = ins;
-    ctx->buffer_data = NULL;
+
+    ctx->log_encoder = flb_log_event_encoder_create(FLB_LOG_EVENT_FORMAT_DEFAULT);
+
+    if (ctx->log_encoder == NULL) {
+        flb_plg_error(ins, "could not initialize event encoder");
+        syslog_conf_destroy(ctx);
+
+        return NULL;
+    }
+
     mk_list_init(&ctx->connections);
 
+    ret = flb_input_config_map_set(ins, (void *)ctx);
+    if (ret == -1) {
+        flb_log_event_encoder_destroy(ctx->log_encoder);
+
+        flb_plg_error(ins, "unable to load configuration");
+        flb_free(ctx);
+
+        return NULL;
+    }
+
     /* Syslog mode: unix_udp, unix_tcp, tcp or udp */
-    tmp = flb_input_get_property("mode", ins);
-    if (tmp) {
-        if (strcasecmp(tmp, "unix_tcp") == 0) {
+    if (ctx->mode_str) {
+#ifdef FLB_SYSTEM_WINDOWS
+        if (strcasestr(ctx->mode_str, "unix") != NULL) {
+            flb_log_event_encoder_destroy(ctx->log_encoder);
+
+            flb_plg_error(ins, "unix sockets are note available in windows");
+            flb_free(ctx);
+
+            return NULL;
+        }
+
+#undef FLB_SYSLOG_UNIX_UDP
+#define FLB_SYSLOG_UNIX_UDP FLB_SYSLOG_UDP
+#endif
+        if (strcasecmp(ctx->mode_str, "unix_tcp") == 0) {
             ctx->mode = FLB_SYSLOG_UNIX_TCP;
         }
-        else if (strcasecmp(tmp, "unix_udp") == 0) {
+        else if (strcasecmp(ctx->mode_str, "unix_udp") == 0) {
             ctx->mode = FLB_SYSLOG_UNIX_UDP;
         }
-        else if (strcasecmp(tmp, "tcp") == 0) {
+        else if (strcasecmp(ctx->mode_str, "tcp") == 0) {
             ctx->mode = FLB_SYSLOG_TCP;
         }
-        else if (strcasecmp(tmp, "udp") == 0) {
+        else if (strcasecmp(ctx->mode_str, "udp") == 0) {
             ctx->mode = FLB_SYSLOG_UDP;
         }
         else {
-            flb_error("[in_syslog] Unknown syslog mode %s", tmp);
+            flb_log_event_encoder_destroy(ctx->log_encoder);
+
+            flb_error("[in_syslog] Unknown syslog mode %s", ctx->mode_str);
             flb_free(ctx);
             return NULL;
         }
@@ -83,41 +118,46 @@ struct flb_syslog *syslog_conf_create(struct flb_input_instance *ins,
 
     /* Unix socket path and permission */
     if (ctx->mode == FLB_SYSLOG_UNIX_UDP || ctx->mode == FLB_SYSLOG_UNIX_TCP) {
-        tmp = flb_input_get_property("path", ins);
-        if (tmp) {
-            ctx->unix_path = flb_strdup(tmp);
-        }
-
-        tmp = flb_input_get_property("unix_perm", ins);
-        if (tmp) {
-            ctx->unix_perm = strtol(tmp, NULL, 8) & 07777;
+        if (ctx->unix_perm_str) {
+            ctx->unix_perm = strtol(ctx->unix_perm_str, NULL, 8) & 07777;
         } else {
             ctx->unix_perm = 0644;
         }
     }
 
     /* Buffer Chunk Size */
-    tmp = flb_input_get_property("buffer_chunk_size", ins);
-    if (!tmp) {
-        ctx->buffer_chunk_size = FLB_SYSLOG_CHUNK; /* 32KB */
-    }
-    else {
-        ctx->buffer_chunk_size = flb_utils_size_to_bytes(tmp);
+    if (ctx->buffer_chunk_size == -1) {
+        flb_log_event_encoder_destroy(ctx->log_encoder);
+
+        flb_plg_error(ins, "invalid buffer_chunk_size");
+        flb_free(ctx);
+        return NULL; 
     }
 
     /* Buffer Max Size */
-    tmp = flb_input_get_property("buffer_max_size", ins);
-    if (!tmp) {
+    if (ctx->buffer_max_size == -1) {
+        flb_log_event_encoder_destroy(ctx->log_encoder);
+
+        flb_plg_error(ins, "invalid buffer_max_size");
+        flb_free(ctx);
+        return NULL;
+    }
+    else if (ctx->buffer_max_size == 0) {
         ctx->buffer_max_size = ctx->buffer_chunk_size;
     }
-    else {
-        ctx->buffer_max_size  = flb_utils_size_to_bytes(tmp);
+
+    /* Socket rcv buffer size */
+    if (ctx->receive_buffer_size == -1 || ctx->receive_buffer_size>INT_MAX) {
+        flb_log_event_encoder_destroy(ctx->log_encoder);
+
+        flb_plg_error(ins, "invalid receive_buffer_size");
+        flb_free(ctx);
+        return NULL;
     }
 
     /* Parser */
-    tmp = flb_input_get_property("parser", ins);
-    if (tmp) {
-        ctx->parser = flb_parser_get(tmp, config);
+    if (ctx->parser_name) {
+        ctx->parser = flb_parser_get(ctx->parser_name, config);
     }
     else {
         if (ctx->mode == FLB_SYSLOG_TCP || ctx->mode == FLB_SYSLOG_UDP) {
@@ -129,6 +169,8 @@ struct flb_syslog *syslog_conf_create(struct flb_input_instance *ins,
     }
 
     if (!ctx->parser) {
+        flb_log_event_encoder_destroy(ctx->log_encoder);
+
         flb_error("[in_syslog] parser not set");
         syslog_conf_destroy(ctx);
         return NULL;
@@ -139,11 +181,12 @@ struct flb_syslog *syslog_conf_create(struct flb_input_instance *ins,
 
 int syslog_conf_destroy(struct flb_syslog *ctx)
 {
-    if (ctx->buffer_data) {
-        flb_free(ctx->buffer_data);
-        ctx->buffer_data = NULL;
+    if (ctx->log_encoder != NULL) {
+        flb_log_event_encoder_destroy(ctx->log_encoder);
     }
+
     syslog_server_destroy(ctx);
+
     flb_free(ctx);
 
     return 0;
