@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -69,26 +68,53 @@ static int in_tail_collect_pending(struct flb_input_instance *ins,
     struct flb_tail_config *ctx = in_context;
     struct flb_tail_file *file;
     struct stat st;
+    uint64_t pre;
+    uint64_t total_processed = 0;
 
     /* Iterate promoted event files with pending bytes */
     mk_list_foreach_safe(head, tmp, &ctx->files_event) {
         file = mk_list_entry(head, struct flb_tail_file, _head);
 
-        /* Gather current file size */
-        ret = fstat(file->fd, &st);
-        if (ret == -1) {
-            flb_errno();
-            flb_tail_file_remove(file);
-            continue;
+        if (file->watch_fd == -1 ||
+            (file->offset >= file->size)) {
+            ret = fstat(file->fd, &st);
+
+            if (ret == -1) {
+                flb_errno();
+                flb_tail_file_remove(file);
+                continue;
+            }
+
+            file->size = st.st_size;
+            file->pending_bytes = (file->size - file->offset);
+        }
+        else {
+            memset(&st, 0, sizeof(struct stat));
+        }
+
+        if (file->pending_bytes <= 0) {
+            file->pending_bytes = (file->size - file->offset);
         }
-        file->size = st.st_size;
-        file->pending_bytes = (file->size - file->offset);
 
         if (file->pending_bytes <= 0) {
             continue;
         }
 
+        if (ctx->event_batch_size > 0 &&
+            total_processed >= ctx->event_batch_size) {
+            break;
+        }
+
+        /* get initial offset to calculate the number of processed bytes later */
+        pre = file->offset;
+
         ret = flb_tail_file_chunk(file);
+
+        /* Update the total number of bytes processed */
+        if (file->offset > pre) {
+            total_processed += (file->offset - pre);
+        }
+
         switch (ret) {
         case FLB_TAIL_ERROR:
             /* Could not longer read the file */
@@ -100,8 +126,8 @@ static int in_tail_collect_pending(struct flb_input_instance *ins,
              * Adjust counter to verify if we need a further read(2) later.
              * For more details refer to tail_fs_inotify.c:96.
              */
-            if (file->offset < st.st_size) {
-                file->pending_bytes = (st.st_size - file->offset);
+            if (file->offset < file->size) {
+                file->pending_bytes = (file->size - file->offset);
                 active++;
             }
             else {
@@ -190,7 +216,7 @@ static int in_tail_collect_static(struct flb_input_instance *ins,
             if (file->config->exit_on_eof) {
                 flb_plg_info(ctx->ins, "inode=%"PRIu64" file=%s ended, stop",
                              file->inode, file->name);
-                if (mk_list_size(&ctx->files_static) == 1) {
+                if (ctx->files_static_count == 1) {
                     flb_engine_exit(config);
                 }
             }
@@ -227,7 +253,7 @@ static int in_tail_collect_static(struct flb_input_instance *ins,
              * when to stop processing the static list.
              */
             if (alter_size == 0) {
-                pre_size = mk_list_size(&ctx->files_static);
+                pre_size = ctx->files_static_count;
             }
             ret = flb_tail_file_to_event(file);
             if (ret == -1) {
@@ -237,7 +263,7 @@ static int in_tail_collect_static(struct flb_input_instance *ins,
             }
 
             if (alter_size == 0) {
-                pos_size = mk_list_size(&ctx->files_static);
+                pos_size = ctx->files_static_count;
                 if (pre_size == pos_size) {
                     alter_size++;
                 }
@@ -460,6 +486,7 @@ static int in_tail_exit(void *data, struct flb_config *config)
     struct flb_tail_config *ctx = data;
 
     flb_tail_file_remove_all(ctx);
+    flb_tail_fs_exit(ctx);
     flb_tail_config_destroy(ctx);
 
     return 0;
@@ -552,6 +579,14 @@ static struct flb_config_map config_map[] = {
      FLB_CONFIG_MAP_TIME, "watcher_interval", "2s",
      0, FLB_TRUE, offsetof(struct flb_tail_config, watcher_interval),
     },
+    {
+     FLB_CONFIG_MAP_TIME, "progress_check_interval", "2s",
+     0, FLB_TRUE, offsetof(struct flb_tail_config, progress_check_interval),
+    },
+    {
+     FLB_CONFIG_MAP_INT, "progress_check_interval_nsec", "0",
+     0, FLB_TRUE, offsetof(struct flb_tail_config, progress_check_interval_nsec),
+    },
     {
      FLB_CONFIG_MAP_TIME, "rotate_wait", FLB_TAIL_ROTATE_WAIT,
      0, FLB_TRUE, offsetof(struct flb_tail_config, rotate_wait),
@@ -591,10 +626,8 @@ static struct flb_config_map config_map[] = {
     {
      FLB_CONFIG_MAP_TIME, "ignore_older", "0",
      0, FLB_TRUE, offsetof(struct flb_tail_config, ignore_older),
-     "ignore records older than 'ignore_older'. Supports m,h,d (minutes, "
-     "hours, days) syntax. Default behavior is to read all records. Option "
-     "only available when a Parser is specified and it can parse the time "
-     "of a record."
+     "ignore files older than 'ignore_older'. Supports m,h,d (minutes, "
+     "hours, days) syntax. Default behavior is to read all the files."
     },
     {
      FLB_CONFIG_MAP_SIZE, "buffer_chunk_size", FLB_TAIL_CHUNK,
@@ -618,6 +651,15 @@ static struct flb_config_map config_map[] = {
      "in question set's the maximum number of bytes to process per iteration "
      "for the static files monitored."
     },
+    {
+     FLB_CONFIG_MAP_SIZE, "event_batch_size", FLB_TAIL_EVENT_BATCH_SIZE,
+     0, FLB_TRUE, offsetof(struct flb_tail_config, event_batch_size),
+     "When Fluent Bit is processing files in event based mode the amount of"
+     "data available for consumption could be too much and cause the input plugin "
+     "to over extend and smother other plugins"
+     "The configuration property sets the maximum number of bytes to process per iteration "
+     "for the files monitored (in event mode)."
+    },
     {
      FLB_CONFIG_MAP_BOOL, "skip_long_lines", "false",
      0, FLB_TRUE, offsetof(struct flb_tail_config, skip_long_lines),
