@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -34,22 +33,32 @@
 static int in_tcp_collect(struct flb_input_instance *in,
                           struct flb_config *config, void *in_context)
 {
-    int fd;
-    struct flb_in_tcp_config *ctx = in_context;
-    struct tcp_conn *conn;
+    struct flb_connection    *connection;
+    struct tcp_conn          *conn;
+    struct flb_in_tcp_config *ctx;
 
-    /* Accept the new connection */
-    fd = flb_net_accept(ctx->server_fd);
-    if (fd == -1) {
+    ctx = in_context;
+
+    connection = flb_downstream_conn_get(ctx->downstream);
+
+    if (connection == NULL) {
         flb_plg_error(ctx->ins, "could not accept new connection");
+
         return -1;
     }
 
-    flb_plg_trace(ctx->ins, "new TCP connection arrived FD=%i", fd);
-    conn = tcp_conn_add(fd, ctx);
-    if (!conn) {
+    flb_plg_trace(ctx->ins, "new TCP connection arrived FD=%i", connection->fd);
+
+    conn = tcp_conn_add(connection, ctx);
+
+    if (conn == NULL) {
+        flb_plg_error(ctx->ins, "could not accept new connection");
+
+        flb_downstream_conn_release(connection);
+
         return -1;
     }
+
     return 0;
 }
 
@@ -57,8 +66,10 @@ static int in_tcp_collect(struct flb_input_instance *in,
 static int in_tcp_init(struct flb_input_instance *in,
                       struct flb_config *config, void *data)
 {
-    int ret;
+    unsigned short int        port;
+    int                       ret;
     struct flb_in_tcp_config *ctx;
+
     (void) data;
 
     /* Allocate space for the configuration */
@@ -66,38 +77,49 @@ static int in_tcp_init(struct flb_input_instance *in,
     if (!ctx) {
         return -1;
     }
+    ctx->collector_id = -1;
     ctx->ins = in;
     mk_list_init(&ctx->connections);
 
     /* Set the context */
     flb_input_set_context(in, ctx);
 
-    /* Create TCP server */
-    ctx->server_fd = flb_net_server(ctx->tcp_port, ctx->listen);
-    if (ctx->server_fd > 0) {
-        flb_plg_info(ctx->ins, "listening on %s:%s", ctx->listen, ctx->tcp_port);
-    }
-    else {
-        flb_plg_error(ctx->ins, "could not bind address %s:%s. Aborting",
+    port = (unsigned short int) strtoul(ctx->tcp_port, NULL, 10);
+
+    ctx->downstream = flb_downstream_create(FLB_TRANSPORT_TCP,
+                                            in->flags,
+                                            ctx->listen,
+                                            port,
+                                            in->tls,
+                                            config,
+                                            &in->net_setup);
+
+    if (ctx->downstream == NULL) {
+        flb_plg_error(ctx->ins,
+                      "could not initialize downstream on %s:%s. Aborting",
                       ctx->listen, ctx->tcp_port);
+
         tcp_config_destroy(ctx);
+
         return -1;
     }
-    flb_net_socket_nonblocking(ctx->server_fd);
 
-    ctx->evl = config->evl;
+    flb_input_downstream_set(ctx->downstream, ctx->ins);
 
     /* Collect upon data available on the standard input */
     ret = flb_input_set_collector_socket(in,
                                          in_tcp_collect,
-                                         ctx->server_fd,
+                                         ctx->downstream->server_fd,
                                          config);
     if (ret == -1) {
         flb_plg_error(ctx->ins, "Could not set collector for IN_TCP input plugin");
         tcp_config_destroy(ctx);
+
         return -1;
     }
 
+    ctx->collector_id = ret;
+
     return 0;
 }
 
@@ -105,19 +127,49 @@ static int in_tcp_exit(void *data, struct flb_config *config)
 {
     struct mk_list *tmp;
     struct mk_list *head;
-    (void) *config;
-    struct flb_in_tcp_config *ctx = data;
+    struct flb_in_tcp_config *ctx;
     struct tcp_conn *conn;
 
+    (void) *config;
+
+    ctx = data;
+
     mk_list_foreach_safe(head, tmp, &ctx->connections) {
         conn = mk_list_entry(head, struct tcp_conn, _head);
+
         tcp_conn_del(conn);
     }
 
     tcp_config_destroy(ctx);
+
     return 0;
 }
 
+static struct flb_config_map config_map[] = {
+    {
+     FLB_CONFIG_MAP_STR, "format", (char *)NULL,
+     0, FLB_TRUE, offsetof(struct flb_in_tcp_config, format_name),
+     "Set the format: json or none"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "separator", (char *)NULL,
+     0, FLB_TRUE, offsetof(struct flb_in_tcp_config, raw_separator),
+     "Set separator"
+    },
+    {
+      FLB_CONFIG_MAP_STR, "chunk_size", (char *)NULL,
+      0, FLB_TRUE, offsetof(struct flb_in_tcp_config, chunk_size_str),
+      "Set the chunk size"
+    },
+    {
+      FLB_CONFIG_MAP_STR, "buffer_size", (char *)NULL,
+      0, FLB_TRUE, offsetof(struct flb_in_tcp_config, buffer_size_str),
+      "Set the buffer size"
+    },
+    /* EOF */
+    {0}
+};
+
 /* Plugin reference */
 struct flb_input_plugin in_tcp_plugin = {
     .name         = "tcp",
@@ -127,5 +179,6 @@ struct flb_input_plugin in_tcp_plugin = {
     .cb_collect   = in_tcp_collect,
     .cb_flush_buf = NULL,
     .cb_exit      = in_tcp_exit,
-    .flags        = FLB_INPUT_NET,
+    .config_map   = config_map,
+    .flags        = FLB_INPUT_NET_SERVER | FLB_IO_OPT_TLS
 };
