@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -31,11 +30,13 @@
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_output.h>
 #include <fluent-bit/flb_sds.h>
+#include <fluent-bit/flb_record_accessor.h>
 
 struct flb_azure {
     /* account setup */
     flb_sds_t customer_id;
     flb_sds_t log_type;
+    flb_sds_t log_type_key;
     flb_sds_t shared_key;
     flb_sds_t dec_shared_key;
 
@@ -46,6 +47,10 @@ struct flb_azure {
 
     /* records */
     flb_sds_t time_key;
+    struct flb_record_accessor *ra_prefix_key;
+
+    /* time_generated: on/off */
+    int time_generated;
 
     /* Upstream connection to the backend server */
     struct flb_upstream *u;
