@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -25,8 +24,7 @@
 #include <fluent-bit/flb_pack.h>
 #include <fluent-bit/flb_config_map.h>
 #include <fluent-bit/flb_gzip.h>
-#include <mbedtls/sha256.h>
-#include <mbedtls/base64.h>
+#include <fluent-bit/flb_base64.h>
 
 #include <msgpack.h>
 
@@ -43,6 +41,7 @@ static int azure_blob_format(struct flb_config *config,
                              struct flb_input_instance *ins,
                              void *plugin_context,
                              void *flush_ctx,
+                             int event_type,
                              const char *tag, int tag_len,
                              const void *data, size_t bytes,
                              void **out_data, size_t *out_size)
@@ -81,7 +80,7 @@ static int send_blob(struct flb_config *config,
     void *payload_buf;
     size_t payload_size;
     struct flb_http_client *c;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
 
     if (ctx->btype == AZURE_BLOB_APPENDBLOB) {
         uri = azb_append_blob_uri(ctx, tag);
@@ -113,6 +112,7 @@ static int send_blob(struct flb_config *config,
     /* Format the data */
     ret = azure_blob_format(config, i_ins,
                             ctx, NULL,
+                            FLB_EVENT_TYPE_LOGS,
                             tag, tag_len,
                             data, bytes,
                             &out_buf, &out_size);
@@ -200,6 +200,12 @@ static int send_blob(struct flb_config *config,
         return FLB_OK;
     }
     else if (c->resp.status == 404) {
+        /* delete "&sig=..." in the c->uri for security */
+        char *p = strstr(c->uri, "&sig=");
+        if (p) {
+            *p = '\0';
+        }
+
         flb_plg_info(ctx->ins, "blob not found: %s", c->uri);
         flb_http_client_destroy(c);
         return CREATE_BLOB;
@@ -226,7 +232,7 @@ static int create_blob(struct flb_azure_blob *ctx, char *name)
     size_t b_sent;
     flb_sds_t uri = NULL;
     struct flb_http_client *c;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
 
     uri = azb_uri_create_blob(ctx, name);
     if (!uri) {
@@ -269,6 +275,11 @@ static int create_blob(struct flb_azure_blob *ctx, char *name)
     }
 
     if (c->resp.status == 201) {
+        /* delete "&sig=..." in the c->uri for security */
+        char *p = strstr(c->uri, "&sig=");
+        if (p) {
+            *p = '\0';
+        }
         flb_plg_info(ctx->ins, "blob created successfully: %s", c->uri);
     }
     else {
@@ -296,7 +307,7 @@ static int create_container(struct flb_azure_blob *ctx, char *name)
     size_t b_sent;
     flb_sds_t uri;
     struct flb_http_client *c;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
 
     /* Get upstream connection */
     u_conn = flb_upstream_conn_get(ctx->u);
@@ -375,7 +386,7 @@ static int ensure_container(struct flb_azure_blob *ctx)
     size_t b_sent;
     flb_sds_t uri;
     struct flb_http_client *c;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
 
     uri = azb_uri_ensure_or_create_container(ctx);
     if (!uri) {
@@ -454,8 +465,8 @@ static int cb_azure_blob_init(struct flb_output_instance *ins,
     return 0;
 }
 
-static void cb_azure_blob_flush(const void *data, size_t bytes,
-                                const char *tag, int tag_len,
+static void cb_azure_blob_flush(struct flb_event_chunk *event_chunk,
+                                struct flb_output_flush *out_flush,
                                 struct flb_input_instance *i_ins,
                                 void *out_context,
                                 struct flb_config *config)
@@ -471,13 +482,19 @@ static void cb_azure_blob_flush(const void *data, size_t bytes,
         FLB_OUTPUT_RETURN(FLB_RETRY);
     }
 
-    ret = send_blob(config, i_ins, ctx, (char *) tag,
-                    (char *) tag, tag_len, (char *) data, bytes);
+    ret = send_blob(config, i_ins, ctx,
+                    (char *) event_chunk->tag,  /* use tag as 'name' */
+                    (char *) event_chunk->tag, flb_sds_len(event_chunk->tag),
+                    (char *) event_chunk->data, event_chunk->size);
+
     if (ret == CREATE_BLOB) {
-        ret = create_blob(ctx, (char *) tag);
+        ret = create_blob(ctx, event_chunk->tag);
         if (ret == FLB_OK) {
-            ret = send_blob(config, i_ins, ctx, (char *) tag,
-                            (char *) tag, tag_len, (char *) data, bytes);
+            ret = send_blob(config, i_ins, ctx,
+                            (char *) event_chunk->tag,  /* use tag as 'name' */
+                            (char *) event_chunk->tag,
+                            flb_sds_len(event_chunk->tag),
+                            (char *) event_chunk->data, event_chunk->size);
         }
     }
 
@@ -566,6 +583,18 @@ static struct flb_config_map config_map[] = {
      "Name of the key that will have the record timestamp"
     },
 
+    {
+     FLB_CONFIG_MAP_STR, "auth_type", "key",
+     0, FLB_TRUE, offsetof(struct flb_azure_blob, auth_type),
+     "Set the auth type: key or sas"
+    },
+
+    {
+     FLB_CONFIG_MAP_STR, "sas_token", NULL,
+     0, FLB_TRUE, offsetof(struct flb_azure_blob, sas_token),
+     "Azure Blob SAS token"
+    },
+
     /* EOF */
     {0}
 };
