@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -94,65 +93,6 @@ static void append_labels(struct flb_calyptia *ctx, struct cmt *cmt)
     }
 }
 
-static flb_sds_t sha256_to_hex(unsigned char *sha256)
-{
-    int i;
-    flb_sds_t hex;
-    flb_sds_t tmp;
-
-    hex = flb_sds_create_size(64);
-    if (!hex) {
-        return NULL;
-    }
-
-    for (i = 0; i < 32; i++) {
-        tmp = flb_sds_printf(&hex, "%02x", sha256[i]);
-        if (!tmp) {
-            flb_sds_destroy(hex);
-            return NULL;
-        }
-        hex = tmp;
-    }
-
-    flb_sds_len_set(hex, 64);
-    return hex;
-}
-
-static int get_machine_id(struct flb_calyptia *ctx, char **out_buf, size_t *out_size)
-{
-    int ret;
-    char *buf;
-    flb_sds_t s_buf;
-    size_t s;
-    unsigned char sha256_buf[64] = {0};
-    mbedtls_sha256_context sha256_ctx;
-
-    /* retrieve raw machine id */
-    ret = flb_utils_get_machine_id(&buf, &s);
-    if (ret == -1) {
-        flb_plg_error(ctx->ins, "could not obtain machine id");
-        return -1;
-    }
-
-    /* perform sha256 */
-    mbedtls_sha256_init(&sha256_ctx);
-    mbedtls_sha256_starts(&sha256_ctx, 0);
-    mbedtls_sha256_update(&sha256_ctx, (const unsigned char *) buf, s);
-    mbedtls_sha256_finish(&sha256_ctx, sha256_buf);
-    flb_free(buf);
-
-    /* convert to hex */
-    s_buf = sha256_to_hex(sha256_buf);
-    if (!s_buf) {
-        return -1;
-    }
-
-    *out_buf = s_buf;
-    *out_size = flb_sds_len(s_buf);
-
-    return 0;
-}
-
 static void pack_str(msgpack_packer *mp_pck, char *str)
 {
     int len;
@@ -307,6 +247,16 @@ static flb_sds_t get_agent_metadata(struct flb_calyptia *ctx)
     msgpack_pack_str(&mp_pck, len);
     msgpack_pack_str_body(&mp_pck, ctx->machine_id, len);
 
+    /* fleetID */
+    if (ctx->fleet_id) {
+        flb_mp_map_header_append(&mh);
+        msgpack_pack_str(&mp_pck, 7);
+        msgpack_pack_str_body(&mp_pck, "fleetID", 7);
+        len = flb_sds_len(ctx->fleet_id);
+        msgpack_pack_str(&mp_pck, len);
+        msgpack_pack_str_body(&mp_pck, ctx->fleet_id, len);
+    }
+
     /* pack environment metadata */
     pack_env_metadata(config->env, &mh, &mp_pck);
 
@@ -357,6 +307,18 @@ static int calyptia_http_do(struct flb_calyptia *ctx, struct flb_http_client *c,
                             sizeof(CALYPTIA_H_AGENT_TOKEN) - 1,
                             ctx->agent_token, flb_sds_len(ctx->agent_token));
     }
+#ifdef FLB_HAVE_CHUNK_TRACE
+    else if (type == CALYPTIA_ACTION_TRACE)  {
+        flb_http_add_header(c,
+                            CALYPTIA_H_CTYPE, sizeof(CALYPTIA_H_CTYPE) - 1,
+                            CALYPTIA_H_CTYPE_JSON, sizeof(CALYPTIA_H_CTYPE_JSON) - 1);
+
+        flb_http_add_header(c,
+                            CALYPTIA_H_AGENT_TOKEN,
+                            sizeof(CALYPTIA_H_AGENT_TOKEN) - 1,
+                            ctx->agent_token, flb_sds_len(ctx->agent_token));
+    }
+#endif
 
     /* Map debug callbacks */
     flb_http_client_debug(c, ctx->ins->callback);
@@ -404,7 +366,7 @@ static flb_sds_t get_agent_info(char *buf, size_t size, char *k)
 
     len = strlen(k);
 
-    ret = flb_pack_json(buf, size, &out_buf, &out_size, &type);
+    ret = flb_pack_json(buf, size, &out_buf, &out_size, &type, NULL);
     if (ret != 0) {
         return NULL;
     }
@@ -473,7 +435,7 @@ static int store_session_set(struct flb_calyptia *ctx, char *buf, size_t size)
                              FLB_VERSION_STR "\n", sizeof(FLB_VERSION_STR) - 1);
 
     /* encode */
-    ret = flb_pack_json(buf, size, &mp_buf, &mp_size, &type);
+    ret = flb_pack_json(buf, size, &mp_buf, &mp_size, &type, NULL);
     if (ret < 0) {
         flb_plg_error(ctx->ins, "could not encode session information");
         return -1;
@@ -589,7 +551,7 @@ static int api_agent_create(struct flb_config *config, struct flb_calyptia *ctx)
     char uri[1024];
     flb_sds_t meta;
     struct flb_upstream *u;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
     struct flb_http_client *c;
 
     /* Meta */
@@ -612,7 +574,7 @@ static int api_agent_create(struct flb_config *config, struct flb_calyptia *ctx)
     }
 
     /* Make it synchronous */
-    u->flags &= ~(FLB_IO_ASYNC);
+    flb_stream_disable_async_mode(&u->base);
 
     /* Get upstream connection */
     u_conn = flb_upstream_conn_get(u);
@@ -694,8 +656,6 @@ static struct flb_calyptia *config_init(struct flb_output_instance *ins,
 {
     int ret;
     int flags;
-    size_t size;
-    char *machine_id;
     struct flb_calyptia *ctx;
 
     /* Calyptia plugin context */
@@ -742,14 +702,10 @@ static struct flb_calyptia *config_init(struct flb_output_instance *ins,
         }
     }
 
-    /* If no machine_id has been provided via a configuration option get it from the local machine-id. */
+    /* the machine-id is provided by custom calyptia, which invokes this plugin. */
     if (!ctx->machine_id) {
-        /* machine id */
-        ret = get_machine_id(ctx, &machine_id, &size);
-        if (ret == -1) {
-            return NULL;
-        }
-        ctx->machine_id = (flb_sds_t) machine_id;
+        flb_plg_error(ctx->ins, "machine_id has not been set");
+        return NULL;
     }
 
     flb_plg_debug(ctx->ins, "machine_id=%s", ctx->machine_id);
@@ -800,6 +756,12 @@ static int cb_calyptia_init(struct flb_output_instance *ins,
     ctx->metrics_endpoint = flb_sds_create_size(256);
     flb_sds_printf(&ctx->metrics_endpoint, CALYPTIA_ENDPOINT_METRICS,
                    ctx->agent_id);
+
+#ifdef FLB_HAVE_CHUNK_TRACE
+    ctx->trace_endpoint = flb_sds_create_size(256);
+    flb_sds_printf(&ctx->trace_endpoint, CALYPTIA_ENDPOINT_TRACE, 
+                   ctx->pipeline_id);
+#endif /* FLB_HAVE_CHUNK_TRACE */
     return 0;
 }
 
@@ -808,7 +770,7 @@ static void debug_payload(struct flb_calyptia *ctx, void *data, size_t bytes)
     int ret;
     size_t off = 0;
     struct cmt *cmt;
-    cmt_sds_t out;
+    cfl_sds_t out;
 
     ret = cmt_decode_msgpack_create(&cmt, (char *) data, bytes, &off);
     if (ret != CMT_DECODE_MSGPACK_SUCCESS) {
@@ -822,8 +784,8 @@ static void debug_payload(struct flb_calyptia *ctx, void *data, size_t bytes)
     cmt_destroy(cmt);
 }
 
-static void cb_calyptia_flush(const void *data, size_t bytes,
-                              const char *tag, int tag_len,
+static void cb_calyptia_flush(struct flb_event_chunk *event_chunk,
+                              struct flb_output_flush *out_flush,
                               struct flb_input_instance *i_ins,
                               void *out_context,
                               struct flb_config *config)
@@ -832,7 +794,13 @@ static void cb_calyptia_flush(const void *data, size_t bytes,
     size_t off = 0;
     size_t out_size = 0;
     char *out_buf = NULL;
-    struct flb_upstream_conn *u_conn;
+
+/* used to create records for reporting traces to the cloud. */
+#ifdef FLB_HAVE_CHUNK_TRACE
+    flb_sds_t json;
+#endif /* FLB_HAVE_CHUNK_TRACE */
+
+    struct flb_connection *u_conn;
     struct flb_http_client *c;
     struct flb_calyptia *ctx = out_context;
     struct cmt *cmt;
@@ -845,55 +813,102 @@ static void cb_calyptia_flush(const void *data, size_t bytes,
         FLB_OUTPUT_RETURN(FLB_RETRY);
     }
 
-    /* if we have labels append them */
-    if (ctx->add_labels && mk_list_size(ctx->add_labels) > 0) {
-        ret = cmt_decode_msgpack_create(&cmt, (char *) data, bytes, &off);
-        if (ret != CMT_DECODE_MSGPACK_SUCCESS) {
-            flb_upstream_conn_release(u_conn);
-            FLB_OUTPUT_RETURN(FLB_ERROR);
-        }
+    if (event_chunk->type == FLB_EVENT_TYPE_METRICS) {
+        /* if we have labels append them */
+        if (ctx->add_labels && mk_list_size(ctx->add_labels) > 0) {
+            ret = cmt_decode_msgpack_create(&cmt,
+                                            (char *) event_chunk->data,
+                                            event_chunk->size,
+                                            &off);
+            if (ret != CMT_DECODE_MSGPACK_SUCCESS) {
+                flb_upstream_conn_release(u_conn);
+                FLB_OUTPUT_RETURN(FLB_ERROR);
+            }
 
-        /* append labels set by config */
-        append_labels(ctx, cmt);
+            /* append labels set by config */
+            append_labels(ctx, cmt);
 
-        /* encode back to msgpack */
-        ret = cmt_encode_msgpack_create(cmt, &out_buf, &out_size);
-        if (ret != 0) {
+            /* encode back to msgpack */
+            ret = cmt_encode_msgpack_create(cmt, &out_buf, &out_size);
+            if (ret != 0) {
+                cmt_destroy(cmt);
+                flb_upstream_conn_release(u_conn);
+                FLB_OUTPUT_RETURN(FLB_ERROR);
+            }
             cmt_destroy(cmt);
-            flb_upstream_conn_release(u_conn);
-            FLB_OUTPUT_RETURN(FLB_ERROR);
         }
-        cmt_destroy(cmt);
-    }
-    else {
-        out_buf = (char *) data;
-        out_size = bytes;
-    }
+        else {
+            out_buf = (char *) event_chunk->data;
+            out_size = event_chunk->size;
+        }
 
-    /* Compose HTTP Client request */
-    c = flb_http_client(u_conn, FLB_HTTP_POST, ctx->metrics_endpoint,
-                        out_buf, out_size, NULL, 0, NULL, 0);
-    if (!c) {
-        if (out_buf != data) {
+        /* Compose HTTP Client request */
+        c = flb_http_client(u_conn, FLB_HTTP_POST, ctx->metrics_endpoint,
+                            out_buf, out_size, NULL, 0, NULL, 0);
+        if (!c) {
+            if (out_buf != event_chunk->data) {
+                cmt_encode_msgpack_destroy(out_buf);
+            }
+            flb_upstream_conn_release(u_conn);
+            FLB_OUTPUT_RETURN(FLB_RETRY);
+        }
+
+        /* perform request: 'ret' might be FLB_OK, FLB_ERROR or FLB_RETRY */
+        ret = calyptia_http_do(ctx, c, CALYPTIA_ACTION_METRICS);
+        if (ret == FLB_OK) {
+            flb_plg_debug(ctx->ins, "metrics delivered OK");
+        }
+        else if (ret == FLB_ERROR) {
+            flb_plg_error(ctx->ins, "could not deliver metrics");
+            debug_payload(ctx, out_buf, out_size);
+        }
+
+        if (out_buf != event_chunk->data) {
             cmt_encode_msgpack_destroy(out_buf);
         }
-        flb_upstream_conn_release(u_conn);
-        FLB_OUTPUT_RETURN(FLB_RETRY);
     }
+    
+#ifdef FLB_HAVE_CHUNK_TRACE
+    if (event_chunk->type == (FLB_EVENT_TYPE_LOGS | FLB_EVENT_TYPE_HAS_TRACE)) {
+        json = flb_pack_msgpack_to_json_format(event_chunk->data,
+                                               event_chunk->size,
+                                               FLB_PACK_JSON_FORMAT_STREAM,
+                                               FLB_PACK_JSON_DATE_DOUBLE,
+                                               NULL);
+        if (json == NULL) {
+            flb_upstream_conn_release(u_conn);
+            FLB_OUTPUT_RETURN(FLB_RETRY);
+        }
+        out_buf = (char *)json;
+        out_size = flb_sds_len(json);
 
-    /* perform request: 'ret' might be FLB_OK, FLB_ERROR or FLB_RETRY */
-    ret = calyptia_http_do(ctx, c, CALYPTIA_ACTION_METRICS);
-    if (ret == FLB_OK) {
-        flb_plg_debug(ctx->ins, "metrics delivered OK");
-    }
-    else if (ret == FLB_ERROR) {
-        flb_plg_error(ctx->ins, "could not deliver metrics");
-        debug_payload(ctx, out_buf, out_size);
-    }
-
-    if (out_buf != data) {
-        cmt_encode_msgpack_destroy(out_buf);
+        if (flb_sds_printf(&ctx->metrics_endpoint, CALYPTIA_ENDPOINT_METRICS,
+                       ctx->agent_id) == NULL) {
+            flb_upstream_conn_release(u_conn);
+            flb_sds_destroy(json);
+            FLB_OUTPUT_RETURN(FLB_RETRY);
+        }
+        c = flb_http_client(u_conn, FLB_HTTP_POST, ctx->trace_endpoint,
+                            out_buf, out_size, NULL, 0, NULL, 0);
+        if (!c) {
+            flb_upstream_conn_release(u_conn);
+            flb_sds_destroy(json);
+            flb_sds_destroy(ctx->metrics_endpoint);
+            FLB_OUTPUT_RETURN(FLB_RETRY);
+        }
+        
+        /* perform request: 'ret' might be FLB_OK, FLB_ERROR or FLB_RETRY */
+        ret = calyptia_http_do(ctx, c, CALYPTIA_ACTION_TRACE);
+        if (ret == FLB_OK) {
+            flb_plg_debug(ctx->ins, "trace delivered OK");
+        }
+        else if (ret == FLB_ERROR) {
+            flb_plg_error(ctx->ins, "could not deliver trace");
+            debug_payload(ctx, out_buf, out_size);
+        }
+        flb_sds_destroy(json);
     }
+#endif /* FLB_HAVE_CHUNK_TRACE */
 
     flb_upstream_conn_release(u_conn);
     flb_http_client_destroy(c);
@@ -920,10 +935,6 @@ static int cb_calyptia_exit(void *data, struct flb_config *config)
         flb_sds_destroy(ctx->agent_token);
     }
 
-    if (ctx->machine_id) {
-        flb_sds_destroy(ctx->machine_id);
-    }
-
     if (ctx->env) {
         flb_env_destroy(ctx->env);
     }
@@ -932,6 +943,12 @@ static int cb_calyptia_exit(void *data, struct flb_config *config)
         flb_sds_destroy(ctx->metrics_endpoint);
     }
 
+#ifdef FLB_HAVE_CHUNK_TRACE
+    if (ctx->trace_endpoint) {
+        flb_sds_destroy(ctx->trace_endpoint);
+    }
+#endif /* FLB_HAVE_CHUNK_TRACE */
+
     if (ctx->fs) {
         flb_fstore_destroy(ctx->fs);
     }
@@ -966,6 +983,11 @@ static struct flb_config_map config_map[] = {
      0, FLB_TRUE, offsetof(struct flb_calyptia, machine_id),
      "Custom machine_id to be used when registering agent"
     },
+    {
+     FLB_CONFIG_MAP_STR, "fleet_id", NULL,
+     0, FLB_TRUE, offsetof(struct flb_calyptia, fleet_id),
+     "Fleet ID for identifying as part of a managed fleet"
+    },
 
     {
      FLB_CONFIG_MAP_STR, "store_path", NULL,
@@ -979,6 +1001,14 @@ static struct flb_config_map config_map[] = {
      "Label to append to the generated metric."
     },
 
+#ifdef FLB_HAVE_CHUNK_TRACE
+    {
+     FLB_CONFIG_MAP_STR, "pipeline_id", NULL,
+     0, FLB_TRUE, offsetof(struct flb_calyptia, pipeline_id),
+     "Pipeline ID for calyptia core traces."
+    },
+#endif
+
     /* EOF */
     {0}
 };
