@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -25,6 +24,9 @@
 #include <fluent-bit/flb_pack.h>
 #include <fluent-bit/flb_time.h>
 #include <fluent-bit/flb_gzip.h>
+#include <fluent-bit/flb_config_map.h>
+#include <fluent-bit/flb_version.h>
+#include <fluent-bit/flb_log_event_decoder.h>
 
 #include <msgpack.h>
 
@@ -85,37 +87,55 @@ static int datadog_format(struct flb_config *config,
                           struct flb_input_instance *ins,
                           void *plugin_context,
                           void *flush_ctx,
+                          int event_type,
                           const char *tag, int tag_len,
                           const void *data, size_t bytes,
                           void **out_data, size_t *out_size)
 {
     int i;
     int ind;
-    int byte_cnt;
+    int byte_cnt = 64;
     int remap_cnt;
+    int ret;
     /* for msgpack global structs */
-    int array_size = 0;
-    size_t off = 0;
-    msgpack_unpacked result;
+    size_t array_size = 0;
     msgpack_sbuffer mp_sbuf;
     msgpack_packer mp_pck;
     /* for sub msgpack objs */
     int map_size;
-    struct flb_time tms;
     int64_t timestamp;
-    msgpack_object *obj;
     msgpack_object map;
-    msgpack_object root;
     msgpack_object k;
     msgpack_object v;
     struct flb_out_datadog *ctx = plugin_context;
+    struct flb_event_chunk *event_chunk;
 
     /* output buffer */
     flb_sds_t out_buf;
     flb_sds_t remapped_tags = NULL;
+    flb_sds_t tmp = NULL;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
 
-    /* Count number of records */
-    array_size = flb_mp_count(data, bytes);
+    /* in normal flush callback we have the event_chunk set as flush context
+     * so we don't need to calculate the event len.
+     * But in test mode the formatter won't get the event_chunk as flush_ctx
+     */ 
+    if (flush_ctx != NULL) {
+        event_chunk = flush_ctx; 
+        array_size = event_chunk->total_events;
+    } else {
+        array_size = flb_mp_count(data, bytes);
+    }
+
+    ret = flb_log_event_decoder_init(&log_decoder, (char *) data, bytes);
+
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
+
+        return -1;
+    }
 
     /* Create temporary msgpack buffer */
     msgpack_sbuffer_init(&mp_sbuf);
@@ -124,16 +144,12 @@ static int datadog_format(struct flb_config *config,
     /* Prepare array for all entries */
     msgpack_pack_array(&mp_pck, array_size);
 
-    off = 0;
-    msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
-        root = result.data;
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
+        timestamp = timestamp_format(&log_event.timestamp);
 
-        /* Get timestamp and object */
-        flb_time_pop_from_msgpack(&tms, &result, &obj);
-        timestamp = timestamp_format(&tms);
-
-        map = root.via.array.ptr[1];
+        map = *log_event.body;
         map_size = map.via.map.size;
 
         /*
@@ -160,6 +176,23 @@ static int datadog_format(struct flb_config *config,
 
             if (!remapped_tags) {
                 remapped_tags = flb_sds_create_size(byte_cnt);
+                if (!remapped_tags) {
+                    flb_errno();
+                    msgpack_sbuffer_destroy(&mp_sbuf);
+                    flb_log_event_decoder_destroy(&log_decoder);
+                    return -1;
+                }
+            }
+            else if (flb_sds_len(remapped_tags) < byte_cnt) {
+                tmp = flb_sds_increase(remapped_tags, byte_cnt - flb_sds_len(remapped_tags));
+                if (!tmp) {
+                    flb_errno();
+                    flb_sds_destroy(remapped_tags);
+                    msgpack_sbuffer_destroy(&mp_sbuf);
+                    flb_log_event_decoder_destroy(&log_decoder);
+                    return -1;
+                }
+                remapped_tags = tmp;
             }
 
             /*
@@ -226,8 +259,11 @@ static int datadog_format(struct flb_config *config,
              * (so they won't be packed as attr)
              */
             if (ctx->remap && (ind = dd_attr_need_remapping(k, v)) >=0 ) {
-                remapping[ind].remap_to_tag(remapping[ind].remap_tag_name, v,
-                                            remapped_tags);
+                ret = remapping[ind].remap_to_tag(remapping[ind].remap_tag_name, v,
+                                                  &remapped_tags);
+                if (ret < 0) {
+                    flb_plg_error(ctx->ins, "Failed to remap tag: %s, skipping", remapping[ind].remap_tag_name);
+                }
                 continue;
             }
 
@@ -249,9 +285,25 @@ static int datadog_format(struct flb_config *config,
         /* here we concatenate ctx->dd_tags and remapped_tags, depending on their presence */
         if (remap_cnt) {
             if (ctx->dd_tags != NULL) {
-                flb_sds_cat(remapped_tags, FLB_DATADOG_TAG_SEPERATOR,
-                            strlen(FLB_DATADOG_TAG_SEPERATOR));
+                tmp = flb_sds_cat(remapped_tags, FLB_DATADOG_TAG_SEPERATOR,
+                                  strlen(FLB_DATADOG_TAG_SEPERATOR));
+                if (!tmp) {
+                    flb_errno();
+                    flb_sds_destroy(remapped_tags);
+                    msgpack_sbuffer_destroy(&mp_sbuf);
+                    flb_log_event_decoder_destroy(&log_decoder);
+                    return -1;
+                }
+                remapped_tags = tmp;
                 flb_sds_cat(remapped_tags, ctx->dd_tags, strlen(ctx->dd_tags));
+                if (!tmp) {
+                    flb_errno();
+                    flb_sds_destroy(remapped_tags);
+                    msgpack_sbuffer_destroy(&mp_sbuf);
+                    flb_log_event_decoder_destroy(&log_decoder);
+                    return -1;
+                }
+                remapped_tags = tmp;
             }
             dd_msgpack_pack_key_value_str(&mp_pck,
                                           FLB_DATADOG_DD_TAGS_KEY,
@@ -275,7 +327,7 @@ static int datadog_format(struct flb_config *config,
         if (remapped_tags) {
             flb_sds_destroy(remapped_tags);
         }
-        msgpack_unpacked_destroy(&result);
+        flb_log_event_decoder_destroy(&log_decoder);
         return -1;
     }
 
@@ -283,7 +335,8 @@ static int datadog_format(struct flb_config *config,
     *out_size = flb_sds_len(out_buf);
 
     /* Cleanup */
-    msgpack_unpacked_destroy(&result);
+    flb_log_event_decoder_destroy(&log_decoder);
+
     if (remapped_tags) {
         flb_sds_destroy(remapped_tags);
     }
@@ -291,14 +344,14 @@ static int datadog_format(struct flb_config *config,
     return 0;
 }
 
-static void cb_datadog_flush(const void *data, size_t bytes,
-                             const char *tag, int tag_len,
+static void cb_datadog_flush(struct flb_event_chunk *event_chunk,
+                             struct flb_output_flush *out_flush,
                              struct flb_input_instance *i_ins,
                              void *out_context,
                              struct flb_config *config)
 {
     struct flb_out_datadog *ctx = out_context;
-    struct flb_upstream_conn *upstream_conn;
+    struct flb_connection *upstream_conn;
     struct flb_http_client *client;
     void *out_buf;
     size_t out_size;
@@ -319,8 +372,9 @@ static void cb_datadog_flush(const void *data, size_t bytes,
     /* Convert input data into a Datadog JSON payload */
     ret = datadog_format(config, i_ins,
                          ctx, NULL,
-                         tag, tag_len,
-                         data, bytes,
+                         event_chunk->type,
+                         event_chunk->tag, flb_sds_len(event_chunk->tag),
+                         event_chunk->data, event_chunk->size,
                          &out_buf, &out_size);
     if (ret == -1) {
         flb_upstream_conn_release(upstream_conn);
@@ -354,7 +408,11 @@ static void cb_datadog_flush(const void *data, size_t bytes,
         FLB_OUTPUT_RETURN(FLB_ERROR);
     }
 
+    /* Add the required headers to the URI */
     flb_http_add_header(client, "User-Agent", 10, "Fluent-Bit", 10);
+    flb_http_add_header(client, FLB_DATADOG_API_HDR, sizeof(FLB_DATADOG_API_HDR) - 1, ctx->api_key, flb_sds_len(ctx->api_key));
+    flb_http_add_header(client, FLB_DATADOG_ORIGIN_HDR, sizeof(FLB_DATADOG_ORIGIN_HDR) - 1, "Fluent-Bit", 10);
+    flb_http_add_header(client, FLB_DATADOG_ORIGIN_VERSION_HDR, sizeof(FLB_DATADOG_ORIGIN_VERSION_HDR) - 1, FLB_VERSION_STR, sizeof(FLB_VERSION_STR) - 1);
     flb_http_add_header(client,
                         FLB_DATADOG_CONTENT_TYPE, sizeof(FLB_DATADOG_CONTENT_TYPE) - 1,
                         FLB_DATADOG_MIME_JSON, sizeof(FLB_DATADOG_MIME_JSON) - 1);
@@ -376,12 +434,12 @@ static void cb_datadog_flush(const void *data, size_t bytes,
         }
         else {
             if (client->resp.payload) {
-                flb_plg_info(ctx->ins, "%s%s, port=%i, HTTP status=%i payload=%s",
+                flb_plg_debug(ctx->ins, "%s%s, port=%i, HTTP status=%i payload=%s",
                              ctx->scheme, ctx->host, ctx->port,
                              client->resp.status, client->resp.payload);
             }
             else {
-                flb_plg_info(ctx->ins, "%s%s, port=%i, HTTP status=%i",
+                flb_plg_debug(ctx->ins, "%s%s, port=%i, HTTP status=%i",
                              ctx->scheme, ctx->host, ctx->port,
                              client->resp.status);
             }
@@ -422,6 +480,76 @@ static int cb_datadog_exit(void *data, struct flb_config *config)
     return 0;
 }
 
+static struct flb_config_map config_map[] = {
+    {
+     FLB_CONFIG_MAP_STR, "compress", "false",
+     0, FLB_FALSE, 0,
+     "compresses the payload in GZIP format, "
+     "Datadog supports and recommends setting this to 'gzip'."
+    },
+    {
+     FLB_CONFIG_MAP_STR, "apikey", NULL,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, api_key),
+     "Datadog API key"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "dd_service", NULL,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, dd_service),
+     "The human readable name for your service generating the logs "
+     "- the name of your application or database."
+    },
+    {
+     FLB_CONFIG_MAP_STR, "dd_source", NULL,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, dd_source),
+     "A human readable name for the underlying technology of your service. "
+     "For example, 'postgres' or 'nginx'."
+    },
+    {
+     FLB_CONFIG_MAP_STR, "dd_tags", NULL,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, dd_tags),
+     "The tags you want to assign to your logs in Datadog."
+    },
+
+    {
+     FLB_CONFIG_MAP_STR, "proxy", NULL,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, proxy),
+     "Specify an HTTP Proxy. The expected format of this value is http://host:port. "
+     "Note that https is not supported yet."
+    },
+    {
+     FLB_CONFIG_MAP_BOOL, "include_tag_key", "false",
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, include_tag_key),
+     "If enabled, tag is appended to output. "
+     "The key name is used 'tag_key' property."
+    },
+    {
+     FLB_CONFIG_MAP_STR, "tag_key", FLB_DATADOG_DEFAULT_TAG_KEY,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, tag_key),
+     "The key name of tag. If 'include_tag_key' is false, "
+     "This property is ignored"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "dd_message_key", NULL,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, dd_message_key),
+     "By default, the plugin searches for the key 'log' "
+     "and remap the value to the key 'message'. "
+     "If the property is set, the plugin will search the property name key."
+    },
+    {
+     FLB_CONFIG_MAP_STR, "provider", NULL,
+     0, FLB_FALSE, 0,
+     "To activate the remapping, specify configuration flag provider with value 'ecs'"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "json_date_key", FLB_DATADOG_DEFAULT_TIME_KEY,
+     0, FLB_TRUE, offsetof(struct flb_out_datadog, json_date_key),
+     "Date key name for output."
+    },
+
+    /* EOF */
+    {0}
+};
+
 struct flb_output_plugin out_datadog_plugin = {
     .name         = "datadog",
     .description  = "Send events to DataDog HTTP Event Collector",
@@ -432,6 +560,9 @@ struct flb_output_plugin out_datadog_plugin = {
     /* Test */
     .test_formatter.callback = datadog_format,
 
+    /* Config map */
+    .config_map   = config_map,
+
     /* Plugin flags */
     .flags        = FLB_OUTPUT_NET | FLB_IO_OPT_TLS,
 };
