@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -24,6 +23,7 @@
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_time.h>
 #include <fluent-bit/flb_metrics.h>
+#include <fluent-bit/flb_log_event_decoder.h>
 #include <msgpack.h>
 
 #include <stdio.h>
@@ -31,10 +31,16 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 
+#ifdef FLB_SYSTEM_WINDOWS
+#include <Shlobj.h>
+#include <Shlwapi.h>
+#endif
+
 #include "file.h"
 
 #ifdef FLB_SYSTEM_WINDOWS
 #define NEWLINE "\r\n"
+#define S_ISDIR(m)      (((m) & S_IFMT) == S_IFDIR)
 #else
 #define NEWLINE "\n"
 #endif
@@ -47,6 +53,7 @@ struct flb_file_conf {
     const char *template;
     int format;
     int csv_column_names;
+    int mkdir;
     struct flb_output_instance *ins;
 };
 
@@ -123,6 +130,10 @@ static int cb_file_init(struct flb_output_instance *ins,
         else if (!strcasecmp(tmp, "template")) {
             ctx->format    = FLB_OUT_FILE_FMT_TEMPLATE;
         }
+        else if (!strcasecmp(tmp, "out_file")) {
+            /* for explicit setting */
+            ctx->format = FLB_OUT_FILE_FMT_JSON;
+        }
         else {
             flb_plg_error(ctx->ins, "unknown format %s. abort.", tmp);
             flb_free(ctx);
@@ -330,7 +341,7 @@ static void print_metrics_text(struct flb_output_instance *ins,
 {
     int ret;
     size_t off = 0;
-    cmt_sds_t text;
+    cfl_sds_t text;
     struct cmt *cmt = NULL;
 
     /* get cmetrics context */
@@ -350,8 +361,92 @@ static void print_metrics_text(struct flb_output_instance *ins,
     cmt_encode_text_destroy(text);
 }
 
-static void cb_file_flush(const void *data, size_t bytes,
-                          const char *tag, int tag_len,
+static int mkpath(struct flb_output_instance *ins, const char *dir)
+{
+    struct stat st;
+    char *dup_dir = NULL;
+#ifdef FLB_SYSTEM_MACOS
+    char *parent_dir = NULL;
+#endif
+
+    int ret;
+
+    if (!dir) {
+        errno = EINVAL;
+        return -1;
+    }
+
+    if (strlen(dir) == 0) {
+        errno = EINVAL;
+        return -1;
+    }
+
+    if (stat(dir, &st) == 0) {
+        if (S_ISDIR (st.st_mode)) {
+            return 0;
+        }
+        flb_plg_error(ins, "%s is not a directory", dir);
+        errno = ENOTDIR;
+        return -1;
+    }
+
+#ifdef FLB_SYSTEM_WINDOWS
+    char path[MAX_PATH];
+
+    if (_fullpath(path, dir, MAX_PATH) == NULL) {
+        return -1;
+    }
+
+    if (SHCreateDirectoryExA(NULL, path, NULL) != ERROR_SUCCESS) {
+        return -1;
+    }
+    return 0;
+#elif FLB_SYSTEM_MACOS
+    dup_dir = strdup(dir);
+    if (!dup_dir) {
+        return -1;
+    }
+
+    /* macOS's dirname(3) should return current directory when slash
+     * charachter is not included in passed string.
+     * And note that macOS's dirname(3) does not modify passed string.
+     */
+    parent_dir = dirname(dup_dir);
+    if (stat(parent_dir, &st) == 0 && strncmp(parent_dir, ".", 1)) {
+        if (S_ISDIR (st.st_mode)) {
+            flb_plg_debug(ins, "creating directory %s", dup_dir);
+            ret = mkdir(dup_dir, 0755);
+            free(dup_dir);
+            return ret;
+        }
+    }
+
+    ret = mkpath(ins, dirname(dup_dir));
+    if (ret != 0) {
+        free(dup_dir);
+        return ret;
+    }
+    flb_plg_debug(ins, "creating directory %s", dup_dir);
+    ret = mkdir(dup_dir, 0755);
+    free(dup_dir);
+    return ret;
+#else
+    dup_dir = strdup(dir);
+    if (!dup_dir) {
+        return -1;
+    }
+    ret = mkpath(ins, dirname(dup_dir));
+    free(dup_dir);
+    if (ret != 0) {
+        return ret;
+    }
+    flb_plg_debug(ins, "creating directory %s", dir);
+    return mkdir(dir, 0755);
+#endif
+}
+
+static void cb_file_flush(struct flb_event_chunk *event_chunk,
+                          struct flb_output_flush *out_flush,
                           struct flb_input_instance *ins,
                           void *out_context,
                           struct flb_config *config)
@@ -359,18 +454,18 @@ static void cb_file_flush(const void *data, size_t bytes,
     int ret;
     int column_names;
     FILE * fp;
-    msgpack_unpacked result;
     size_t off = 0;
     size_t last_off = 0;
     size_t alloc_size = 0;
     size_t total;
     char out_file[PATH_MAX];
     char *buf;
-    char *tag_buf;
     long file_pos;
-    msgpack_object *obj;
     struct flb_file_conf *ctx = out_context;
-    struct flb_time tm;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
+    char* out_file_copy;
+
     (void) config;
 
     /* Set the right output file */
@@ -381,7 +476,7 @@ static void cb_file_flush(const void *data, size_t bytes,
         }
         else {
             snprintf(out_file, PATH_MAX - 1, "%s/%s",
-                     ctx->out_path, tag);
+                     ctx->out_path, event_chunk->tag);
         }
     }
     else {
@@ -389,12 +484,27 @@ static void cb_file_flush(const void *data, size_t bytes,
             snprintf(out_file, PATH_MAX - 1, "%s", ctx->out_file);
         }
         else {
-            snprintf(out_file, PATH_MAX - 1, "%s", tag);
+            snprintf(out_file, PATH_MAX - 1, "%s", event_chunk->tag);
         }
     }
 
     /* Open output file with default name as the Tag */
     fp = fopen(out_file, "ab+");
+    if (ctx->mkdir == FLB_TRUE && fp == NULL && errno == ENOENT) {
+        out_file_copy = strdup(out_file);
+        if (out_file_copy) {
+#ifdef FLB_SYSTEM_WINDOWS
+            PathRemoveFileSpecA(out_file_copy);
+            ret = mkpath(ctx->ins, out_file_copy);
+#else
+            ret = mkpath(ctx->ins, dirname(out_file_copy));
+#endif
+            free(out_file_copy);
+            if (ret == 0) {
+                fp = fopen(out_file, "ab+");
+            }
+        }
+    }
     if (fp == NULL) {
         flb_errno();
         flb_plg_error(ctx->ins, "error opening: %s", out_file);
@@ -408,22 +518,13 @@ static void cb_file_flush(const void *data, size_t bytes,
     file_pos = ftell(fp);
 
     /* Check if the event type is metrics, handle the payload differently */
-    if (flb_input_event_type_is_metric(ins)) {
-        print_metrics_text(ctx->ins, fp, (char *) data, bytes);
+    if (event_chunk->type == FLB_INPUT_METRICS) {
+        print_metrics_text(ctx->ins, fp,
+                           event_chunk->data, event_chunk->size);
         fclose(fp);
         FLB_OUTPUT_RETURN(FLB_OK);
     }
 
-    tag_buf = flb_malloc(tag_len + 1);
-    if (!tag_buf) {
-        flb_errno();
-        fclose(fp);
-        FLB_OUTPUT_RETURN(FLB_RETRY);
-    }
-    memcpy(tag_buf, tag, tag_len);
-    tag_buf[tag_len] = '\0';
-
-
     /*
      * Msgpack output format used to create unit tests files, useful for
      * Fluent Bit developers.
@@ -433,46 +534,55 @@ static void cb_file_flush(const void *data, size_t bytes,
         total = 0;
 
         do {
-            ret = fwrite((char *)data + off, 1, bytes - off, fp);
+            ret = fwrite((char *) event_chunk->data + off, 1,
+                         event_chunk->size - off, fp);
             if (ret < 0) {
                 flb_errno();
                 fclose(fp);
-                flb_free(tag_buf);
                 FLB_OUTPUT_RETURN(FLB_RETRY);
             }
             total += ret;
-        } while (total < bytes);
+        } while (total < event_chunk->size);
 
         fclose(fp);
-        flb_free(tag_buf);
         FLB_OUTPUT_RETURN(FLB_OK);
     }
 
+    ret = flb_log_event_decoder_init(&log_decoder,
+                                     (char *) event_chunk->data,
+                                     event_chunk->size);
+
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
+
+        fclose(fp);
+        FLB_OUTPUT_RETURN(FLB_ERROR);
+    }
+
     /*
      * Upon flush, for each array, lookup the time and the first field
      * of the map to use as a data point.
      */
-    msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
         alloc_size = (off - last_off) + 128; /* JSON is larger than msgpack */
         last_off = off;
 
-        flb_time_pop_from_msgpack(&tm, &result, &obj);
-
         switch (ctx->format){
         case FLB_OUT_FILE_FMT_JSON:
-            buf = flb_msgpack_to_json_str(alloc_size, obj);
+            buf = flb_msgpack_to_json_str(alloc_size, log_event.body);
             if (buf) {
                 fprintf(fp, "%s: [%"PRIu64".%09lu, %s]" NEWLINE,
-                        tag_buf,
-                        tm.tm.tv_sec, tm.tm.tv_nsec,
+                        event_chunk->tag,
+                        log_event.timestamp.tm.tv_sec, log_event.timestamp.tm.tv_nsec,
                         buf);
                 flb_free(buf);
             }
             else {
-                msgpack_unpacked_destroy(&result);
+                flb_log_event_decoder_destroy(&log_decoder);
                 fclose(fp);
-                flb_free(tag_buf);
                 FLB_OUTPUT_RETURN(FLB_RETRY);
             }
             break;
@@ -484,22 +594,30 @@ static void cb_file_flush(const void *data, size_t bytes,
             else {
                 column_names = FLB_FALSE;
             }
-            csv_output(fp, column_names, &tm, obj, ctx);
+            csv_output(fp, column_names,
+                       &log_event.timestamp,
+                       log_event.body, ctx);
             break;
         case FLB_OUT_FILE_FMT_LTSV:
-            ltsv_output(fp, &tm, obj, ctx);
+            ltsv_output(fp,
+                        &log_event.timestamp,
+                        log_event.body, ctx);
             break;
         case FLB_OUT_FILE_FMT_PLAIN:
-            plain_output(fp, obj, alloc_size);
+            plain_output(fp, log_event.body, alloc_size);
+
             break;
         case FLB_OUT_FILE_FMT_TEMPLATE:
-            template_output(fp, &tm, obj, ctx);
+            template_output(fp,
+                            &log_event.timestamp,
+                            log_event.body, ctx);
+
             break;
         }
     }
 
-    flb_free(tag_buf);
-    msgpack_unpacked_destroy(&result);
+    flb_log_event_decoder_destroy(&log_decoder);
+
     fclose(fp);
 
     FLB_OUTPUT_RETURN(FLB_OK);
@@ -564,6 +682,12 @@ static struct flb_config_map config_map[] = {
      "Add column names (keys) in the first line of the target file"
     },
 
+    {
+     FLB_CONFIG_MAP_BOOL, "mkdir", "false",
+     0, FLB_TRUE, offsetof(struct flb_file_conf, mkdir),
+     "Recursively create output directory if it does not exist. Permissions set to 0755"
+    },
+
     /* EOF */
     {0}
 };
@@ -575,6 +699,7 @@ struct flb_output_plugin out_file_plugin = {
     .cb_flush     = cb_file_flush,
     .cb_exit      = cb_file_exit,
     .flags        = 0,
+    .workers      = 1,
     .event_type   = FLB_OUTPUT_LOGS | FLB_OUTPUT_METRICS,
     .config_map   = config_map,
 };
