@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -25,6 +25,8 @@
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_network.h>
 #include <fluent-bit/flb_random.h>
+#include <fluent-bit/flb_config_map.h>
+#include <fluent-bit/flb_log_event_decoder.h>
 #include <msgpack.h>
 
 #include <stdio.h>
@@ -226,8 +228,8 @@ static int gelf_send_udp(struct flb_out_gelf_config *ctx, char *msg,
     return 0;
 }
 
-static void cb_gelf_flush(const void *data, size_t bytes,
-                          const char *tag, int tag_len,
+static void cb_gelf_flush(struct flb_event_chunk *event_chunk,
+                          struct flb_output_flush *out_flush,
                           struct flb_input_instance *i_ins,
                           void *out_context,
                           struct flb_config *config)
@@ -235,17 +237,15 @@ static void cb_gelf_flush(const void *data, size_t bytes,
     int ret;
     flb_sds_t s;
     flb_sds_t tmp;
-    msgpack_unpacked result;
     size_t off = 0;
     size_t prev_off = 0;
     size_t size = 0;
     size_t bytes_sent;
-    msgpack_object root;
     msgpack_object map;
-    msgpack_object *obj;
-    struct flb_time tm;
-    struct flb_upstream_conn *u_conn = NULL;
+    struct flb_connection *u_conn = NULL;
     struct flb_out_gelf_config *ctx = out_context;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
 
     if (ctx->mode != FLB_GELF_UDP) {
         u_conn = flb_upstream_conn_get(ctx->u);
@@ -255,38 +255,52 @@ static void cb_gelf_flush(const void *data, size_t bytes,
         }
     }
 
-    msgpack_unpacked_init(&result);
+    ret = flb_log_event_decoder_init(&log_decoder,
+                                     (char *) event_chunk->data,
+                                     event_chunk->size);
 
-    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
+
+        if (ctx->mode != FLB_GELF_UDP) {
+            flb_upstream_conn_release(u_conn);
+        }
+
+        FLB_OUTPUT_RETURN(FLB_RETRY);
+    }
+
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
+        off = log_decoder.offset;
         size = off - prev_off;
         prev_off = off;
-        if (result.data.type != MSGPACK_OBJECT_ARRAY) {
-            continue;
-        }
 
-        root = result.data;
-        if (root.via.array.size != 2) {
-            continue;
-        }
-
-        flb_time_pop_from_msgpack(&tm, &result, &obj);
-        map = root.via.array.ptr[1];
+        map = *log_event.body;
 
         size = (size * 1.4);
         s = flb_sds_create_size(size);
         if (s == NULL) {
-            msgpack_unpacked_destroy(&result);
+            flb_log_event_decoder_destroy(&log_decoder);
             FLB_OUTPUT_RETURN(FLB_ERROR);
         }
 
-        tmp = flb_msgpack_to_gelf(&s, &map, &tm, &(ctx->fields));
+        tmp = flb_msgpack_to_gelf(&s, &map, &log_event.timestamp,
+                                  &(ctx->fields));
         if (tmp != NULL) {
             s = tmp;
             if (ctx->mode == FLB_GELF_UDP) {
                 ret = gelf_send_udp(ctx, s, flb_sds_len(s));
                 if (ret == -1) {
-                    msgpack_unpacked_destroy(&result);
+                    if (ctx->mode != FLB_GELF_UDP) {
+                        flb_upstream_conn_release(u_conn);
+                    }
+
+                    flb_log_event_decoder_destroy(&log_decoder);
+
                     flb_sds_destroy(s);
+
                     FLB_OUTPUT_RETURN(FLB_RETRY);
                 }
             }
@@ -296,9 +310,15 @@ static void cb_gelf_flush(const void *data, size_t bytes,
                                        s, flb_sds_len(s) + 1, &bytes_sent);
                 if (ret == -1) {
                     flb_errno();
-                    flb_upstream_conn_release(u_conn);
-                    msgpack_unpacked_destroy(&result);
+
+                    if (ctx->mode != FLB_GELF_UDP) {
+                        flb_upstream_conn_release(u_conn);
+                    }
+
+                    flb_log_event_decoder_destroy(&log_decoder);
+
                     flb_sds_destroy(s);
+
                     FLB_OUTPUT_RETURN(FLB_RETRY);
                 }
             }
@@ -310,7 +330,7 @@ static void cb_gelf_flush(const void *data, size_t bytes,
         flb_sds_destroy(s);
     }
 
-    msgpack_unpacked_destroy(&result);
+    flb_log_event_decoder_destroy(&log_decoder);
 
     if (ctx->mode != FLB_GELF_UDP) {
         flb_upstream_conn_release(u_conn);
@@ -322,6 +342,7 @@ static void cb_gelf_flush(const void *data, size_t bytes,
 static int cb_gelf_init(struct flb_output_instance *ins, struct flb_config *config,
                         void *data)
 {
+    int ret;
     const char *tmp;
     struct flb_out_gelf_config *ctx = NULL;
 
@@ -336,6 +357,13 @@ static int cb_gelf_init(struct flb_output_instance *ins, struct flb_config *conf
     }
     ctx->ins = ins;
 
+    ret = flb_output_config_map_set(ins, (void *) ctx);
+    if (ret == -1) {
+        flb_plg_error(ins, "flb_output_config_map_set failed");
+        flb_free(ctx);
+        return -1;
+    }
+
     /* Config Mode */
     tmp = flb_output_get_property("mode", ins);
     if (tmp) {
@@ -388,24 +416,6 @@ static int cb_gelf_init(struct flb_output_instance *ins, struct flb_config *conf
         ctx->fields.level_key = flb_sds_create(tmp);
     }
 
-    /* Config UDP Packet_Size */
-    tmp = flb_output_get_property("packet_size", ins);
-    if (tmp != NULL && atoi(tmp) >= 0) {
-        ctx->pckt_size = atoi(tmp);
-    }
-    else {
-        ctx->pckt_size = 1420;
-    }
-
-    /* Config UDP Compress */
-    tmp = flb_output_get_property("compress", ins);
-    if (tmp) {
-        ctx->compress = flb_utils_bool(tmp);
-    }
-    else {
-        ctx->compress = FLB_TRUE;
-    }
-
     /* init random seed */
     if (flb_random_bytes((unsigned char *) &ctx->seed, sizeof(int))) {
         ctx->seed = time(NULL);
@@ -458,6 +468,10 @@ static int cb_gelf_exit(void *data, struct flb_config *config)
 {
     struct flb_out_gelf_config *ctx = data;
 
+    if (ctx == NULL) {
+        return 0;
+    }
+
     if (ctx->u) {
         flb_upstream_destroy(ctx->u);
     }
@@ -477,6 +491,58 @@ static int cb_gelf_exit(void *data, struct flb_config *config)
     return 0;
 }
 
+
+static struct flb_config_map config_map[] = {
+    {
+     FLB_CONFIG_MAP_STR, "mode", "udp",
+     0, FLB_FALSE, 0,
+     "The protocol to use. 'tls', 'tcp' or 'udp'"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "gelf_short_message_key", NULL,
+     0, FLB_FALSE, 0,
+     "A short descriptive message (MUST be set in GELF)"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "gelf_timestamp_key", NULL,
+     0, FLB_FALSE, 0,
+     "Timestamp key name (SHOULD be set in GELF)"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "gelf_host_key", NULL,
+     0, FLB_FALSE, 0,
+     "Key which its value is used as the name of the host,"
+     "source or application that sent this message. (MUST be set in GELF) "
+    },
+    {
+     FLB_CONFIG_MAP_STR, "gelf_full_message_key", NULL,
+     0, FLB_FALSE, 0,
+     "Key to use as the long message that can i.e. contain a backtrace. "
+     "(Optional in GELF)"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "gelf_level_key", NULL,
+     0, FLB_FALSE, 0,
+     "Key to be used as the log level. "
+     "Its value must be in standard syslog levels (between 0 and 7). "
+     "(Optional in GELF)"
+    },
+    {
+     FLB_CONFIG_MAP_INT, "packet_size", "1420",
+     0, FLB_TRUE, offsetof(struct flb_out_gelf_config, pckt_size),
+     "If transport protocol is udp, you can set the size of packets to be sent."
+    },
+    {
+     FLB_CONFIG_MAP_BOOL, "compress", "true",
+     0, FLB_TRUE, offsetof(struct flb_out_gelf_config, compress),
+     "If transport protocol is udp, "
+     "you can set this if you want your UDP packets to be compressed."
+    },
+
+    /* EOF */
+    {0}
+};
+
 /* Plugin reference */
 struct flb_output_plugin out_gelf_plugin = {
     .name           = "gelf",
@@ -486,4 +552,5 @@ struct flb_output_plugin out_gelf_plugin = {
     .cb_flush       = cb_gelf_flush,
     .cb_exit        = cb_gelf_exit,
     .flags          = FLB_OUTPUT_NET | FLB_IO_OPT_TLS,
+    .config_map     = config_map
 };
