@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -28,11 +27,27 @@
 #define FLB_HTTP_MIME_MSGPACK   "application/msgpack"
 #define FLB_HTTP_MIME_JSON      "application/json"
 
+#ifdef FLB_HAVE_SIGNV4
+#ifdef FLB_HAVE_AWS
+#define FLB_HTTP_AWS_CREDENTIAL_PREFIX "aws_"
+#endif
+#endif
+
 struct flb_out_http {
     /* HTTP Auth */
     char *http_user;
     char *http_passwd;
 
+    /* AWS Auth */
+#ifdef FLB_HAVE_SIGNV4
+#ifdef FLB_HAVE_AWS
+    int has_aws_auth;
+    struct flb_aws_provider *aws_provider;
+    const char *aws_region;
+    const char *aws_service;
+#endif
+#endif
+
     /* Proxy */
     const char *proxy;
     char *proxy_host;
@@ -53,6 +68,16 @@ struct flb_out_http {
     /* GELF fields */
     struct flb_gelf_fields gelf_fields;
 
+    /* which record key to use as body */
+    flb_sds_t body_key;
+
+    struct flb_record_accessor *body_ra;
+
+    /* override headers with contents of the map in the key specified here */
+    flb_sds_t headers_key;
+
+    struct flb_record_accessor *headers_ra;
+
     /* Include tag in header */
     flb_sds_t header_tag;
 
