@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -25,6 +24,7 @@
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_time.h>
 #include <fluent-bit/flb_http_client.h>
+#include <fluent-bit/flb_log_event_decoder.h>
 
 #include "newrelic.h"
 
@@ -144,15 +144,13 @@ static int package_record(struct flb_time *ts, msgpack_object *map,
 static flb_sds_t newrelic_compose_payload(struct flb_newrelic *ctx,
                                           const void *data, size_t bytes)
 {
-    int mp_ok = MSGPACK_UNPACK_SUCCESS;
     int total_records;
-    size_t off = 0;
     flb_sds_t json;
-    struct flb_time tms;
-    msgpack_unpacked result;
     msgpack_packer mp_pck;
     msgpack_sbuffer mp_sbuf;
-    msgpack_object *obj;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
+    int ret;
 
     /*
      * Following the New Relic Fluentd implementation, this is the
@@ -171,11 +169,19 @@ static flb_sds_t newrelic_compose_payload(struct flb_newrelic *ctx,
      *     ]}
      */
 
+    ret = flb_log_event_decoder_init(&log_decoder, (char *) data, bytes);
+
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
+
+        return NULL;
+    }
+
     /* Count number of records */
     total_records = flb_mp_count(data, bytes);
 
     /* Initialize msgpack buffers */
-    msgpack_unpacked_init(&result);
     msgpack_sbuffer_init(&mp_sbuf);
     msgpack_packer_init(&mp_pck, &mp_sbuf, msgpack_sbuffer_write);
 
@@ -217,18 +223,18 @@ static flb_sds_t newrelic_compose_payload(struct flb_newrelic *ctx,
     msgpack_pack_str_body(&mp_pck, "logs", 4);
     msgpack_pack_array(&mp_pck, total_records);
 
-    /* Iterate each record and pack it */
-    while (msgpack_unpack_next(&result, data, bytes, &off) == mp_ok) {
-        /* Retrive timestamp of the record */
-        flb_time_pop_from_msgpack(&tms, &result, &obj);
-
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
         /* Package the record */
-        package_record(&tms, obj, &mp_pck);
+        package_record(&log_event.timestamp, log_event.body, &mp_pck);
     }
 
+    flb_log_event_decoder_destroy(&log_decoder);
+
     json = flb_msgpack_raw_to_json_sds(mp_sbuf.data, mp_sbuf.size);
+
     msgpack_sbuffer_destroy(&mp_sbuf);
-    msgpack_unpacked_destroy(&result);
 
     return json;
 }
@@ -343,8 +349,8 @@ static int cb_newrelic_init(struct flb_output_instance *ins,
     return 0;
 }
 
-static void cb_newrelic_flush(const void *data, size_t bytes,
-                              const char *tag, int tag_len,
+static void cb_newrelic_flush(struct flb_event_chunk *event_chunk,
+                              struct flb_output_flush *out_flush,
                               struct flb_input_instance *i_ins,
                               void *out_context,
                               struct flb_config *config)
@@ -357,11 +363,12 @@ static void cb_newrelic_flush(const void *data, size_t bytes,
     void *payload_buf = NULL;
     size_t payload_size = 0;
     struct flb_newrelic *ctx = out_context;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
     struct flb_http_client *c;
 
     /* Format the data to the expected Newrelic Payload */
-    payload = newrelic_compose_payload(ctx, data, bytes);
+    payload = newrelic_compose_payload(ctx,
+                                       event_chunk->data, event_chunk->size);
     if (!payload) {
         flb_plg_error(ctx->ins, "cannot compose request payload");
         FLB_OUTPUT_RETURN(FLB_RETRY);
