@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -22,14 +21,17 @@
 #define FLB_PROMETHEUS_EXPORTER_H
 
 #include <fluent-bit/flb_output_plugin.h>
-#include <fluent-bit/flb_hash.h>
+#include <fluent-bit/flb_hash_table.h>
 
 /* Plugin context */
 struct prom_exporter {
     void *http;
 
     /* hash table for metrics reported */
-    struct flb_hash *ht_metrics;
+    struct flb_hash_table *ht_metrics;
+
+    /* add timestamp to every metric */
+    int add_timestamp;
 
     /* config reader for 'add_label' */
     struct mk_list *add_labels;
