@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2015-2022 The Fluent Bit Authors
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -49,6 +49,7 @@
 #define MONITORED_RESOURCE_KEY "logging.googleapis.com/monitored_resource"
 #define LOCAL_RESOURCE_ID_KEY "logging.googleapis.com/local_resource_id"
 #define DEFAULT_LABELS_KEY "logging.googleapis.com/labels"
+#define DEFAULT_PROJECT_ID_KEY "logging.googleapis.com/projectId"
 #define DEFAULT_SEVERITY_KEY "logging.googleapis.com/severity"
 #define DEFAULT_TRACE_KEY "logging.googleapis.com/trace"
 #define DEFAULT_SPAN_ID_KEY "logging.googleapis.com/spanId"
@@ -86,6 +87,11 @@
 #define FLB_STACKDRIVER_FAILED_REQUESTS      1001   /* failed requests */
 #endif
 
+// https://grpc.github.io/grpc/core/md_doc_statuscodes.html
+#define GRPC_STATUS_CODES_SIZE 17
+#define PARTIAL_SUCCESS_GRPC_TYPE "type.googleapis.com/google.logging.v2.WriteLogEntriesPartialErrors"
+#define PARTIAL_SUCCESS_GRPC_TYPE_SIZE 66
+
 struct flb_stackdriver_oauth_credentials {
     /* parsed credentials file */
     flb_sds_t type;
@@ -165,6 +171,7 @@ struct flb_stackdriver {
 
     /* other */
     flb_sds_t export_to_project_id;
+    flb_sds_t project_id_key;
     flb_sds_t resource;
     flb_sds_t severity_key;
     flb_sds_t trace_key;
@@ -174,6 +181,7 @@ struct flb_stackdriver {
     flb_sds_t http_request_key;
     int http_request_key_size;
     bool autoformat_stackdriver_trace;
+    bool test_log_entry_format;
 
     flb_sds_t stackdriver_agent;
 
