@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -21,7 +20,9 @@
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_api.h>
 #include <fluent-bit/flb_mem.h>
+#include <fluent-bit/flb_log.h>
 
+#include <fluent-bit/flb_input.h>
 #include <fluent-bit/flb_output.h>
 
 struct flb_api *flb_api_create()
@@ -35,6 +36,17 @@ struct flb_api *flb_api_create()
     }
 
     api->output_get_property = flb_output_get_property;
+    api->input_get_property = flb_input_get_property;
+
+#ifdef FLB_HAVE_METRICS
+    api->output_get_cmt_instance = flb_output_get_cmt_instance;
+    api->input_get_cmt_instance = flb_input_get_cmt_instance;
+#endif
+
+    api->log_print = flb_log_print;
+    api->input_log_check = flb_input_log_check;
+    api->output_log_check = flb_output_log_check;
+
     return api;
 }
 
