@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -18,10 +17,13 @@
  *  limitations under the License.
  */
 
+#include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
 
 #include <monkey/mk_core.h>
+#include <fluent-bit/flb_bucket_queue.h>
+#include <fluent-bit/flb_event_loop.h>
 #include <fluent-bit/flb_time.h>
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_bits.h>
@@ -45,7 +47,12 @@
 #include <fluent-bit/flb_sosreport.h>
 #include <fluent-bit/flb_storage.h>
 #include <fluent-bit/flb_http_server.h>
+#include <fluent-bit/flb_pack.h>
 #include <fluent-bit/flb_metrics.h>
+#include <fluent-bit/flb_version.h>
+#include <fluent-bit/flb_upstream.h>
+#include <fluent-bit/flb_downstream.h>
+#include <fluent-bit/flb_ring_buffer.h>
 
 #ifdef FLB_HAVE_METRICS
 #include <fluent-bit/flb_metrics_exporter.h>
@@ -61,12 +68,19 @@
 extern struct flb_aws_error_reporter *error_reporter;
 #endif
 
+#include <ctraces/ctr_version.h>
+
+static pthread_once_t local_thread_engine_evl_init = PTHREAD_ONCE_INIT;
 FLB_TLS_DEFINE(struct mk_event_loop, flb_engine_evl);
 
+static void flb_engine_evl_init_private()
+{
+    FLB_TLS_INIT(flb_engine_evl);
+}
 
 void flb_engine_evl_init()
 {
-    FLB_TLS_INIT(flb_engine_evl);
+    pthread_once(&local_thread_engine_evl_init, flb_engine_evl_init_private);
 }
 
 struct mk_event_loop *flb_engine_evl_get()
@@ -98,6 +112,40 @@ int flb_engine_destroy_tasks(struct mk_list *tasks)
     return c;
 }
 
+void flb_engine_reschedule_retries(struct flb_config *config)
+{
+    int ret;
+    struct mk_list *head;
+    struct mk_list *t_head;
+    struct mk_list *rt_head;
+    struct mk_list *tmp_task;
+    struct mk_list *tmp_retry_task;
+    struct flb_task *task;
+    struct flb_input_instance *ins;
+    struct flb_task_retry *retry;
+
+    /* Invalidate and reschedule all retry tasks to be retried immediately */
+    mk_list_foreach(head, &config->inputs) {
+        ins = mk_list_entry(head, struct flb_input_instance, _head);
+        mk_list_foreach_safe(t_head, tmp_task, &ins->tasks) {
+            task = mk_list_entry(t_head, struct flb_task, _head);
+            mk_list_foreach_safe(rt_head, tmp_retry_task, &task->retries) {
+                retry = mk_list_entry(rt_head, struct flb_task_retry, _head);
+                flb_sched_request_invalidate(config, retry);
+                ret = flb_sched_retry_now(config, retry);
+                if (ret == -1) {
+                    /* Can't do much here, just continue on */
+                    flb_warn("[engine] failed to immediately re-schedule retry=%p "
+                             "for task %i. Err: %d", retry, task->id, flb_errno());
+                } else {
+                    flb_debug("[engine] re-scheduled retry=%p for task %i",
+                        retry, task->id);
+                }
+            }
+        }
+    }
+}
+
 int flb_engine_flush(struct flb_config *config,
                      struct flb_input_plugin *in_force)
 {
@@ -123,26 +171,20 @@ static void cb_engine_sched_timer(struct flb_config *ctx, void *data)
 {
     (void) data;
 
-    /* Upstream connections timeouts handling */
+    /* Upstream timeout handling */
     flb_upstream_conn_timeouts(&ctx->upstreams);
+
+    /* Downstream timeout handling */
+    flb_downstream_conn_timeouts(&ctx->downstreams);
 }
 
-static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
-                                      struct flb_config *config)
+static inline int handle_input_event(flb_pipefd_t fd, uint64_t ts,
+                                     struct flb_config *config)
 {
-    int ret;
     int bytes;
-    int task_id;
-    int out_id;
-    int retries;
-    int retry_seconds;
     uint32_t type;
-    uint32_t key;
+    uint32_t ins_id;
     uint64_t val;
-    char *name;
-    struct flb_task *task;
-    struct flb_task_retry *retry;
-    struct flb_output_instance *ins;
 
     bytes = flb_pipe_r(fd, &val, sizeof(val));
     if (bytes == -1) {
@@ -150,6 +192,49 @@ static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
         return -1;
     }
 
+    /* Get type and key */
+    type   = FLB_BITS_U64_HIGH(val);
+    ins_id = FLB_BITS_U64_LOW(val);
+
+    /* At the moment we only support events coming from an input coroutine */
+    if (type != FLB_ENGINE_IN_CORO) {
+        flb_error("[engine] invalid event type %i for input handler",
+                  type);
+        return -1;
+    }
+
+    flb_input_coro_finished(config, ins_id);
+    return 0;
+}
+
+static inline double calculate_chunk_capacity_percent(struct flb_output_instance *ins)
+{
+    /* Currently, total_limit_size 0(K|M)B will be translated as no
+     * limit. So, we need to handle this situation to be unlimited. */
+    if (ins->total_limit_size <= 0) {
+        return 100.0;
+    }
+
+    return 100 * (1.0 - (ins->fs_backlog_chunks_size + ins->fs_chunks_size)/
+                  ((double)ins->total_limit_size));
+}
+
+static inline int handle_output_event(uint64_t ts,
+                                      struct flb_config *config,
+                                      uint64_t val)
+{
+    int ret;
+    int task_id;
+    int out_id;
+    int retries;
+    int retry_seconds;
+    uint32_t type;
+    uint32_t key;
+    char *name;
+    struct flb_task *task;
+    struct flb_task_retry *retry;
+    struct flb_output_instance *ins;
+
     /* Get type and key */
     type = FLB_BITS_U64_HIGH(val);
     key  = FLB_BITS_U64_LOW(val);
@@ -193,20 +278,29 @@ static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
     }
     name = (char *) flb_output_name(ins);
 
+    /* If we are in synchronous mode, flush the next waiting task */
+    if (ins->flags & FLB_OUTPUT_SYNCHRONOUS) {
+        if (ret == FLB_OK || ret == FLB_RETRY || ret == FLB_ERROR) {
+            flb_output_task_singleplex_flush_next(ins->singleplex_queue);
+        }
+    }
+
     /* A task has finished, delete it */
     if (ret == FLB_OK) {
         /* cmetrics */
-        cmt_counter_add(ins->cmt_proc_records, ts, task->records,
+        cmt_counter_add(ins->cmt_proc_records, ts, task->event_chunk->total_events,
                         1, (char *[]) {name});
 
-        cmt_counter_add(ins->cmt_proc_bytes, ts, task->size,
+        cmt_counter_add(ins->cmt_proc_bytes, ts, task->event_chunk->size,
                         1, (char *[]) {name});
 
         /* [OLD API] Update metrics */
 #ifdef FLB_HAVE_METRICS
         if (ins->metrics) {
-            flb_metrics_sum(FLB_METRIC_OUT_OK_RECORDS, task->records, ins->metrics);
-            flb_metrics_sum(FLB_METRIC_OUT_OK_BYTES, task->size, ins->metrics);
+            flb_metrics_sum(FLB_METRIC_OUT_OK_RECORDS,
+                            task->event_chunk->total_events, ins->metrics);
+            flb_metrics_sum(FLB_METRIC_OUT_OK_BYTES,
+                            task->event_chunk->size, ins->metrics);
         }
 #endif
         /* Inform the user if a 'retry' succedeed */
@@ -229,15 +323,24 @@ static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
                      flb_input_name(task->i_ins),
                      flb_output_name(ins), out_id);
         }
+
+        cmt_gauge_set(ins->cmt_chunk_available_capacity_percent, ts,
+                      calculate_chunk_capacity_percent(ins),
+                      1, (char *[]) {name});
+
         flb_task_retry_clean(task, ins);
         flb_task_users_dec(task, FLB_TRUE);
     }
-    else if (ret == FLB_RETRY && config->is_running && !config->is_shutting_down) {
+    else if (ret == FLB_RETRY) {
         if (ins->retry_limit == FLB_OUT_RETRY_NONE) {
             /* cmetrics: output_dropped_records_total */
             cmt_counter_add(ins->cmt_dropped_records, ts, task->records,
                             1, (char *[]) {name});
 
+            cmt_gauge_set(ins->cmt_chunk_available_capacity_percent, ts,
+                          calculate_chunk_capacity_percent(ins),
+                          1, (char *[]) {name});
+
             /* OLD metrics API */
 #ifdef FLB_HAVE_METRICS
             flb_metrics_sum(FLB_METRIC_OUT_DROPPED_RECORDS, task->records, ins->metrics);
@@ -248,7 +351,10 @@ static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
                      task_id,
                      flb_input_name(task->i_ins),
                      flb_output_name(ins), out_id);
+
+            flb_task_retry_clean(task, ins);
             flb_task_users_dec(task, FLB_TRUE);
+
             return 0;
         }
 
@@ -267,20 +373,26 @@ static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
             cmt_counter_add(ins->cmt_dropped_records, ts, task->records,
                             1, (char *[]) {name});
 
+            cmt_gauge_set(ins->cmt_chunk_available_capacity_percent, ts,
+                          calculate_chunk_capacity_percent(ins),
+                          1, (char *[]) {name});
+
             /* OLD metrics API */
 #ifdef FLB_HAVE_METRICS
             flb_metrics_sum(FLB_METRIC_OUT_RETRY_FAILED, 1, ins->metrics);
             flb_metrics_sum(FLB_METRIC_OUT_DROPPED_RECORDS, task->records, ins->metrics);
 #endif
             /* Notify about this failed retry */
-            flb_warn("[engine] chunk '%s' cannot be retried: "
-                     "task_id=%i, input=%s > output=%s",
-                     flb_input_chunk_get_name(task->ic),
-                     task_id,
-                     flb_input_name(task->i_ins),
-                     flb_output_name(ins));
+            flb_error("[engine] chunk '%s' cannot be retried: "
+                      "task_id=%i, input=%s > output=%s",
+                      flb_input_chunk_get_name(task->ic),
+                      task_id,
+                      flb_input_name(task->i_ins),
+                      flb_output_name(ins));
 
+            flb_task_retry_clean(task, ins);
             flb_task_users_dec(task, FLB_TRUE);
+
             return 0;
         }
 
@@ -321,6 +433,10 @@ static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
             cmt_counter_add(ins->cmt_retried_records, ts, task->records,
                             1, (char *[]) {name});
 
+            cmt_gauge_set(ins->cmt_chunk_available_capacity_percent, ts,
+                          calculate_chunk_capacity_percent(ins),
+                          1, (char *[]) {name});
+
             /* OLD metrics API: update the metrics since a new retry is coming */
 #ifdef FLB_HAVE_METRICS
             flb_metrics_sum(FLB_METRIC_OUT_RETRY, 1, ins->metrics);
@@ -334,17 +450,69 @@ static inline int handle_output_event(flb_pipefd_t fd, uint64_t ts,
         cmt_counter_add(ins->cmt_dropped_records, ts, task->records,
                         1, (char *[]) {name});
 
+        cmt_gauge_set(ins->cmt_chunk_available_capacity_percent, ts,
+                      calculate_chunk_capacity_percent(ins),
+                      1, (char *[]) {name});
+
         /* OLD API */
 #ifdef FLB_HAVE_METRICS
         flb_metrics_sum(FLB_METRIC_OUT_ERROR, 1, ins->metrics);
         flb_metrics_sum(FLB_METRIC_OUT_DROPPED_RECORDS, task->records, ins->metrics);
 #endif
+
+        flb_task_retry_clean(task, ins);
         flb_task_users_dec(task, FLB_TRUE);
     }
 
     return 0;
 }
 
+static inline int handle_output_events(flb_pipefd_t fd,
+                                       struct flb_config *config)
+{
+    uint64_t values[FLB_ENGINE_OUTPUT_EVENT_BATCH_SIZE];
+    int      result;
+    int      bytes;
+    size_t   limit;
+    size_t   index;
+    uint64_t ts;
+
+    memset(&values, 0, sizeof(values));
+
+    bytes = flb_pipe_r(fd, &values, sizeof(values));
+
+    if (bytes == -1) {
+        flb_errno();
+        return -1;
+    }
+
+    limit = floor(bytes / sizeof(uint64_t));
+
+    ts = cfl_time_now();
+
+    for (index = 0 ;
+         index < limit &&
+         index < (sizeof(values) / sizeof(values[0])) ;
+         index++) {
+        if (values[index] == 0) {
+            break;
+        }
+
+        result = handle_output_event(ts, config, values[index]);
+    }
+
+    /* This is wrong, in one hand, if handle_output_event_ fails we should
+     * stop, on the other, we have already consumed the signals from the pipe
+     * so we have to do whatever we can with them.
+     *
+     * And a side effect is that since we have N results but we are not aborting
+     * as soon as we get an error there could be N results to this function which
+     * not only are we not ready to handle but is not even checked at the moment.
+    */
+
+    return result;
+}
+
 static inline int flb_engine_manager(flb_pipefd_t fd, struct flb_config *config)
 {
     int bytes;
@@ -371,10 +539,6 @@ static inline int flb_engine_manager(flb_pipefd_t fd, struct flb_config *config)
             return FLB_ENGINE_STOP;
         }
     }
-    else if (type == FLB_ENGINE_IN_THREAD) {
-        /* Event coming from an input thread */
-        flb_input_coro_destroy_id(key, config);
-    }
 
     return 0;
 }
@@ -455,6 +619,7 @@ int flb_engine_failed(struct flb_config *config)
 
     /* Check the channel is valid (enabled by library mode) */
     if (config->ch_notif[1] <= 0) {
+        flb_error("[engine] no channel to notify FAILED message");
         return -1;
     }
 
@@ -464,6 +629,9 @@ int flb_engine_failed(struct flb_config *config)
         flb_error("[engine] fail to dispatch FAILED message");
     }
 
+    /* Waiting flushing log */
+    sleep(1);
+
     return ret;
 }
 
@@ -495,33 +663,80 @@ static int flb_engine_log_start(struct flb_config *config)
     return 0;
 }
 
+static void flb_engine_drain_ring_buffer_signal_channel(flb_pipefd_t fd)
+{
+    static char signal_buffer[512];
+
+    flb_pipe_r(fd, signal_buffer, sizeof(signal_buffer));
+}
+
+
+#ifdef FLB_HAVE_IN_STORAGE_BACKLOG
 extern int sb_segregate_chunks(struct flb_config *config);
+#else
+int sb_segregate_chunks(struct flb_config *config)
+{
+    return 0;
+}
+#endif
 
 int flb_engine_start(struct flb_config *config)
 {
     int ret;
     uint64_t ts;
     char tmp[16];
+    int rb_flush_flag;
     struct flb_time t_flush;
     struct mk_event *event;
     struct mk_event_loop *evl;
+    struct flb_bucket_queue *evl_bktq;
     struct flb_sched *sched;
     struct flb_net_dns dns_ctx;
 
     /* Initialize the networking layer */
     flb_net_lib_init();
-
     flb_net_ctx_init(&dns_ctx);
     flb_net_dns_ctx_init();
     flb_net_dns_ctx_set(&dns_ctx);
 
+    flb_pack_init(config);
+
     /* Create the event loop and set it in the global configuration */
     evl = mk_event_loop_create(256);
     if (!evl) {
+        fprintf(stderr, "[log] could not create event loop\n");
         return -1;
     }
     config->evl = evl;
 
+    /* Create the bucket queue (FLB_ENGINE_PRIORITY_COUNT priorities) */
+    evl_bktq = flb_bucket_queue_create(FLB_ENGINE_PRIORITY_COUNT);
+    if (!evl_bktq) {
+        return -1;
+    }
+    config->evl_bktq = evl_bktq;
+
+    /*
+     * Event loop channel to ingest flush events from flb_engine_flush()
+     *
+     *  - FLB engine uses 'ch_self_events[1]' to dispatch tasks to self
+     *  - Self to receive message on ch_parent_events[0]
+     *
+     * The mk_event_channel_create() will attach the pipe read end ch_self_events[0]
+     * to the local event loop 'evl'.
+     */
+    ret = mk_event_channel_create(config->evl,
+                                    &config->ch_self_events[0],
+                                    &config->ch_self_events[1],
+                                    &config->event_thread_init);
+    if (ret == -1) {
+        flb_error("[engine] could not create engine thread channel");
+        return -1;
+    }
+    /* Signal type to indicate a "flush" request */
+    config->event_thread_init.type = FLB_ENGINE_EV_THREAD_ENGINE;
+    config->event_thread_init.priority = FLB_ENGINE_PRIORITY_THREAD;
+
     /* Register the event loop on this thread */
     flb_engine_evl_init();
     flb_engine_evl_set(evl);
@@ -529,10 +744,12 @@ int flb_engine_start(struct flb_config *config)
     /* Start the Logging service */
     ret = flb_engine_log_start(config);
     if (ret == -1) {
+        fprintf(stderr, "[engine] log start failed\n");
         return -1;
     }
 
-    flb_info("[engine] started (pid=%i)", getpid());
+    flb_info("[fluent bit] version=%s, commit=%.10s, pid=%i",
+             FLB_VERSION_STR, FLB_GIT_HASH, getpid());
 
     /* Debug coroutine stack size */
     flb_utils_bytes_to_human_readable_size(config->coro_stack_size,
@@ -548,7 +765,7 @@ int flb_engine_start(struct flb_config *config)
     ret = mk_event_channel_create(config->evl,
                                   &config->ch_manager[0],
                                   &config->ch_manager[1],
-                                  config);
+                                  &config->ch_event);
     if (ret != 0) {
         flb_error("[engine] could not create manager channels");
         return -1;
@@ -563,12 +780,14 @@ int flb_engine_start(struct flb_config *config)
     /* Start the Storage engine */
     ret = flb_storage_create(config);
     if (ret == -1) {
+        flb_error("[engine] storage creation failed");
         return -1;
     }
 
     /* Init Metrics engine */
     cmt_initialize();
     flb_info("[cmetrics] version=%s", cmt_version());
+    flb_info("[ctraces ] version=%s", ctr_version());
 
     /* Initialize the scheduler */
     sched = flb_sched_create(config, config->evl);
@@ -582,16 +801,17 @@ int flb_engine_start(struct flb_config *config)
     flb_sched_ctx_init();
     flb_sched_ctx_set(sched);
 
-
     /* Initialize input plugins */
     ret = flb_input_init_all(config);
     if (ret == -1) {
+        flb_error("[engine] input initialization failed");
         return -1;
     }
 
     /* Initialize filter plugins */
     ret = flb_filter_init_all(config);
     if (ret == -1) {
+        flb_error("[engine] filter initialization failed");
         return -1;
     }
 
@@ -601,6 +821,7 @@ int flb_engine_start(struct flb_config *config)
     /* Initialize output plugins */
     ret = flb_output_init_all(config);
     if (ret == -1) {
+        flb_error("[engine] output initialization failed");
         return -1;
     }
 
@@ -617,6 +838,7 @@ int flb_engine_start(struct flb_config *config)
                                                t_flush.tm.tv_sec,
                                                t_flush.tm.tv_nsec,
                                                event);
+    event->priority = FLB_ENGINE_PRIORITY_FLUSH;
     if (config->flush_fd == -1) {
         flb_utils_error(FLB_ERR_CFG_FLUSH_CREATE);
     }
@@ -628,9 +850,6 @@ int flb_engine_start(struct flb_config *config)
     }
 #endif
 
-    /* Initialize collectors */
-    flb_input_collectors_start(config);
-
     /* Prepare routing paths */
     ret = flb_router_io_set(config);
     if (ret == -1) {
@@ -666,6 +885,9 @@ int flb_engine_start(struct flb_config *config)
     }
 #endif
 
+    /* Initialize collectors */
+    flb_input_collectors_start(config);
+
     /*
      * Sched a permanent callback triggered every 1.5 second to let other
      * Fluent Bit components run tasks at that interval.
@@ -678,28 +900,65 @@ int flb_engine_start(struct flb_config *config)
         return -1;
     }
 
+    /* DEV/TEST change only */
+    int rb_ms;
+    char *rb_env;
+
+    rb_env = getenv("FLB_DEV_RB_MS");
+    if (!rb_env) {
+        rb_ms = 250;
+    }
+    else {
+        rb_ms = atoi(rb_env);
+    }
+
+    /* Input instance / Ring buffer collector */
+    ret = flb_sched_timer_cb_create(config->sched,
+                                    FLB_SCHED_TIMER_CB_PERM,
+                                    rb_ms, flb_input_chunk_ring_buffer_collector,
+                                    config, NULL);
+    if (ret == -1) {
+        flb_error("[engine] could not schedule permanent callback");
+        return -1;
+    }
+
     /* Signal that we have started */
     flb_engine_started(config);
 
     ret = sb_segregate_chunks(config);
 
-    if (ret)
-    {
+    if (ret) {
         flb_error("[engine] could not segregate backlog chunks");
         return -2;
     }
 
     while (1) {
-        mk_event_wait(evl);
-        mk_event_foreach(event, evl) {
+        rb_flush_flag = FLB_FALSE;
+
+        mk_event_wait(evl); /* potentially conditional mk_event_wait or mk_event_wait_2 based on bucket queue capacity for one shot events */
+        flb_event_priority_live_foreach(event, evl_bktq, evl, FLB_ENGINE_LOOP_MAX_ITER) {
             if (event->type == FLB_ENGINE_EV_CORE) {
                 ret = flb_engine_handle_event(event->fd, event->mask, config);
                 if (ret == FLB_ENGINE_STOP) {
                     if (config->grace_count == 0) {
-                        flb_warn("[engine] service will shutdown in max %u seconds",
+                        if (config->grace >= 0) {
+                            flb_warn("[engine] service will shutdown in max %u seconds",
                                  config->grace);
+                        } else {
+                            flb_warn("[engine] service will shutdown when all remaining tasks are flushed");
+                        }
+
+                        /* Reschedule retry tasks to be retried immediately */
+                        flb_engine_reschedule_retries(config);
                     }
 
+                    /* mark the runtime as the ingestion is not active and that we are in shutting down mode */
+                    config->is_ingestion_active = FLB_FALSE;
+                    config->is_shutting_down = FLB_TRUE;
+
+                    /* pause all input plugin instances */
+                    flb_input_pause_all(config);
+
                     /*
                      * We are preparing to shutdown, we give a graceful time
                      * of 'config->grace' seconds to process any pending event.
@@ -721,6 +980,7 @@ int flb_engine_start(struct flb_config *config)
                                                                   1,
                                                                   0,
                                                                   event);
+                    event->priority = FLB_ENGINE_PRIORITY_SHUTDOWN;
                 }
                 else if (ret == FLB_ENGINE_SHUTDOWN) {
                     if (config->shutdown_fd > 0) {
@@ -732,14 +992,16 @@ int flb_engine_start(struct flb_config *config)
                     config->grace_count++;
 
                     /*
-                     * Grace period has finished, but we need to check if there is
+                     * Grace timeout has finished, but we need to check if there is
                      * any pending running task. A running task is associated to an
                      * output co-routine, since we don't know what's the state or
                      * resources allocated by that co-routine, the best thing is to
                      * wait again for the grace period and re-check again.
+                     * If grace period is set to -1, keep trying to shut down until all
+                     * tasks and retries get flushed.
                      */
                     ret = flb_task_running_count(config);
-                    if (ret > 0 && config->grace_count < config->grace) {
+                    if (ret > 0 && (config->grace_count < config->grace || config->grace == -1)) {
                         if (config->grace_count == 1) {
                             flb_task_running_print(config);
                         }
@@ -762,33 +1024,58 @@ int flb_engine_start(struct flb_config *config)
                 /* Event type registered by the Scheduler */
                 flb_sched_event_handler(config, event);
             }
+            else if (event->type == FLB_ENGINE_EV_THREAD_ENGINE) {
+                struct flb_output_flush *output_flush;
+
+                /* Read the coroutine reference */
+                ret = flb_pipe_r(event->fd, &output_flush, sizeof(struct flb_output_flush *));
+                if (ret <= 0 || output_flush == 0) {
+                    flb_errno();
+                    continue;
+                }
+
+                /* Init coroutine */
+                flb_coro_resume(output_flush->coro);
+            }
             else if (event->type == FLB_ENGINE_EV_CUSTOM) {
                 event->handler(event);
             }
             else if (event->type == FLB_ENGINE_EV_THREAD) {
-                struct flb_upstream_conn *u_conn;
-                struct flb_coro *co;
+                struct flb_connection *connection;
 
                 /*
                  * Check if we have some co-routine associated to this event,
                  * if so, resume the co-routine
                  */
-                u_conn = (struct flb_upstream_conn *) event;
-                co = u_conn->coro;
-                if (co) {
-                    flb_trace("[engine] resuming coroutine=%p", co);
-                    flb_coro_resume(co);
+
+                connection = (struct flb_connection *) event;
+
+                if (connection->coroutine) {
+                    flb_trace("[engine] resuming coroutine=%p", connection->coroutine);
+
+                    flb_coro_resume(connection->coroutine);
                 }
             }
             else if (event->type == FLB_ENGINE_EV_OUTPUT) {
-                ts = cmt_time_now();
-
                 /*
                  * Event originated by an output plugin. likely a Task return
                  * status.
                  */
-                handle_output_event(event->fd, ts, config);
+                handle_output_events(event->fd, config);
             }
+            else if (event->type == FLB_ENGINE_EV_INPUT) {
+                ts = cfl_time_now();
+                handle_input_event(event->fd, ts, config);
+            }
+            else if(event->type == FLB_ENGINE_EV_THREAD_INPUT) {
+                flb_engine_drain_ring_buffer_signal_channel(event->fd);
+
+                rb_flush_flag = FLB_TRUE;
+            }
+        }
+
+        if (rb_flush_flag) {
+            flb_input_chunk_ring_buffer_collector(config, NULL);
         }
 
         /* Cleanup functions associated to events and timers */
@@ -796,6 +1083,7 @@ int flb_engine_start(struct flb_config *config)
             flb_net_dns_lookup_context_cleanup(&dns_ctx);
             flb_sched_timer_cleanup(config->sched);
             flb_upstream_conn_pending_destroy_list(&config->upstreams);
+            flb_downstream_conn_pending_destroy_list(&config->downstreams);
 
             /*
             * depend on main thread to clean up expired message
@@ -828,9 +1116,9 @@ int flb_engine_shutdown(struct flb_config *config)
 
     /* cleanup plugins */
     flb_filter_exit(config);
-    flb_input_exit_all(config);
     flb_output_exit(config);
     flb_custom_exit(config);
+    flb_input_exit_all(config);
 
     /* Destroy the storage context */
     flb_storage_destroy(config);
@@ -854,12 +1142,7 @@ int flb_engine_shutdown(struct flb_config *config)
 int flb_engine_exit(struct flb_config *config)
 {
     int ret;
-    uint64_t val = FLB_ENGINE_EV_STOP;
-
-    config->is_ingestion_active = FLB_FALSE;
-    config->is_shutting_down = FLB_TRUE;
-
-    flb_input_pause_all(config);
+    uint64_t val;
 
     val = FLB_ENGINE_EV_STOP;
     ret = flb_pipe_w(config->ch_manager[1], &val, sizeof(uint64_t));
