@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -29,12 +28,15 @@
 #include <fluent-bit/flb_coro.h>
 #include <fluent-bit/flb_engine.h>
 #include <fluent-bit/flb_task.h>
+#include <fluent-bit/flb_event.h>
+
 
 /* It creates a new output thread using a 'Retry' context */
 int flb_engine_dispatch_retry(struct flb_task_retry *retry,
                               struct flb_config *config)
 {
     int ret;
+    char *buf_data;
     size_t buf_size;
     struct flb_task *task;
 
@@ -61,21 +63,37 @@ int flb_engine_dispatch_retry(struct flb_task_retry *retry,
     }
 
     /* There is a match, get the buffer */
-    task->buf = flb_input_chunk_flush(task->ic, &buf_size);
-    task->size = buf_size;
-
-    if (!task->buf) {
+    buf_data = (char *) flb_input_chunk_flush(task->ic, &buf_size);
+    if (!buf_data) {
         /* Could not retrieve chunk content */
         flb_error("[engine_dispatch] could not retrieve chunk content, removing retry");
         flb_task_retry_destroy(retry);
         return -1;
     }
 
-    ret = flb_output_task_flush(task, retry->o_ins, config);
-    if (ret == -1) {
-        flb_task_retry_destroy(retry);
-        return -1;
+    /* Update the buffer reference */
+    flb_event_chunk_update(task->event_chunk, buf_data, buf_size);
+
+    /* flush the task */
+    if (retry->o_ins->flags & FLB_OUTPUT_SYNCHRONOUS) {
+        /*
+         * If the plugin doesn't allow for multiplexing.
+         * singleplex_enqueue deletes retry context on flush or delayed flush failure
+         */
+        ret = flb_output_task_singleplex_enqueue(retry->o_ins->singleplex_queue, retry,
+                                                 task, retry->o_ins, config);
+        if (ret == -1) {
+            return -1;
+        }
     }
+    else {
+        ret = flb_output_task_flush(task, retry->o_ins, config);
+        if (ret == -1) {
+            flb_task_retry_destroy(retry);
+            return -1;
+        }
+    }
+
     return 0;
 }
 
@@ -89,16 +107,19 @@ static void test_run_formatter(struct flb_config *config,
     void *out_buf = NULL;
     size_t out_size = 0;
     struct flb_test_out_formatter *otf;
+    struct flb_event_chunk *evc;
 
     otf = &o_ins->test_formatter;
+    evc = task->event_chunk;
 
     /* Invoke the output plugin formatter test callback */
     ret = otf->callback(config,
                         i_ins,
                         o_ins->context,
                         flush_ctx,
-                        task->tag, task->tag_len,
-                        task->buf, task->size,
+                        evc->type,
+                        evc->tag, flb_sds_len(evc->tag),
+                        evc->data, evc->size,
                         &out_buf, &out_size);
 
     /* Call the runtime test callback checker */
@@ -177,10 +198,20 @@ static int tasks_start(struct flb_input_instance *in,
             hits++;
 
             /*
-             * We have the Task and the Route, created a thread context for the
-             * data handling.
+             * If the plugin is in synchronous mode, enqueue the task and flush
+             * when appropriate.
              */
-            flb_output_task_flush(task, route->out, config);
+            if (out->flags & FLB_OUTPUT_SYNCHRONOUS) {
+                flb_output_task_singleplex_enqueue(route->out->singleplex_queue, NULL,
+                                                   task, route->out, config);
+            }
+            else {
+                /*
+                 * We have the Task and the Route, created a thread context for the
+                 * data handling.
+                 */
+                flb_output_task_flush(task, route->out, config);
+            }
 
             /*
             th = flb_output_thread(task,
