@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -103,12 +102,16 @@ struct flb_metrics *flb_metrics_create(const char *title)
 int flb_metrics_title(const char *title, struct flb_metrics *metrics)
 {
     int ret;
+    size_t size = sizeof(metrics->title) - 1;
 
-    ret = snprintf(metrics->title, sizeof(metrics->title) - 1, "%s", title);
+    ret = snprintf(metrics->title, size, "%s", title);
     if (ret == -1) {
         flb_errno();
         return -1;
     }
+    else if (ret >= size){
+        flb_warn("[%s] title '%s' was truncated", __FUNCTION__, title);
+    }
     metrics->title_len = strlen(metrics->title);
     return 0;
 }
@@ -117,6 +120,7 @@ int flb_metrics_add(int id, const char *title, struct flb_metrics *metrics)
 {
     int ret;
     struct flb_metric *m;
+    size_t size;
 
     /* Create context */
     m = flb_malloc(sizeof(struct flb_metric));
@@ -125,14 +129,19 @@ int flb_metrics_add(int id, const char *title, struct flb_metrics *metrics)
         return -1;
     }
     m->val = 0;
+    size = sizeof(m->title) - 1;
 
     /* Write title */
-    ret = snprintf(m->title, sizeof(m->title) - 1, "%s", title);
+    ret = snprintf(m->title, size, "%s", title);
     if (ret == -1) {
         flb_errno();
         flb_free(m);
         return -1;
     }
+    else if (ret >= size) {
+        flb_warn("[%s] title '%s' was truncated", __FUNCTION__, title);
+    }
+
     m->title_len = strlen(m->title);
 
     /* Assign an ID */
@@ -311,6 +320,25 @@ static int attach_build_info(struct flb_config *ctx, struct cmt *cmt, uint64_t t
     return 0;
 }
 
+static int attach_hot_reload_info(struct flb_config *ctx, struct cmt *cmt, uint64_t ts,
+                                  char *hostname)
+{
+    double val;
+    struct cmt_gauge *g;
+
+    g = cmt_gauge_create(cmt, "fluentbit", "", "hot_reloaded_times",
+                         "Collect the count of hot reloaded times.",
+                         1, (char *[]) {"hostname"});
+    if (!g) {
+        return -1;
+    }
+
+    val = (double) ctx->hot_reloaded_count;
+
+    cmt_gauge_set(g, ts, val, 1, (char *[]) {hostname});
+    return 0;
+}
+
 /* Append internal Fluent Bit metrics to context */
 int flb_metrics_fluentbit_add(struct flb_config *ctx, struct cmt *cmt)
 {
@@ -319,19 +347,19 @@ int flb_metrics_fluentbit_add(struct flb_config *ctx, struct cmt *cmt)
     char hostname[128];
 
     /* current timestamp */
-    ts = cmt_time_now();
+    ts = cfl_time_now();
 
     /* get hostname */
     ret = gethostname(hostname, sizeof(hostname) - 1);
     if (ret == -1) {
-        strncpy(hostname, "unknown", 7);
-        hostname[7] = '\0';
+        strcpy(hostname, "unknown");
     }
 
     /* Attach metrics to cmetrics context */
     attach_uptime(ctx, cmt, ts, hostname);
     attach_process_start_time_seconds(ctx, cmt, ts, hostname);
     attach_build_info(ctx, cmt, ts, hostname);
+    attach_hot_reload_info(ctx, cmt, ts, hostname);
 
     return 0;
 }
