@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -25,6 +24,7 @@
 #include <fluent-bit/flb_mp.h>
 #include <fluent-bit/flb_slist.h>
 #include <fluent-bit/flb_record_accessor.h>
+#include <fluent-bit/flb_metrics.h>
 
 #include <msgpack.h>
 #include <mpack/mpack.h>
@@ -33,23 +33,84 @@
 #define pack_uint16(buf, d) _msgpack_store16(buf, (uint16_t) d)
 #define pack_uint32(buf, d) _msgpack_store32(buf, (uint32_t) d)
 
+/* Return the number of msgpack serialized events in the buffer */
 int flb_mp_count(const void *data, size_t bytes)
 {
+    return flb_mp_count_remaining(data, bytes, NULL);
+}
+
+int flb_mp_count_remaining(const void *data, size_t bytes, size_t *remaining_bytes)
+{
+    size_t remaining;
     int count = 0;
     mpack_reader_t reader;
 
     mpack_reader_init_data(&reader, (const char *) data, bytes);
-    while (mpack_reader_remaining(&reader, NULL) > 0) {
-        count++;
+    for (;;) {
+        remaining = mpack_reader_remaining(&reader, NULL);
+        if (!remaining) {
+            break;
+        }
         mpack_discard(&reader);
+        if (mpack_reader_error(&reader)) {
+            break;
+        }
+        count++;
     }
 
+    if (remaining_bytes) {
+        *remaining_bytes = remaining;
+    }
     mpack_reader_destroy(&reader);
     return count;
 }
 
-int flb_mp_validate_chunk(const void *data, size_t bytes,
-                          int *out_records, size_t *processed_bytes)
+int flb_mp_validate_metric_chunk(const void *data, size_t bytes,
+                                 int *out_series, size_t *processed_bytes)
+{
+    int ret;
+    int ok = CMT_DECODE_MSGPACK_SUCCESS;
+    int count = 0;
+    size_t off = 0;
+    size_t pre_off = 0;
+    struct cmt *cmt;
+
+    while ((ret = cmt_decode_msgpack_create(&cmt,
+                                            (char *) data, bytes, &off)) == ok) {
+        cmt_destroy(cmt);
+        count++;
+        pre_off = off;
+    }
+
+    switch (ret) {
+        case CMT_DECODE_MSGPACK_INVALID_ARGUMENT_ERROR:
+        case CMT_DECODE_MSGPACK_CORRUPT_INPUT_DATA_ERROR:
+        case CMT_DECODE_MSGPACK_CONSUME_ERROR:
+        case CMT_DECODE_MSGPACK_ENGINE_ERROR:
+        case CMT_DECODE_MSGPACK_PENDING_MAP_ENTRIES:
+        case CMT_DECODE_MSGPACK_PENDING_ARRAY_ENTRIES:
+        case CMT_DECODE_MSGPACK_UNEXPECTED_KEY_ERROR:
+        case CMT_DECODE_MSGPACK_UNEXPECTED_DATA_TYPE_ERROR:
+        case CMT_DECODE_MSGPACK_DICTIONARY_LOOKUP_ERROR:
+        case CMT_DECODE_MSGPACK_VERSION_ERROR:
+            goto error;
+    }
+
+    if (ret == CMT_DECODE_MSGPACK_INSUFFICIENT_DATA && off == bytes) {
+        *out_series = count;
+        *processed_bytes = pre_off;
+        return 0;
+    }
+
+error:
+    *out_series = count;
+    *processed_bytes = pre_off;
+
+    return -1;
+}
+
+int flb_mp_validate_log_chunk(const void *data, size_t bytes,
+                              int *out_records, size_t *processed_bytes)
 {
     int ret;
     int count = 0;
@@ -59,7 +120,9 @@ int flb_mp_validate_chunk(const void *data, size_t bytes,
     unsigned char *ptr;
     msgpack_object array;
     msgpack_object ts;
+    msgpack_object header;
     msgpack_object record;
+    msgpack_object metadata;
     msgpack_unpacked result;
 
     msgpack_unpacked_init(&result);
@@ -101,9 +164,25 @@ int flb_mp_validate_chunk(const void *data, size_t bytes,
             goto error;
         }
 
-        ts = array.via.array.ptr[0];
+        header = array.via.array.ptr[0];
         record = array.via.array.ptr[1];
 
+        if (header.type == MSGPACK_OBJECT_ARRAY) {
+            if (header.via.array.size != 2) {
+                goto error;
+            }
+
+            ts = header.via.array.ptr[0];
+            metadata = header.via.array.ptr[1];
+
+            if (metadata.type != MSGPACK_OBJECT_MAP) {
+                goto error;
+            }
+        }
+        else {
+            ts = header;
+        }
+
         if (ts.type != MSGPACK_OBJECT_POSITIVE_INTEGER &&
             ts.type != MSGPACK_OBJECT_FLOAT &&
             ts.type != MSGPACK_OBJECT_EXT) {
@@ -127,6 +206,7 @@ int flb_mp_validate_chunk(const void *data, size_t bytes,
     msgpack_unpacked_destroy(&result);
     *out_records = count;
     *processed_bytes = pre_off;
+
     return -1;
 }
 
@@ -288,6 +368,37 @@ void flb_mp_array_header_end(struct flb_mp_map_header *mh)
     flb_mp_set_array_header_size(ptr, mh->entries);
 }
 
+static int insert_by_subkey_count(struct flb_record_accessor *ra, struct flb_mp_accessor *mpa)
+{
+    int subkey_count;
+    int count;
+    struct mk_list *h;
+    struct flb_record_accessor *val_ra;
+
+    /*
+     * sort flb_record_accessor by number of subkey
+     *
+     *  e.g.
+     *    $kubernetes
+     *    $kubernetes[2]['a']
+     *    $kubernetes[2]['annotations']['fluentbit.io/tag']
+     */
+    subkey_count = flb_ra_subkey_count(ra);
+    mk_list_foreach(h, &mpa->ra_list) {
+        val_ra = mk_list_entry(h, struct flb_record_accessor, _head);
+        count = flb_ra_subkey_count(val_ra);
+        if (count >=  subkey_count) {
+            mk_list_add_before(&ra->_head, &val_ra->_head, &mpa->ra_list);
+            return 0;
+        }
+    }
+
+    /* add to tail of list */
+    mk_list_add(&ra->_head, &mpa->ra_list);
+    return 0;
+}
+
+
 /*
  * Create an 'mp accessor' context: this context allows to create a list of
  * record accessor patterns based on a 'slist' context, where every slist string
@@ -320,7 +431,7 @@ struct flb_mp_accessor *flb_mp_accessor_create(struct mk_list *slist_patterns)
             flb_mp_accessor_destroy(mpa);
             return NULL;
         }
-        mk_list_add(&ra->_head, &mpa->ra_list);
+        insert_by_subkey_count(ra, mpa);
     }
 
     if (mk_list_size(&mpa->ra_list) == 0) {
