@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -20,6 +19,7 @@
 
 #include <fluent-bit/flb_info.h>
 #include <fluent-bit/flb_log.h>
+#include <fluent-bit/flb_event_loop.h>
 #include <fluent-bit/flb_network.h>
 #include <fluent-bit/flb_scheduler.h>
 #include <fluent-bit/flb_output_plugin.h>
@@ -116,7 +116,7 @@ static int upstream_thread_create(struct flb_out_thread_instance *th_ins,
     struct flb_upstream *th_u;
 
     mk_list_foreach(head, &ins->upstreams) {
-        u = mk_list_entry(head, struct flb_upstream, _head);
+        u = mk_list_entry(head, struct flb_upstream, base._head);
 
         th_u = flb_calloc(1, sizeof(struct flb_upstream));
         if (!th_u) {
@@ -125,7 +125,7 @@ static int upstream_thread_create(struct flb_out_thread_instance *th_ins,
         }
         th_u->parent_upstream = u;
         flb_upstream_queue_init(&th_u->queue);
-        mk_list_add(&th_u->_head, &th_ins->upstreams);
+        mk_list_add(&th_u->base._head, &th_ins->upstreams);
     }
 
     return 0;
@@ -138,7 +138,7 @@ int count_upstream_busy_connections(struct flb_out_thread_instance *th_ins)
     struct flb_upstream *u;
 
     mk_list_foreach(head, &th_ins->upstreams) {
-        u = mk_list_entry(head, struct flb_upstream, _head);
+        u = mk_list_entry(head, struct flb_upstream, base._head);
         c += mk_list_size(&u->queue.busy_queue);
     }
 
@@ -152,7 +152,7 @@ static void upstream_thread_destroy(struct flb_out_thread_instance *th_ins)
     struct flb_upstream *th_u;
 
     mk_list_foreach_safe(head, tmp, &th_ins->upstreams) {
-        th_u = mk_list_entry(head, struct flb_upstream, _head);
+        th_u = mk_list_entry(head, struct flb_upstream, base._head);
         flb_upstream_destroy(th_u);
     }
 }
@@ -176,11 +176,11 @@ static void output_thread(void *data)
     struct mk_event *event;
     struct flb_sched *sched;
     struct flb_task *task;
-    struct flb_upstream_conn *u_conn;
+    struct flb_connection *u_conn;
     struct flb_output_instance *ins;
-    struct flb_output_coro *out_coro;
+    struct flb_output_flush *out_flush;
     struct flb_out_thread_instance *th_ins = data;
-    struct flb_out_coro_params *params;
+    struct flb_out_flush_params *params;
     struct flb_net_dns dns_ctx;
 
     /* Register thread instance */
@@ -229,6 +229,8 @@ static void output_thread(void *data)
     snprintf(tmp, sizeof(tmp) - 1, "flb-out-%s-w%i", ins->name, thread_id);
     mk_utils_worker_rename(tmp);
 
+    memset(&event_local, 0, sizeof(struct mk_event));
+
     /* Channel used by flush callbacks to notify it return status */
     ret = mk_event_channel_create(th_ins->evl,
                                   &th_ins->ch_thread_events[0],
@@ -246,7 +248,8 @@ static void output_thread(void *data)
     /* Thread event loop */
     while (running) {
         mk_event_wait(th_ins->evl);
-        mk_event_foreach(event, th_ins->evl) {
+        flb_event_priority_live_foreach(event, th_ins->evl_bktq, th_ins->evl,
+                                      FLB_ENGINE_LOOP_MAX_ITER) {
             /*
              * FIXME
              * -----
@@ -276,7 +279,7 @@ static void output_thread(void *data)
                  * If the address receives 0xdeadbeef, means the thread must
                  * be terminated.
                  */
-                if ((uint64_t) task == 0xdeadbeef) {
+                if (task == (struct flb_task *) 0xdeadbeef) {
                     stopping = FLB_TRUE;
                     flb_plg_info(th_ins->ins, "thread worker #%i stopping...",
                                  thread_id);
@@ -284,17 +287,14 @@ static void output_thread(void *data)
                 }
 
                 /* Start the co-routine with the flush callback */
-                out_coro = flb_output_coro_create(task,
-                                                  task->i_ins,
-                                                  th_ins->ins,
-                                                  th_ins->config,
-                                                  task->buf, task->size,
-                                                  task->tag,
-                                                  task->tag_len);
-                if (!out_coro) {
+                out_flush = flb_output_flush_create(task,
+                                                    task->i_ins,
+                                                    th_ins->ins,
+                                                    th_ins->config);
+                if (!out_flush) {
                     continue;
                 }
-                flb_coro_resume(out_coro->coro);
+                flb_coro_resume(out_flush->coro);
             }
             else if (event->type == FLB_ENGINE_EV_CUSTOM) {
                 event->handler(event);
@@ -304,10 +304,11 @@ static void output_thread(void *data)
                  * Check if we have some co-routine associated to this event,
                  * if so, resume the co-routine
                  */
-                u_conn = (struct flb_upstream_conn *) event;
-                if (u_conn->coro) {
-                    flb_trace("[engine] resuming coroutine=%p", u_conn->coro);
-                    flb_coro_resume(u_conn->coro);
+                u_conn = (struct flb_connection *) event;
+
+                if (u_conn->coroutine) {
+                    flb_trace("[engine] resuming coroutine=%p", u_conn->coroutine);
+                    flb_coro_resume(u_conn->coroutine);
                 }
             }
             else if (event->type == FLB_ENGINE_EV_OUTPUT) {
@@ -332,7 +333,7 @@ static void output_thread(void *data)
         flb_sched_timer_cleanup(sched);
 
         /* Check if we should stop the event loop */
-        if (stopping == FLB_TRUE && mk_list_size(&th_ins->coros) == 0) {
+        if (stopping == FLB_TRUE && mk_list_size(&th_ins->flush_list) == 0) {
             /*
              * If there are no busy network connections (and no coroutines) its
              * safe to stop it.
@@ -356,13 +357,14 @@ static void output_thread(void *data)
     upstream_thread_destroy(th_ins);
     flb_upstream_conn_active_destroy_list(&th_ins->upstreams);
     flb_upstream_conn_pending_destroy_list(&th_ins->upstreams);
-    mk_event_loop_destroy(th_ins->evl);
 
     flb_sched_destroy(sched);
-    params = FLB_TLS_GET(out_coro_params);
+    params = FLB_TLS_GET(out_flush_params);
     if (params) {
         flb_free(params);
     }
+    mk_event_loop_destroy(th_ins->evl);
+    flb_bucket_queue_destroy(th_ins->evl_bktq);
 
     flb_plg_info(ins, "thread worker #%i stopped", thread_id);
 }
@@ -404,6 +406,7 @@ int flb_output_thread_pool_create(struct flb_config *config,
     struct flb_tp *tp;
     struct flb_tp_thread *th;
     struct mk_event_loop *evl;
+    struct flb_bucket_queue *evl_bktq;
     struct flb_out_thread_instance *th_ins;
 
     /* Create the thread pool context */
@@ -431,10 +434,10 @@ int flb_output_thread_pool_create(struct flb_config *config,
 
         th_ins->config = config;
         th_ins->ins = ins;
-        th_ins->coro_id = 0;
-        mk_list_init(&th_ins->coros);
-        mk_list_init(&th_ins->coros_destroy);
-        pthread_mutex_init(&th_ins->coro_mutex, NULL);
+        th_ins->flush_id = 0;
+        mk_list_init(&th_ins->flush_list);
+        mk_list_init(&th_ins->flush_list_destroy);
+        pthread_mutex_init(&th_ins->flush_mutex, NULL);
         mk_list_init(&th_ins->upstreams);
 
         upstream_thread_create(th_ins, ins);
@@ -446,7 +449,15 @@ int flb_output_thread_pool_create(struct flb_config *config,
             flb_free(th_ins);
             continue;
         }
+        evl_bktq = flb_bucket_queue_create(FLB_ENGINE_PRIORITY_COUNT);
+        if (!evl_bktq) {
+            flb_plg_error(ins, "could not create thread event loop bucket queue");
+            flb_free(evl);
+            flb_free(th_ins);
+            continue;
+        }
         th_ins->evl = evl;
+        th_ins->evl_bktq = evl_bktq;
 
         /*
          * Event loop setup between parent engine and this thread
@@ -464,11 +475,13 @@ int flb_output_thread_pool_create(struct flb_config *config,
         if (ret == -1) {
             flb_plg_error(th_ins->ins, "could not create thread channel");
             mk_event_loop_destroy(th_ins->evl);
+            flb_bucket_queue_destroy(th_ins->evl_bktq);
             flb_free(th_ins);
             continue;
         }
         /* Signal type to indicate a "flush" request */
         th_ins->event.type = FLB_ENGINE_EV_THREAD_OUTPUT;
+        th_ins->event.priority = FLB_ENGINE_PRIORITY_THREAD;
 
         /* Spawn the thread */
         th = flb_tp_thread_create(tp, output_thread, th_ins, config);
@@ -500,9 +513,9 @@ int flb_output_thread_pool_coros_size(struct flb_output_instance *ins)
 
         th_ins = th->params.data;
 
-        pthread_mutex_lock(&th_ins->coro_mutex);
-        n = mk_list_size(&th_ins->coros);
-        pthread_mutex_unlock(&th_ins->coro_mutex);
+        pthread_mutex_lock(&th_ins->flush_mutex);
+        n = mk_list_size(&th_ins->flush_list);
+        pthread_mutex_unlock(&th_ins->flush_mutex);
 
         size += n;
     }
@@ -513,7 +526,7 @@ int flb_output_thread_pool_coros_size(struct flb_output_instance *ins)
 void flb_output_thread_pool_destroy(struct flb_output_instance *ins)
 {
     int n;
-    uint64_t stop = 0xdeadbeef;
+    struct flb_task *stop = (struct flb_task *) 0xdeadbeef;
     struct flb_tp *tp = ins->tp;
     struct mk_list *head;
     struct flb_out_thread_instance *th_ins;
