@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2024 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -143,6 +142,10 @@ void flb_worker_destroy(struct flb_worker *worker)
         return;
     }
 
+    if (worker->log_cache) {
+        flb_log_cache_destroy(worker->log_cache);
+    }
+
     mk_list_del(&worker->_head);
     flb_free(worker);
 }
